/* packet-gsm0808.c
 * Routines for GSM 08.08 message dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm0808.c 1.3 2007/03/16 16:23:22GMT Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"

static void dissect_gsm0808_eie(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0808_eie_extra_information(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0808_eie_current_channel_type_2(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0808_ie_l3info(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0808_ie_l3_msg_contents(int msg_type, tvbuff_t *, packet_info *, proto_tree *, proto_tree *, void *extra);
static void dissect_gsm0808_ie_resource_available(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_layer_3_header_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_encryption_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_chosen_encryption_algorithm(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_channel_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_current_channel_type_1(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_periodicity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_total_resource_accessible(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_resource_situation(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_cell_identifier(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_mobile_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_priority(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_diagnostics(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_chosen_channel(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_cic(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);
static void dissect_gsm0808_ie_rr_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra);

/* Initialize the protocol and registered fields */
static int proto_gsm0808 = -1;
static int hf_gsm0808_msg_type = -1;
static int hf_ie_ei = -1;
static int hf_eie_ei = -1;
static int hf_ie_len8 = -1;
static int hf_ie_len16 = -1;
static int hf_eie_len = -1;
static int hf_ie_cic = -1;
static int hf_ie_cause = -1;
static int hf_ie_rr_cause = -1;
static int hf_ie_ci = -1;
static int hf_ie_lac = -1;
static int hf_ie_sac = -1;
static int hf_ie_rnc_id = -1;

/* Initialize the subtree pointers */
static gint ett_gsm0808 = -1;
static gint ett_ie = -1;
static gint ett_eie = -1;

static dissector_handle_t data_handle;
static dissector_handle_t gsm0808_handle;
static dissector_handle_t gsm0408_handle;
static dissector_handle_t gsm0408_l3_contents_handle;

/* useful statics */
static char hex_digit[]="0123456789abcdef";

/* Value Strings */
static const value_string gsm0808_msg_type_vals[] = {
	{ 0x01, "Assignment Request" },
	{ 0x02, "Assignment Complete" },
	{ 0x03, "Assignment Failure" },
	{ 0x10, "Handover Request" },
	{ 0x11, "Handover Required" },
	{ 0x12, "Handover Request Ack" },
	{ 0x13, "Handover Command" },
	{ 0x14, "Handover Complete" },
	{ 0x15, "Handover Succeeded" },
	{ 0x16, "Handover Failure" },
	{ 0x17, "Handover Performed" },
	{ 0x18, "Handover Candidate Enquire" },
	{ 0x19, "Handover Candidate Response" },
	{ 0x1a, "Handover Required Reject" },
	{ 0x1b, "Handover Detect" },
	{ 0x20, "Clear Command" },
	{ 0x21, "Clear Complete" },
	{ 0x22, "Clear Request" },
	{ 0x23, "Reserved" },
	{ 0x24, "Reserved" },
	{ 0x25, "SAPI N Reject" },
	{ 0x26, "Confusion" },
	{ 0x28, "Suspend" },
	{ 0x29, "Resume" },
	{ 0x2a, "Connection Orientated Information" },
	{ 0x2b, "Perform Location Request" },
	{ 0x2c, "LSA Information" },
	{ 0x2d, "Perform Location Response" },
	{ 0x2e, "Perform Location Abort" },
	{ 0x2f, "Common ID" },
	{ 0x30, "Reset" },
	{ 0x31, "Reset Ack" },
	{ 0x32, "Overload" },
	{ 0x33, "Reserved" },
	{ 0x34, "Reset Circuit" },
	{ 0x35, "Reset Circuit Ack" },
	{ 0x36, "MSC Invoke Trace" },
	{ 0x37, "BSC Invoke Trace" },
	{ 0x40, "Block" },
	{ 0x41, "Block Ack" },
	{ 0x42, "Unblock" },
	{ 0x43, "Unblock Ack" },
	{ 0x44, "Circuit Group Block" },
	{ 0x45, "Circuit Group Block Ack" },
	{ 0x46, "Circuit Group Unblock" },
	{ 0x47, "Circuit Group Unblock Ack" },
	{ 0x48, "Unequipped Circuit" },
	{ 0x4e, "Change Circuit" },
	{ 0x4f, "Change Circuit Ack" },
	{ 0x50, "Resource Request" },
	{ 0x51, "Resource Indication" },
	{ 0x52, "Paging" },
	{ 0x53, "Cipher Mode Command" },
	{ 0x54, "Classmark Update" },
	{ 0x55, "Cipher Mode Complete" },
	{ 0x56, "Queueing Indication" },
	{ 0x57, "Complete Layer 3 Info" },
	{ 0x58, "Classmark Request" },
	{ 0x59, "Cipher Mode Reject" },
	{ 0x5a, "Load Indication" },
	{ 0x04, "VGCS/VBS Setup" },
	{ 0x05, "VGCS/VBS Setup Ack" },
	{ 0x06, "VGCS/VBS Setup Refuse" },
	{ 0x07, "VGCS/VBS Assignment Request" },
	{ 0x1c, "VGCS/VBS Assignment Result" },
	{ 0x1d, "VGCS/VBS Assignment Failure" },
	{ 0x1e, "VGCS/VBS Queueing Indication" },
	{ 0x1f, "Uplink Request" },
	{ 0x27, "Uplink Request Ack" },
	{ 0x49, "Uplink Request Confirm" },
	{ 0x4a, "Uplink Release Indication" },
	{ 0x4b, "Uplink Reject Command" },
	{ 0x4c, "Uplink Release Command" },
	{ 0x4d, "Uplink Seized Command" },
	{ 0, NULL }
};

static const value_string ie_ei_vals[] = {
	{ 0x01, "CIC" },
	{ 0x03, "Resource Available" },
	{ 0x04, "Cause" },
	{ 0x05, "Cell Identifier" },
	{ 0x06, "Priority" },
	{ 0x07, "Layer 3 Header Info" },
	{ 0x08, "IMSI" },
	{ 0x09, "TMSI" },
	{ 0x0a, "Encryption Info" },
	{ 0x0b, "Channel Type" },
	{ 0x0c, "Periodicity" },
	{ 0x0d, "Extended Resource Indicator" },
	{ 0x0e, "Number of MSs" },
	{ 0x12, "Classmark Info Type 2" },
	{ 0x13, "Classmark Info Type 3" },
	{ 0x14, "Interference Band to be Used" },
	{ 0x15, "RR Cause" },
	{ 0x17, "Layer 3 Info" },
	{ 0x18, "DLCI" },
	{ 0x19, "Downlink DTX Flag" },
	{ 0x1a, "Cell Identifier List" },
	{ 0x1b, "Response Request" },
	{ 0x1c, "Resource Indication Method" },
	{ 0x1d, "Classmark Info Type 1" },
	{ 0x1e, "CIC List" },
	{ 0x1f, "Diagnostic" },
	{ 0x20, "Layer 3 Message Contents" },
	{ 0x21, "Chosen Channel" },
	{ 0x22, "Total Resource Available" },
	{ 0x23, "Cipher Response Mode" },
	{ 0x24, "Channel Needed" },
	{ 0x25, "Trace Type" },
	{ 0x26, "Trigger id" },
	{ 0x27, "Trace Ref" },
	{ 0x28, "Transaction Id" },
	{ 0x29, "Mobile Identity" },
	{ 0x2a, "OMD Id" },
	{ 0x2b, "Forward Indicator" },
	{ 0x2c, "Chosen Encryption Algorithm" },
	{ 0x2d, "Circuit Pool" },
	{ 0x2e, "Circuit Pool List" },
	{ 0x2f, "Time Indication" },
	{ 0x30, "Resource Situation" },
	{ 0x31, "Current Channel Type 1" },
	{ 0x32, "Queueing Indicator" },
	{ 0x40, "Speech Version" },
	{ 0x33, "Assignment Requirement" },
	{ 0x35, "Talker Flag" },
	{ 0x36, "Connection Release Requested" },
	{ 0x37, "Group Call Reference" },
	{ 0x38, "eMLPP Priority" },
	{ 0x39, "Configuration Evolution Indicator" },
	{ 0x3a, "Old BSS to New BSS Info" },
	{ 0x3b, "LSA Identifier" },
	{ 0x3c, "LSA Identifier List" },
	{ 0x3d, "LSA Information" },
	{ 0x3e, "LCS QoS" },
	{ 0x3f, "LSA Access Control Suppression" },
	{ 0x43, "LCS Priority" },
	{ 0x44, "LCS Location Type" },
	{ 0x45, "LCS Location Estimate" },
	{ 0x46, "LCS Positioning Data" },
	{ 0x47, "LCS LCS Cause" },
	{ 0x48, "LCS LCS Client Type" },
	{ 0x49, "LCS APDU" },
	{ 0x4a, "LCS Network Element Identity" },
	{ 0x4b, "LCS GPS Assistance Data" },
	{ 0x4c, "LCS Deciphering Keys" },
	{ 0x4d, "Return Error Request" },
	{ 0x4e, "Return Error Cause" },
	{ 0x4f, "Segmentation" },
	{ 0x50, "Service Handover" },
	{ 0x41, "Source RNC to target RNC UMTS" },
	{ 0x42, "Source RNC to target RNC CDMA" },
	{ 0, NULL }
};

static const value_string cause_vals[] = {
	{ 0x00, "Radio interface message failure"},
	{ 0x01, "Radio interface failure"},
	{ 0x02, "Uplink quality"},
	{ 0x03, "Uplink strength"},
	{ 0x04, "Downlink quality"},
	{ 0x05, "Downlink strength"},
	{ 0x06, "Distance"},
	{ 0x07, "O and M intervention"},
	{ 0x08, "Response to MSC invocation"},
	{ 0x09, "Call control"},
	{ 0x0a, "Radio interface failure, reversion to old channel"},
	{ 0x0b, "Handover successful"},
	{ 0x0c, "Better Cell"},
	{ 0x0d, "Directed Retry"},
	{ 0x0e, "Joined group call channel"},
	{ 0x0f, "Traffic"},
	{ 0x20, "Equipment failure"},
	{ 0x21, "No radio resource available"},
	{ 0x22, "Requested terrestrial resource unavailable"},
	{ 0x23, "CCCH overload"},
	{ 0x24, "Processor overload"},
	{ 0x25, "BSS not equipped"},
	{ 0x26, "MS not equipped"},
	{ 0x27, "Invalid cell"},
	{ 0x28, "Traffic Load"},
	{ 0x29, "Preemption"},
	{ 0x30, "Requested transcoding/rate adaption unavailable"},
	{ 0x31, "Circuit pool mismatch"},
	{ 0x32, "Switch circuit pool"},
	{ 0x33, "Requested speech version unavailable"},
	{ 0x34, "LSA not allowed"},
	{ 0x40, "Ciphering algorithm not supported"},
	{ 0x50, "Terrestrial circuit already allocated"},
	{ 0x51, "Invalid message contents"},
	{ 0x52, "Information element or field missing"},
	{ 0x53, "Incorrect value"},
	{ 0x54, "Unknown Message type"},
	{ 0x55, "Unknown Information Element"},
	{ 0x60, "Protocol Error between BSS and MSC"},
	{ 0x61, "VGCS/VBS call non existent"},
	{ 0, NULL }
};

static const value_string rr_cause_vals[] = {
	{ 0x00, "Normal event" },
	{ 0x01, "Abnormal release, unspecified" },
	{ 0x02, "Abnormal release, channel unacceptable" },
	{ 0x03, "Abnormal release, timer expired" },
	{ 0x04, "Abnormal release, no activity on the radio path" },
	{ 0x05, "Preemptive release" },
	{ 0x06, "UTRAN configuration unknown" },
	{ 0x08, "Handover impossible, timing advance out of range" },
	{ 0x09, "Channel mode unacceptable" },
	{ 0x0a, "Frequency not implemented" },
	{ 0x0c, "Lower layer failure" },
	{ 0x21, "Call already cleared" },
	{ 0x5f, "Semantically incorrect message" },
	{ 0x60, "Invalid mandatory information" },
	{ 0x61, "Message type non-existent or not implemented" },
	{ 0x62, "Message type not compatible with protocol state" },
	{ 0x64, "Conditional IE error" },
	{ 0x65, "No cell allocation available" },
	{ 0x6f, "Protocol error unspecified" },
	{ 0, NULL }
};

static const value_string downlink_dtx_flag_vals[] = {
	{ 0x00, "Downlink DTX normal" },
	{ 0x01, "BSS forbidden to use downlink DTX on this channel" },
	{ 0, NULL }
};

static const value_string speech_data_vals[] = {
	{ 0x01, "Speech" },
	{ 0x02, "Data" },
	{ 0x03, "Signalling" },
	{ 0, NULL }
};
#define SD_SPEECH		0x01
#define SD_DATA			0x02
#define SD_SIGNALLING	0x03

static const value_string channel_rate_and_type_vals[] = {
	{ 0x00, "SDCCH or Full or Half rate TCH" },
	{ 0x01, "SDCCH" },
	{ 0x02, "SDCCH or Full rate TCH" },
	{ 0x03, "SDCCH or Half rate TCH" },
	{ 0x08, "Full rate TCH/F" },
	{ 0x09, "Half Rate TCH/H" },
	{ 0x0a, "Full or Half rate TCH, Full rate preferred, changes allowed" },
	{ 0x0b, "Full or Half rate TCH, Half rate preferred, changes allowed" },
	{ 0x1a, "Full or Half rate TCH, Full rate preferred, changes not allowed" },
	{ 0x1b, "Full or Half rate TCH, Half rate preferred, changes not allowed" },
	{ 0x0f, "Full or Half rate TCH, changes allowed" },
	{ 0x1f, "Full or Half rate TCH, changes allowed" },
	{ 0x20, "Multislot 1 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x21, "Multislot 2 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x22, "Multislot 3 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x23, "Multislot 4 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x24, "Multislot 5 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x25, "Multislot 6 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x26, "Multislot 7 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x27, "Multislot 8 TCH/F, changes to number of TCH or air interface rate by BSS allowed" },
	{ 0x30, "Multislot 1 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x31, "Multislot 2 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x32, "Multislot 3 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x33, "Multislot 4 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x34, "Multislot 5 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x35, "Multislot 6 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x36, "Multislot 7 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0x37, "Multislot 8 TCH/F, changes to number of TCH or air interface rate by BSS not allowed" },
	{ 0, NULL }
};

static const value_string speech_version_vals[] = {
	{ 0x01, "Full rate vesion 1 (FR)" },
	{ 0x11, "Full rate vesion 2 (EFR)" },
	{ 0x21, "Full rate vesion 3 (AMR/F)" },
	{ 0x05, "Half rate vesion 1 (HR)" },
	{ 0x15, "Half rate vesion 2 (N/A)" },
	{ 0x25, "Half rate vesion 3 (AMR/H)" },
	{ 0, NULL }
};

static const value_string nt_radio_data_rate_vals[] = {
	{ 0x00, "12 kb/s FR or 6kb/s HR" },
	{ 0x34, "43.5kb/s" },
	{ 0x31, "29kb/s" },
	{ 0x18, "14.5kb/s" },
	{ 0x10, "12kb/s" },
	{ 0x11, "6kb/s" },
	{ 0, NULL }
};

static const value_string t_radio_data_rate_vals[] = {
	{ 0x39, "28.8kb/s" },
	{ 0x18, "14.4kb/s" },
	{ 0x10, "9.6kb/s" },
	{ 0x11, "4.8kb/s" },
	{ 0x12, "2.4kb/s" },
	{ 0x13, "1.2kb/s" },
	{ 0x14, "600b/s" },
	{ 0x15, "1200/75b/s" },
	{ 0, NULL }
};

static const value_string allowed_air_interface_rates_vals[] = {
	{ 0x01, "6kb/s" },
	{ 0x02, "12kb/s" },
	{ 0x08, "14.5kb/s" },
	{ 0x10, "29kb/s" },
	{ 0x40, "43.5kb/s" },
	{ 0, NULL }
};

static const value_string chosen_channel_mode_vals[] = {
	{ 0x00, "no channel mode indication" },
	{ 0x09, "speech (full rate or half rate)" },
	{ 0x0e, "data 14.5kb/s radio interface" },
	{ 0x0b, "data 12.0kb/s radio interface" },
	{ 0x0c, "data 6.0kb/s radio interface" },
	{ 0x0d, "data 3.6kb/s radio interface" },
	{ 0x08, "signalling only" },
	{ 0x01, "data 29.0kb/s radio interface" },
	{ 0x02, "data 32.0kb/s radio interface" },
	{ 0x03, "data 43.5kb/s radio interface" },
	{ 0x04, "data 43.5kb/s downlink and 14.5kb/s uplink" },
	{ 0x05, "data 29.0kb/s downlink and 14.5kb/s uplink" },
	{ 0x06, "data 43.5kb/s downlink and 29.0kb/s uplink" },
	{ 0x07, "data 14.5kb/s downlink and 43.5kb/s uplink" },
	{ 0x0a, "data 14.5kb/s downlink and 29.0kb/s uplink" },
	{ 0x0f, "data 29.0kb/s downlink and 43.5kb/s uplink" },
	{ 0, NULL }
};

static const value_string current_channel_mode_vals[] = {
	{ 0x00, "signalling only" },
	{ 0x01, "speech (full rate or half rate)" },
	{ 0x06, "data 14.5kb/s radio interface" },
	{ 0x03, "data 12.0kb/s radio interface" },
	{ 0x04, "data 6.0kb/s radio interface" },
	{ 0x05, "data 3.6kb/s radio interface" },
	{ 0x07, "data 29.0kb/s radio interface" },
	{ 0x08, "data 32.0kb/s radio interface" },
	{ 0x09, "data 43.5kb/s radio interface" },
	{ 0x02, "data 43.5kb/s downlink and 14.5kb/s uplink" },
	{ 0x0a, "data 29.0kb/s downlink and 14.5kb/s uplink" },
	{ 0x0b, "data 43.5kb/s downlink and 29.0kb/s uplink" },
	{ 0x0c, "data 14.5kb/s downlink and 43.5kb/s uplink" },
	{ 0x0d, "data 14.5kb/s downlink and 29.0kb/s uplink" },
	{ 0x0e, "data 29.0kb/s downlink and 43.5kb/s uplink" },
	{ 0, NULL }
};

static const value_string chosen_channel_vals[] = {
	{ 0x00, "None" },
	{ 0x01, "SDCCH" },
	{ 0x08, "1 Full rate TCH" },
	{ 0x09, "1 Half rate TCH" },
	{ 0x0a, "2 Full rate TCHs" },
	{ 0x0b, "3 Full rate TCHs" },
	{ 0x0c, "4 Full rate TCHs" },
	{ 0x0d, "5 Full rate TCHs" },
	{ 0x0e, "6 Full rate TCHs" },
	{ 0x0f, "7 Full rate TCHs" },
	{ 0x04, "8 Full rate TCHs" },
	{ 0, NULL }
};

static const value_string asymmetry_indication_vals[] = {
	{ 0x00, "Not applicable (symmetric)" },
	{ 0x01, "Uplink biased asymmetry" },
	{ 0x02, "Downlink biased asymmetry" },
	{ 0, NULL }
};

static const value_string cell_id_desc_vals[] = {
	{ 0x00, "CGI" },
	{ 0x01, "LAC+CI" },
	{ 0x02, "CI" },
	{ 0x03, "No cell associated" },
	{ 0x04, "LAI" },
	{ 0x05, "LAC" },
	{ 0x06, "All BSS Cells" },
	{ 0x08, "UMTS PLMN+LAC+RNC" },
	{ 0x09, "UMTS RNC" },
	{ 0x0a, "UMTS LAC+RNC" },
	{ 0x0b, "UMTS SAI" },
	{ 0x0c, "UMTS LAC+RNC-ID+CI" },
	{ 0, NULL }
};
#define CELL_ID_DESC_CGI					0x00
#define CELL_ID_DESC_LAC_CI					0x01
#define CELL_ID_DESC_CI						0x02
#define CELL_ID_DESC_NA						0x03
#define CELL_ID_DESC_LAI					0x04
#define CELL_ID_DESC_LAC					0x05
#define CELL_ID_DESC_ALL_BSS				0x06
#define CELL_ID_DESC_UMTS_PLMN_LAC_RNC		0x08
#define CELL_ID_DESC_UMTS_RNC				0x09
#define CELL_ID_DESC_UMTS_LAC_RNC			0x0a
#define CELL_ID_DESC_UMTS_SAI				0x0b
#define CELL_ID_DESC_UMTS_LAC_RNC_CI		0x0c

static const value_string resource_indication_method_vals[] = {
	{ 0x00, "Method 1: Spontaneous (O&M configured)" },
	{ 0x01, "Method 2: Single response" },
	{ 0x02, "Method 3: Periodic responses" },
	{ 0x03, "Method 4: None" },
	{ 0, NULL }
};

static const value_string channel_needed_vals[] = {
	{ 0x00, "Any channel" },
	{ 0x01, "SDCCH" },
	{ 0x02, "TCH/F (full rate)" },
	{ 0x03, "TCH/H or TCH/F (dual rate)" },
	{ 0, NULL }
};

static const value_string cipher_response_mode_vals[] = {
	{ 0x00, "IMEISV must not be included by the MS" },
	{ 0x01, "IMEISV must be included by the MS" },
	{ 0, NULL }
};

static const value_string forward_indicator_vals[] = {
	{ 0x01, "Forward to subsequent BSS, no trace to MSC" },
	{ 0x02, "Forward to subsequent BSS, and trace to MSC" },
	{ 0, NULL }
};

static const value_string resource_chan_type_vals[] = {
	{ 0x01, "SDCCH" },
	{ 0x08, "TCH/F (full rate)" },
	{ 0x09, "TCH/H (half rate)" },
	{ 0, NULL }
};

static const value_string identity_type_vals[] = {
	{ 0x00, "No Identity" },
	{ 0x01, "IMSI" },
	{ 0x02, "IMEI" },
	{ 0x03, "IMEISV" },
	{ 0x04, "TMSI/P-TMSI" },
	{ 0, NULL }
};
#define MOBILE_IDENTITY_TYPE_TMSI	0x04

static const value_string resource_band_vals[] = {
	{ 0x00, "Total number of accessible channels" },
	{ 0x01, "Number of channels available in Band 1" },
	{ 0x02, "Number of channels available in Band 2" },
	{ 0x03, "Number of channels available in Band 3" },
	{ 0x04, "Number of channels available in Band 4" },
	{ 0x05, "Number of channels available in Band 5" },
	{ 0x0e, "Number of channels available without band classification" },
	{ 0, NULL }
};

static const value_string queueing_indicator_vals[] = {
	{ 0x00, "It is recommended not to allow queueing" },
	{ 0x02, "It is recommended to allow queueing" },
	{ 0, NULL }
};

static const value_string assignment_requirement_vals[] = {
	{ 0x00, "Delay allowed" },
	{ 0x01, "Immediate, and the resources shall not be de-allocated until the end of the call" },
	{ 0x02, "Immediate, and the resources may further be de-allocated by the BSS" },
	{ 0, NULL }
};

static const value_string configuration_evolution_indicator_vals[] = {
	{ 0x00, "No modification is allowed" },
	{ 0x01, "Modification is allowed and maximum number of TCH/F is 1" },
	{ 0x02, "Modification is allowed and maximum number of TCH/F is 2" },
	{ 0x03, "Modification is allowed and maximum number of TCH/F is 3" },
	{ 0x04, "Modification is allowed and maximum number of TCH/F is 4" },
	{ 0, NULL }
};

static const value_string extra_information_lcs_vals[] = {
	{ 0x00, "No ongoing LCS procedure" },
	{ 0x01, "LCS procedure interrupted by handover: new BSS may notify SMLC when handover complete" },
	{ 0, NULL }
};

static const value_string extra_information_prec_vals[] = {
	{ 0x00, "this allocation should not cause a preemption of an existing connection" },
	{ 0x01, "this allocation is allowed to preempt an existing connection" },
	{ 0, NULL }
};

static const ie_def_t ie_defs[] = {
	{ 0x01, "CIC", 2, &dissect_gsm0808_ie_cic, NULL },
	{ 0x03, "Resource Available", 20, &dissect_gsm0808_ie_resource_available, NULL },
	{ 0x04, "Cause", -1, &dissect_gsm0808_ie_cause, (void *)cause_vals },
	{ 0x05, "Cell Identifier", -1, &dissect_gsm0808_ie_cell_identifier, NULL },
	{ 0x06, "Priority", -1, &dissect_gsm0808_ie_priority, NULL },
	{ 0x07, "Layer 3 Header Info", -1, &dissect_gsm0808_ie_layer_3_header_information, NULL },
	{ 0x08, "IMSI", -1, NULL, NULL },
	{ 0x09, "TMSI", -1, NULL, NULL },
	{ 0x0a, "Encryption Info", -1, &dissect_gsm0808_ie_encryption_information, NULL },
	{ 0x0b, "Channel Type", -1, &dissect_gsm0808_ie_channel_type, NULL },
	{ 0x0c, "Periodicity", 1, &dissect_gsm0808_ie_periodicity, NULL },
	{ 0x0d, "Extended Resource Indicator", 1, NULL, NULL },
	{ 0x0e, "Number of MSs", 1, NULL, NULL },
	{ 0x12, "Classmark Info Type 2", -1, NULL, NULL },
	{ 0x13, "Classmark Info Type 3", -1, NULL, NULL },
	{ 0x14, "Interference Band to be Used", 1, NULL, NULL },
	{ 0x15, "RR Cause", 1, &dissect_gsm0808_ie_rr_cause, (void *)rr_cause_vals },
	{ 0x17, "Layer 3 Info", -1, &dissect_gsm0808_ie_l3info, NULL },
	{ 0x18, "DLCI", 1, NULL, NULL },
	{ 0x19, "Downlink DTX Flag", 1, NULL, (void *)downlink_dtx_flag_vals },
	{ 0x1a, "Cell Identifier List", -1, &dissect_gsm0808_ie_cell_identifier, NULL },
	{ 0x1b, "Response Request", 0, NULL, NULL },
	{ 0x1c, "Resource Indication Method", 1, NULL, (void *)resource_indication_method_vals },
	{ 0x1d, "Classmark Info Type 1", 1, NULL, NULL },
	{ 0x1e, "CIC List", -1, NULL, NULL },
	{ 0x1f, "Diagnostic", -1, &dissect_gsm0808_ie_diagnostics, NULL },
	{ 0x20, "Layer 3 Message Contents", -1, &dissect_gsm0808_ie_l3_msg_contents, NULL },
	{ 0x21, "Chosen Channel", 1, &dissect_gsm0808_ie_chosen_channel, NULL },
	{ 0x22, "Total Resource Accessible", 4, &dissect_gsm0808_ie_total_resource_accessible, NULL },
	{ 0x23, "Cipher Response Mode", 1, NULL, (void *)cipher_response_mode_vals },
	{ 0x24, "Channel Needed", 1, NULL, (void *)channel_needed_vals },
	{ 0x25, "Trace Type", 1, NULL, NULL },
	{ 0x26, "Trigger id", -1, NULL, NULL },
	{ 0x27, "Trace Ref", 2, NULL, NULL },
	{ 0x28, "Transaction Id", -1, NULL, NULL },
	{ 0x29, "Mobile Identity", -1, &dissect_gsm0808_ie_mobile_identity, NULL },
	{ 0x2a, "OMD Id", -1, NULL, NULL },
	{ 0x2b, "Forward Indicator", 1, NULL, (void *)forward_indicator_vals },
	{ 0x2c, "Chosen Encryption Algorithm", 1, &dissect_gsm0808_ie_chosen_encryption_algorithm, NULL },
	{ 0x2d, "Circuit Pool", 1, NULL, NULL },
	{ 0x2e, "Circuit Pool List", -1, NULL, NULL },
	{ 0x2f, "Time Indication", 1, NULL, NULL },
	{ 0x30, "Resource Situation", -1, &dissect_gsm0808_ie_resource_situation, NULL },
	{ 0x31, "Current Channel Type 1", 1, &dissect_gsm0808_ie_current_channel_type_1, NULL },
	{ 0x32, "Queueing Indicator", 1, NULL, (void *)queueing_indicator_vals },
	{ 0x40, "Speech Version", 1, NULL, (void *)speech_version_vals },
	{ 0x33, "Assignment Requirement", 1, NULL, (void *)assignment_requirement_vals },
	{ 0x35, "Talker Flag", 0, NULL, NULL },
	{ 0x36, "Connection Release Requested", 0, NULL, NULL },
	{ 0x37, "Group Call Reference", -1, NULL, NULL },
	{ 0x38, "eMLPP Priority", 1, NULL, NULL },
	{ 0x39, "Configuration Evolution Indicator", 1, NULL, (void *)configuration_evolution_indicator_vals },
	{ 0x3a, "Old BSS to New BSS Info", -1, &dissect_gsm0808_eie, NULL },
	{ 0x3b, "LSA Identifier", -1, NULL, NULL },
	{ 0x3c, "LSA Identifier List", -1, NULL, NULL },
	{ 0x3d, "LSA Information", -1, NULL, NULL },
	{ 0x3e, "LCS QoS", -1, NULL, NULL },
	{ 0x3f, "LSA Access Control Suppression", 1, NULL, NULL },
	{ 0x43, "LCS Priority", -1, NULL, NULL },
	{ 0x44, "LCS Location Type", -1, NULL, NULL },
	{ 0x45, "LCS Location Estimate", -1, NULL, NULL },
	{ 0x46, "LCS Positioning Data", -1, NULL, NULL },
	{ 0x47, "LCS LCS Cause", -1, NULL, NULL },
	{ 0x48, "LCS LCS Client Type", -1, NULL, NULL },
	{ 0x49, "LCS APDU", -2, NULL, NULL },
	{ 0x4a, "LCS Network Element Identity", -1, NULL, NULL },
	{ 0x4b, "LCS GPS Assistance Data", -1, NULL, NULL },
	{ 0x4c, "LCS Deciphering Keys", -1, NULL, NULL },
	{ 0x4d, "Return Error Request", -1, NULL, NULL },
	{ 0x4e, "Return Error Cause", -1, NULL, NULL },
	{ 0x4f, "Segmentation", -1, NULL, NULL },
	{ 0x50, "Service Handover", -1, NULL, NULL },
	{ 0x41, "Source RNC to target RNC UMTS", -1, NULL, NULL },
	{ 0x42, "Source RNC to target RNC CDMA", -1, NULL, NULL },
	{ 0, NULL, 0, NULL, NULL },
};

static const value_string eie_ei_vals[] = {
	{ 0x01,  "Extra Information" },
	{ 0x02,  "Current Channel Type 2" },
	{ 0x03,  "Target cell radio information" },
	{ 0x04,  "GPRS Suspend information" },
	{ 0x05,  "Multirate configuration information" },
	{ 0x06,  "Dual Transfer mode information" },
	{ 0x07,  "UE Capability information" },
	{ 0x08,  "cdma2000 Capability information" },
	{ 0, NULL }
};

static const ie_def_t eie_defs[] = {
	{ 0x01,  "Extra Information", -1, &dissect_gsm0808_eie_extra_information, NULL },
	{ 0x02,  "Current Channel Type 2", -1, &dissect_gsm0808_eie_current_channel_type_2, NULL },
	{ 0x03,  "Target cell radio information", -1, NULL, NULL },
	{ 0x04,  "GPRS Suspend information", -1, NULL, NULL },
	{ 0x05,  "Multirate configuration information", -1, NULL, NULL },
	{ 0x06,  "Dual Transfer mode information", -1, NULL, NULL },
	{ 0x07,  "UE Capability information", -1, NULL, NULL },
	{ 0x08,  "cdma2000 Capability information", -1, NULL, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

/* Code to dissect the 08.08 message */
static void
dissect_gsm0808(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "gsm0808";

	/* following call moved up to packet-sccp.c */
	/* gsmstat_set_proto(GSMSTAT_PROTO_GSM0808); */
	
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM");

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm0808, tvb, offset, msg_len, FALSE);
		tree = proto_item_add_subtree(ti, ett_gsm0808);
	}

	if (msg_len>=GSM0808_HEADER_LEN) {
		msg_type = tvb_get_guint8(tvb, offset);
		gsmstat_set_msg_type(msg_type);
		if (tree) {
			proto_item_set_text(ti, "  08.08:%s",rd_match_strval(msg_type,gsm0808_msg_type_vals));
			proto_tree_add_uint(tree, hf_gsm0808_msg_type, tvb, offset, 1, msg_type);
		}
		offset+=1;

		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, "08.08:%s(",rd_match_strval(msg_type,gsm0808_msg_type_vals));

		/* create a new subset TVB for disecting the payload, and call the correct payload disector */
		next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
		if (msg_len>offset) {
			dissect_gsm_generic_ie(msg_type, ie_defs, next_tvb, pinfo, tree, main_tree);
			/* if (tree)
				proto_item_set_text(ti, "08.08:%s",rd_match_strval(msg_type,gsm0808_msg_type_vals));*/
		}
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, ")");
	}
	else { /* Duff 08.08 or ... FIXME - partial message !!! */
		if (tree)
			proto_tree_add_text(main_tree, tvb, offset, tvb_length_remaining(tvb,offset),"Short 08.08 Message!!!");
		if (check_col(pinfo->cinfo, COL_INFO))
			col_append_fstr(pinfo->cinfo, COL_INFO, " Short 08.08 Message!!!");
	}

	return;
}

/* Code to dissect the 08.08 IEs */
static void
dissect_gsm_generic_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree)
{
	int offset=0, msg_len, i;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		ie_ei = tvb_get_guint8(tvb, offset);
		offset += 1;

		for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==ie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				ie_len=0;
				if (main_tree) {
					proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
					tree = proto_item_add_subtree(ti, ett_ie);
					proto_tree_add_uint(tree, hf_ie_ei, tvb, 0, 1, ie_ei);
				}
			}
			else if (p_ie_def->ie_len > 0) {
				ie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)(p_ie_def->info)));
						for (i=2; i<ie_len+1; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ie);
						proto_tree_add_uint(tree, hf_ie_ei, tvb, 0, 1, ie_ei);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				ie_len8 = tvb_get_guint8(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ie_len8, tvb, offset, 1, ie_len8);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+1),(value_string *)(p_ie_def->info)));
						for (i=3; i<ie_len8+2; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ie);
						proto_tree_add_uint(tree, hf_ie_ei, tvb, 0, 1, ie_ei);
					}
				}
				offset += 1;
				ie_len = ie_len8;
			}
			else if (p_ie_def->ie_len == -2) {
				ie_len16 = tvb_get_ntohs(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_ie_len16, tvb, offset, 2, ie_len16);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+2),(value_string *)(p_ie_def->info)));
						for (i=4; i<ie_len16+3; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_ie);
						proto_tree_add_uint(tree, hf_ie_ei, tvb, 0, 1, ie_ei);
					}
				}
				offset += 2;
				ie_len = ie_len16;
			}
			else {			/* should never happen */
				ie_len=0; /* safety! */
			}

			if (ie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle,new_tvb, pinfo, tree);
				offset+=ie_len;
			}
		
				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);
				dissect_gsm_generic_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree, top_tree);
			}
		}
	}
	return;
}

/* Code to dissect the Embedded IEs */
static void
dissect_gsm0808_eie( int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;
	guint8	eie_ei, eie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		eie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_eie_ei, tvb, offset, 1, eie_ei);
		offset += 1;

		for (p_ie_def=eie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==eie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				eie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				eie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s:", p_ie_def->ie_name);
						for (i=1; i<eie_len+1; i++)
							if (i==1)
								proto_item_append_text(ti, " %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
							else
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_eie);
					}
				}
			}
			else if (p_ie_def->ie_len == -1) {
				eie_len = tvb_get_guint8(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_eie_len, tvb, offset, 1, eie_len);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+2, "%s:", p_ie_def->ie_name);
						for (i=2; i<eie_len+2; i++)
							if (i==2)
								proto_item_append_text(ti, " %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
							else
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_eie);
					}
				}
				offset += 1;
			}
			
			if (eie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, eie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle,new_tvb, pinfo, tree);
				offset+=eie_len;
			}
			
				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);

				dissect_gsm0808_eie(msg_type, new_tvb, pinfo, main_tree, top_tree, extra);
			}
		}
	}
	return;
}

static void dissect_gsm0808_ie_l3info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
		call_dissector(gsm0408_handle, tvb, pinfo, top_tree);
}

static void dissect_gsm0808_ie_l3_msg_contents(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;

	if (tvb_length_remaining(tvb,offset)<12) {	/* fudge - BSC and this decoder used to erroneously treat this as a L3 Info - maintain this decode with warning!!! */
		call_dissector(gsm0408_l3_contents_handle, tvb, pinfo, top_tree);	/* correct decode */
	}
	else {
		proto_tree_add_text(top_tree, tvb, 0, tvb_length_remaining(tvb,offset), "Warning: Invalid L3 Contents - try decoding as L3 Info...");
		call_dissector(gsm0408_handle, tvb, pinfo, top_tree);			/* wrong decode */
	}
}

static void dissect_gsm0808_ie_resource_available(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, chans, msg_len, band;

	msg_len = tvb_length_remaining(tvb,offset);

	for (band=1; band<=5; band++) {
		chans = tvb_get_ntohs(tvb, offset);
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 2, "Band %d Full rate Channels available", band, chans);
		}
		offset+=2;
		chans = tvb_get_ntohs(tvb, offset);
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 2, "Band %d Half rate Channels available", band, chans);
		}
		offset+=2;
	}

	return;
}

static void dissect_gsm0808_ie_resource_situation(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, chans, msg_len, band_and_type, chans_len;

	msg_len = tvb_length_remaining(tvb,offset);

	while (offset<msg_len) {
		band_and_type = tvb_get_guint8(tvb, offset);
		chans_len=1;
		chans = tvb_get_guint8(tvb, offset+1);
		if (chans&0x80) {
			chans = (chans&0x7f)<<8 | tvb_get_guint8(tvb, offset+2);
			chans_len++;
		}

		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 1+chans_len, "Channel Type %s: %s: %d", rd_match_strval(band_and_type&0x0f,resource_chan_type_vals), rd_match_strval(band_and_type>>4,resource_band_vals), chans);
		}
		offset+=1+chans_len;
	}

	return;
}

static void dissect_gsm0808_ie_cic(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint16 cic;

	msg_len = tvb_length_remaining(tvb,offset);

	cic = tvb_get_ntohs(tvb, offset);
	gsmstat_set_cic(cic);
	if (tree) {
		proto_tree_add_uint(tree, hf_ie_cic, tvb, offset, 2, cic);
	}
	offset+=2;

	return;
}

static void dissect_gsm0808_ie_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 cause;

	msg_len = tvb_length_remaining(tvb,offset);

	cause = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_uint(tree, hf_ie_cause, tvb, offset, 1, cause);
	}
	offset+=1;

	return;
}

static void dissect_gsm0808_ie_rr_cause(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 rr_cause;

	msg_len = tvb_length_remaining(tvb,offset);

	rr_cause = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_uint(tree, hf_ie_rr_cause, tvb, offset, 1, rr_cause);
	}
	offset+=1;

	return;
}

static void dissect_gsm0808_ie_total_resource_accessible(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, chans, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	chans = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 2, "Full rate Channels available: %d", chans);
	}
	offset+=2;
	chans = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 2, "Half rate Channels available: %d", chans);
	}
	offset+=2;

	return;
}

static void dissect_gsm0808_ie_cell_identifier(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 cell_id_desc, o1, o2, o3;
	guint16 ci, lac, rnc_id, sac;

	msg_len = tvb_length_remaining(tvb,offset);

	cell_id_desc = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Cell identifier descriminator: %s", rd_match_strval(cell_id_desc,cell_id_desc_vals));
	}
	offset++;

	while (offset<msg_len) {
		switch (cell_id_desc) {
		case CELL_ID_DESC_CGI:
		case CELL_ID_DESC_UMTS_PLMN_LAC_RNC:
		case CELL_ID_DESC_LAI:
		case CELL_ID_DESC_UMTS_SAI:
		case CELL_ID_DESC_UMTS_LAC_RNC_CI:
			o1 = tvb_get_guint8(tvb, offset);
			o2 = tvb_get_guint8(tvb, offset+1);
			o3 = tvb_get_guint8(tvb, offset+2);
			if (tree) {
				if (o2>>4==0x0f)	/* 2 digit MNC */
					proto_tree_add_text(tree, tvb, offset, 3, "MCC-MNC: %d%d%d-%d%d", o1&0x0f, o1>>4, o2&0x0f, o3&0x0f, o3>>4);
				else				/* 3 digit MNC */
					proto_tree_add_text(tree, tvb, offset, 3, "MCC-MNC: %d%d%d-%d%d%d", o1&0x0f, o1>>4, o2&0x0f, o2>>4, o3&0x0f, o3>>4);
				}
			offset += 3;
			/* deliberate drop-through - do not put a break here */
		
		case CELL_ID_DESC_LAC_CI:
		case CELL_ID_DESC_UMTS_LAC_RNC:
		case CELL_ID_DESC_LAC:
			if (tree) {
				lac = tvb_get_ntohs(tvb,offset);
				proto_tree_add_uint(tree, hf_ie_lac, tvb, offset, 2, lac);
			}
			offset += 2;

			if ((cell_id_desc == CELL_ID_DESC_LAI)||(cell_id_desc == CELL_ID_DESC_LAC))
				break;

			/* deliberate drop-through for other cases - do not put a break here */

		case CELL_ID_DESC_CI:
		case CELL_ID_DESC_UMTS_RNC:
				if (tree) {
				if (cell_id_desc==CELL_ID_DESC_UMTS_SAI) {
					sac = tvb_get_ntohs(tvb,offset);
					proto_tree_add_uint(tree, hf_ie_sac, tvb, offset, 2, sac);
				}
				else {
					if ((cell_id_desc==CELL_ID_DESC_UMTS_PLMN_LAC_RNC)
						||(cell_id_desc==CELL_ID_DESC_UMTS_LAC_RNC_CI)
						||(cell_id_desc==CELL_ID_DESC_UMTS_LAC_RNC)
						||(cell_id_desc==CELL_ID_DESC_UMTS_RNC)) {
						rnc_id = tvb_get_ntohs(tvb,offset);
						proto_tree_add_uint(tree, hf_ie_sac, tvb, offset, 2, sac);
					}
					if ((cell_id_desc==CELL_ID_DESC_CGI)
						||(cell_id_desc==CELL_ID_DESC_UMTS_LAC_RNC_CI)
						||(cell_id_desc==CELL_ID_DESC_LAC_CI)
						||(cell_id_desc==CELL_ID_DESC_CI)) {
					ci = tvb_get_ntohs(tvb,offset);
					proto_tree_add_uint(tree, hf_ie_ci, tvb, offset, 2, ci);
				}
				}
			}
				offset += 2;
			break;

		default:		/* something wrong - get out and try to carry on */
			offset = msg_len;
		}
	}

	return;
}

static void dissect_gsm0808_ie_layer_3_header_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 pd, ti;

	msg_len = tvb_length_remaining(tvb,offset);

	pd = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Protocol Descriminator: %d", pd);
	}
	offset++;

	ti = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Transaction Identifier: %d", ti);
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_periodicity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 periodicity;

	msg_len = tvb_length_remaining(tvb,offset);

	periodicity = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Periodicity: %dms", 100*periodicity);
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_chosen_channel(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 channel;

	msg_len = tvb_length_remaining(tvb,offset);

	channel = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Channel: %s, Mode: %s", rd_match_strval(channel&0x0f,chosen_channel_vals), rd_match_strval(channel>>4,chosen_channel_mode_vals));
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_mobile_identity(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 type, odd;
	char *id, id_string[17];	/* 16 digits is max size + NULL*/

	id = id_string;

	msg_len = tvb_length_remaining(tvb,offset);

	type = tvb_get_guint8(tvb, offset) & 0x07;
	odd = (tvb_get_guint8(tvb, offset) >> 3) & 0x01;
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Identity Type: %s", rd_match_strval(type,identity_type_vals));
	}

	if (type == MOBILE_IDENTITY_TYPE_TMSI) {
		offset++;	/* TMSIs have pad in rest of octet 3, followed by 32 bit value */
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 4, "Identity: 0x%8.8x", tvb_get_ntohl(tvb,offset));
		}
	}
	else {
		for (; offset<msg_len; offset++) {
			if (offset>0) {	/* first octet only has upper nibble used for digits */
				*id++ = hex_digit[tvb_get_guint8(tvb, offset) & 0x0f];
			}
			if (odd || (offset<msg_len-1)) { /* last octet has last nibble spare if even number of digits */
				*id++ = hex_digit[(tvb_get_guint8(tvb, offset) >> 4) & 0x0f];
			}
		}
		*id = 0;	/* NULL terminator */

		if (tree) {
			proto_tree_add_text(tree, tvb, 0, msg_len, "Identity: %s", id_string);
		}
	}

	return;
}

static void dissect_gsm0808_ie_current_channel_type_1(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 channel;

	msg_len = tvb_length_remaining(tvb,offset);

	channel = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Channel: %s, Mode: %s", rd_match_strval(channel&0x0f,chosen_channel_vals), rd_match_strval(channel>>4,current_channel_mode_vals));
	}
	offset++;

	return;
}

static void dissect_gsm0808_eie_current_channel_type_2(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 mode, channel;

	msg_len = tvb_length_remaining(tvb,offset);

	mode = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Mode: %s", rd_match_strval(mode,current_channel_mode_vals));
	}
	offset++;

	channel = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Channel: %s", rd_match_strval(channel,chosen_channel_vals));
	}
	offset++;

	return;
}

static void dissect_gsm0808_eie_extra_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 extra_info;

	msg_len = tvb_length_remaining(tvb,offset);

	extra_info = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Old BSS recommends %s", rd_match_strval(extra_info&0x01,extra_information_prec_vals));
		proto_tree_add_text(tree, tvb, offset, 1, "%s", rd_match_strval((extra_info>>1)&0x01,extra_information_lcs_vals));
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_priority(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 priority;

	msg_len = tvb_length_remaining(tvb,offset);

	priority = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Priority: %d, Pre-emption %sallowed, Queueing %sallowed, %svulnerable to preemption", (priority>>2)&0x0f, (priority>>6)&0x01?"":"not ", (priority>>1)&0x01?"":"not ",(priority>>0)&0x01?"":"not ");
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_diagnostics(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 byte_ptr, bit_ptr;
	char str[32];		/* enough for a few octets... */

	msg_len = tvb_length_remaining(tvb,offset);

	byte_ptr = tvb_get_guint8(tvb, offset);
	if (tree) {
		switch (byte_ptr) {
		case 253:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found at first octet of BSSAP header (descriminator)");
			break;

		case 254:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found at second octet of BSSAP/DTAP header (DLCI)");
			break;

		case 255:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found at last octet of BSSAP header (length)");
			break;

		default:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found at octet %d of message", byte_ptr);
			break;
		}
	}
	offset++;

	bit_ptr = tvb_get_guint8(tvb, offset);
	if (tree) {
		switch (byte_ptr) {
		case 0:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found in no particular part of that octet");
			break;

		default:
			proto_tree_add_text(tree, tvb, offset, 1, "Error found at bit %d of that octet", bit_ptr);
			break;
		}
	}
	offset++;

	if (tree) {
		proto_tree_add_text(tree, tvb, offset, msg_len-offset, "Message in error: %s", data_str( tvb, offset, msg_len-offset, str, sizeof(str) ));
	}
	offset+=msg_len-offset;

	return;
}

static void dissect_gsm0808_ie_encryption_information(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;
	guint8 algs;
	char str[32];		/* enough for standard 64 bit keys */

	msg_len = tvb_length_remaining(tvb,offset);

	algs = tvb_get_guint8(tvb, offset);
	for (i=0; i<8; i++) {
		if (tree) {
			if (algs&(1<<i)) {
				if (i==0)
					proto_tree_add_text(tree, tvb, offset, 1, "No encryption is supported");
				else
					proto_tree_add_text(tree, tvb, offset, 1, "A5/%d is supported", i);
			}
		}
	}
	offset++;

	if (tree) {
		proto_tree_add_text(tree, tvb, offset, msg_len-1, "Key %s", data_str( tvb, offset, msg_len-1, str, sizeof(str) ));
	}
	offset+=msg_len-1;

	return;
}

static void dissect_gsm0808_ie_chosen_encryption_algorithm(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len;
	guint8 alg;

	msg_len = tvb_length_remaining(tvb,offset);

	alg = tvb_get_guint8(tvb, offset);
	if (tree) {
		if (alg==0)
			proto_tree_add_text(tree, tvb, offset, 1, "No encryption used");
		else
			proto_tree_add_text(tree, tvb, offset, 1, "A5/%d used", alg-1);
	}
	offset++;

	return;
}

static void dissect_gsm0808_ie_channel_type(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i;
	guint8 speech_data, chan_rate_type, data_rate, rates, octet5b;

	msg_len = tvb_length_remaining(tvb,offset);

	speech_data = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "%s", rd_match_strval(speech_data,speech_data_vals));
	}
	offset++;

	chan_rate_type = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Channel Rate and Type: %s", rd_match_strval(chan_rate_type,channel_rate_and_type_vals));
	}
	offset++;


	if (speech_data == SD_SPEECH) {
		for (; offset<msg_len; offset++) {
			if (tree) {
				proto_tree_add_text(tree, tvb, offset, 1, "Speech algorithm: %s", rd_match_strval(tvb_get_guint8(tvb, offset)&0x7f,speech_version_vals));
			}

		}
	}
	else if (speech_data == SD_DATA) {
		data_rate = tvb_get_guint8(tvb, offset);	/* octet 5 */
		if (tree) {
			if (data_rate&0x40)
				proto_tree_add_text(tree, tvb, offset, 1, "Non-Transparent rate: %s", rd_match_strval(data_rate&0x3f,nt_radio_data_rate_vals));
			else
				proto_tree_add_text(tree, tvb, offset, 1, "Transparent rate: %s", rd_match_strval(data_rate&0x3f,t_radio_data_rate_vals));
		}
		offset++;
		if (data_rate&0x80 && offset<msg_len) {	/* octet 5a */
			rates = tvb_get_guint8(tvb, offset);
			for (i=0; i<7; i++) {
				if (tree) {
					proto_tree_add_text(tree, tvb, offset, 1, "%s", rd_match_strval(rates&(1<<i),allowed_air_interface_rates_vals));
				}
			}
			offset++;
			if (rates&0x80 && offset<msg_len) {	/* octet 5b */
				octet5b = tvb_get_guint8(tvb, offset);
				if (tree) {
					proto_tree_add_text(tree, tvb, offset, 1, "%s", rd_match_strval((octet5b>>5)&0x03,asymmetry_indication_vals));
				}
				offset++;
			}
		}
	}

	return;
}


/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm0808(void)
{		 

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_gsm0808_msg_type,
			{ "Msg Type", "gsm0808.msg_type",
			FT_UINT8, BASE_HEX, VALS(gsm0808_msg_type_vals), 0x0,
			"08.08 Message Type" }
		},
		{ &hf_ie_ei,
			{ "IE EI", "gsm0808.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ie_ei_vals), 0x0,
			"IE Element Identifier" }
		},
		{ &hf_ie_len8,
			{ "IE Len", "gsm0808.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"IE Length" }
		},
		{ &hf_ie_len16,
			{ "IE Len", "gsm0808.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"IE Length" }
		},
		{ &hf_eie_ei,
			{ "EIE", "gsm0808.eie.ei",
			FT_UINT8, BASE_HEX, VALS(eie_ei_vals), 0x0,
			"Embedded IE Element Identifier", HFILL }
		},
		{ &hf_eie_len,
			{ "EIE Len", "gsm0808.eie.len",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"EIE Length", HFILL }
		},
		{ &hf_ie_cic,
			{ "CIC", "gsm0808.ie.cic",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Circuit Identity Code", HFILL }
		},
		{ &hf_ie_cause,
			{ "Cause", "gsm0808.ie.cause",
			FT_UINT8, BASE_HEX, VALS(cause_vals), 0x0,
			"Cause", HFILL }
		},
		{ &hf_ie_rr_cause,
			{ "RR Cause", "gsm0808.ie.rr_cause",
			FT_UINT8, BASE_HEX, VALS(rr_cause_vals), 0x0,
			"RR Cause", HFILL }
		},
		{ &hf_ie_ci,
			{ "Cell Id", "gsm0808.ie.ci",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Cell Identifier", HFILL }
		},
		{ &hf_ie_lac,
			{ "LAC", "gsm0808.ie.lac",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Location Area Code", HFILL }
		},
		{ &hf_ie_sac,
			{ "SAC", "gsm0808.ie.sac",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Service Area Code", HFILL }
		},
		{ &hf_ie_rnc_id,
			{ "RNC-ID", "gsm0808.ie.rnc_id",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"RNC Idertifier", HFILL }
		},
};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm0808,
		&ett_ie,
		&ett_eie,
	};

	proto_gsm0808 = proto_register_protocol("GSM 08.08 A Interface Layer 3", "GSM 08.08", "gsm0808");

	register_dissector("gsm0808", dissect_gsm0808, proto_gsm0808);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm0808, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm0808(void)
{
    data_handle = find_dissector("data");
    gsm0808_handle = find_dissector("gsm0808");
    gsm0408_handle = find_dissector("gsm0408");
    gsm0408_l3_contents_handle = find_dissector("gsm0408_l3_contents");
}
