/* packet-gsm1221.c
 * Routines for ip.access generic A-bis dissection
 * Copyright 2001, ip.access Ltd <Neil.Piercy@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm1221.c 1.5 2008/06/06 05:11:40BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include <epan/reassemble.h>

#include "packet-gsm.h"
#include "tap.h"

static void dissect_gsm1221_form(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void dissect_gsm1221_manu(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void dissect_gsm1221_ie(guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree);
static void dissect_gsm1221_eie(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *extra);
static void dissect_ie_text(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_measurement_result(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_unt8_vals(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_arfcn_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_bcch_arfcn(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_octet_units(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_short_units(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_recurse(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_interference_level_boundaries(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_t200(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_cgi(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_ip_address(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_connection_failure_criterion(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_physical_config(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_test_report_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_ie_hw_configuration(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);
static void dissect_eie_rtp_features(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info);

/* Initialize the protocol and registered fields */
static int proto_gsm1221 = -1;
static int hf_gsm1221_msg_desc = -1;
static int hf_gsm1221_seq_no = -1;
static int hf_gsm1221_placement = -1;
static int hf_gsm1221_len = -1;
static int hf_gsm1221_msg_type = -1;
static int hf_gsm1221_obj_class = -1;
static int hf_gsm1221_obj_bts = -1;
static int hf_gsm1221_obj_trx = -1;
static int hf_gsm1221_obj_chan = -1;
static int hf_gsm1221_ie_ei = -1;
static int hf_gsm1221_eie_ei = -1;
static int hf_gsm1221_eie_len = -1;
static int hf_gsm1221_ie_len8 = -1;
static int hf_gsm1221_ie_len16 = -1;
static int hf_gsm1221_text = -1;
static int hf_gsm1221_arfcn = -1;
static int hf_gsm1221_ip_address = -1;
static int hf_gsm1221_connection_failure_criterion = -1;
static int hf_gsm1221_em = -1;
static int hf_gsm1221_emi = -1;
static int hf_gsm1221_em_len = -1;
static int hf_gsm1221_em_cause = -1;
static int hf_gsm1221_em_value = -1;
static int hf_gsm1221_reassembled_in = -1;

/* Initialize the subtree pointers */
static gint ett_gsm1221 = -1;
static gint ett_gsm1221_ie = -1;
static gint ett_gsm1221_eie = -1;

static dissector_handle_t gsm1221_handle;
static dissector_handle_t data_handle;

static GHashTable *fragment_table = NULL;

/* Value Strings */
static const value_string msg_desc_vals[] = {
	{ 0x80, "O&M" },
	{ 0x10, "Manu_O&M" },
	{ 0, NULL }
};
#define MSG_DESC_FORM	0x80
#define MSG_DESC_MANU	0x10

static const value_string placement_vals[] = {
	{ 0x80, "Full Message" },
	{ 0x40, "First Segment" },
	{ 0x20, "Continuation Segment" },
	{ 0x10, "Last Segment" },
	{ 0, NULL }
};
#define PLACEMENT_FULL          0x80
#define PLACEMENT_FIRST         0x40
#define PLACEMENT_CONTINUATION  0x20
#define PLACEMENT_LAST          0x10

static const value_string msg_type_vals[] = {
	{ 0x01,  "Load Data Initiate" },
	{ 0x02,  "Load Data Initiate Ack" },
	{ 0x03,  "Load Data Initiate Nack" },
	{ 0x04,  "Load Data Segment" },
	{ 0x05,  "Load Data Segment Ack" },
	{ 0x06,  "Load Data Abort" },
	{ 0x07,  "Load Data End" },
	{ 0x08,  "Load Data End Ack" },
	{ 0x09,  "Load Data End Nack" },
	{ 0x0a,  "SW Activate Request" },
	{ 0x0b,  "SW Activate Request Ack" },
	{ 0x0c,  "SW Activate Request Nack" },
	{ 0x0d,  "Activate SW" },
	{ 0x0e,  "Activate SW Ack" },
	{ 0x0f,  "Activate SW Nack" },
	{ 0x10,  "SW Activated Report" },
	{ 0x21,  "Establish TEI" },
	{ 0x22,  "Establish TEI Ack" },
	{ 0x23,  "Establish TEI Nack" },
	{ 0x24,  "Connect Terrestrial Signalling" },
	{ 0x25,  "Connect Terrestrial Signalling Ack" },
	{ 0x26,  "Connect Terrestrial Signalling Nack" },
	{ 0x27,  "Disconnect Terrestrial Signalling" },
	{ 0x28,  "Disconnect Terrestrial Signalling Ack" },
	{ 0x29,  "Disconnect Terrestrial Signalling Nack" },
	{ 0x2a,  "Connect Terrestrial Traffic" },
	{ 0x2b,  "Connect Terrestrial Traffic Ack" },
	{ 0x2c,  "Connect Terrestrial Traffic Nack" },
	{ 0x2d,  "Disconnect Terrestrial Traffic" },
	{ 0x2e,  "Disconnect Terrestrial Traffic Ack" },
	{ 0x2f,  "Disconnect Terrestrial Traffic Nack" },
	{ 0x31,  "Connect Multi-Drop Link" },
	{ 0x32,  "Connect Multi-Drop Link Ack" },
	{ 0x33,  "Connect Multi-Drop Link Nack" },
	{ 0x34,  "Disconnect Multi-Drop Link" },
	{ 0x35,  "Disconnect Multi-Drop Link Ack" },
	{ 0x36,  "Disconnect Multi-Drop Link Nack" },
	{ 0x41,  "Set BTS Attributes" },
	{ 0x42,  "Set BTS Attributes Ack" },
	{ 0x43,  "Set BTS Attributes Nack" },
	{ 0x44,  "Set Radio Carrier Attributes" },
	{ 0x45,  "Set Radio Carrier Attributes Ack" },
	{ 0x46,  "Set Radio Carrier Attributes Nack" },
	{ 0x47,  "Set Channel Attributes" },
	{ 0x48,  "Set Channel Attributes Ack" },
	{ 0x49,  "Set Channel Attributes Nack" },
	{ 0x51,  "Perform Test" },
	{ 0x52,  "Perform Test Ack" },
	{ 0x53,  "Perform Test Nack" },
	{ 0x54,  "Test Report" },
	{ 0x55,  "Send Test Report" },
	{ 0x56,  "Send Test Report Ack" },
	{ 0x57,  "Send Test Report Nack" },
	{ 0x58,  "Stop Test" },
	{ 0x59,  "Stop Test Ack" },
	{ 0x5a,  "Stop Test Nack" },
	{ 0x61,  "State Changed Event Report" },
	{ 0x62,  "Failure Event Report" },
	{ 0x63,  "Stop Sending Event Reports" },
	{ 0x64,  "Stop Sending Event Reports Ack" },
	{ 0x65,  "Stop Sending Event Reports Nack" },
	{ 0x66,  "Restart Sending Event Reports" },
	{ 0x67,  "Restart Sending Event Reports Ack" },
	{ 0x68,  "Restart Sending Event Reports Nack" },
	{ 0x69,  "Change Administrative State" },
	{ 0x6a,  "Change Administrative State Ack" },
	{ 0x6b,  "Change Administrative State Nack" },
	{ 0x6c,  "Change Administrative State Request" },
	{ 0x6d,  "Change Administrative State Request Ack" },
	{ 0x6e,  "Change Administrative State Request Nack" },
	{ 0x71,  "Changeover" },
	{ 0x72,  "Changeover Ack" },
	{ 0x73,  "Changeover Nack" },
	{ 0x74,  "OpStart" },
	{ 0x75,  "OpStart Ack" },
	{ 0x76,  "OpStart Nack" },
	{ 0x81,  "Get Attributes" },
	{ 0x82,  "Get Attributes Response" },
	{ 0x83,  "Get Attributes Nack" },
	{ 0x84,  "Set Alarm Threshold" },
	{ 0x85,  "Set Alarm Threshold Ack" },
	{ 0x86,  "Set Alarm Threshold Nack" },
	{ 0x87,  "Reinitialise" },
	{ 0x88,  "Reinitialise Ack" },
	{ 0x89,  "Reinitialise Nack" },
	{ 0x8a,  "Measurement Result Request" },
	{ 0x8b,  "Measurement Result Response" },
	{ 0x8c,  "Stop Measurement" },
	{ 0x8d,  "Start Measurement" },
	{ 0x90,  "Change HW Configuration" },
	{ 0x91,  "Change HW Configuration Ack" },
	{ 0x92,  "Change HW Configuration Nack" },
	{ 0x93,  "Report Outstanding Alarms" },
	{ 0x94,  "Report Outstanding Alarms Ack" },
	{ 0x95,  "Report Outstanding Alarms Nack" },

/* ip.access specials */
	{ 0xDE,  "Start Meaurement Ack" },
	{ 0xDF,  "Stop Meaurement Ack" },
	{ 0xE0,  "Connect IP Signalling" },
	{ 0xE1,  "Connect IP Signalling Ack" },
	{ 0xE2,  "Connect IP Signalling Nack" },
	{ 0xE3,  "Disconnect IP Signalling" },
	{ 0xE4,  "Disconnect IP Signalling Ack" },
	{ 0xE5,  "Disconnect IP Signalling Nack" },
	{ 0xE6,  "Connect IP Traffic" },
	{ 0xE7,  "Connect IP Traffic Ack" },
	{ 0xE8,  "Connect IP Traffic Nack" },
	{ 0xE9,  "Disconnect IP Traffic" },
	{ 0xEA,  "Disconnect IP Traffic Ack" },
	{ 0xEB,  "Disconnect IP Traffic Nack" },
	{ 0xEC,  "Set Default SW Req" },
	{ 0xED,  "Set Default SW Ack" },
	{ 0xEE,  "Set Default SW Nack" },
	{ 0xEF,  "Set NV Attributes" },
	{ 0xF0,  "Set NV Attributes Ack" },
	{ 0xF1,  "Set NV Attributes Nack" },
	{ 0xF2,  "Get NV Attributes" },
	{ 0xF3,  "Get NV Attributes Response" },
	{ 0xF4,  "Get NV Attributes Nack" },
	{ 0xF5,  "Set Attributes" },
	{ 0xF6,  "Set Attributes Ack" },
	{ 0xF7,  "Set Attributes Nack" },
	{ 0xF8,  "Attribute Value Change Event" },
	{ 0xF9,  "SW Deactivate" },
	{ 0xFA,  "SW Deactivate Ack" },
	{ 0xFB,  "SW Deactivate Nack" },
	{ 0xFC,  "Measurement Result Request Nack" },
	{ 0xFD,  "Start Meaurement Nack" },
	{ 0xFE,  "Stop Meaurement Nack" },
	{ 0, NULL }
};

static const value_string obj_class_vals[] = {
	{ 0x00,  "Site Mgr" },
	{ 0x01,  "BTS" },
	{ 0x02,  "Radio Carrier" },
	{ 0x03,  "Channel" },
	{ 0x04,  "Baseband TRX" },
	{ 0x0e,  "E1 Trunk Module" },
	{ 0x0f,  "E1 Port" },
	{ 0x10,  "E1 Channel" },
	{ 0x22,  "Clock Module" },
	{ 0xF0,  "GPRS NSE" },
	{ 0xF1,  "GPRS Cell" },
	{ 0xF2,  "GPRS NSVC" },
	{ 0, NULL }
};
#define OBJ_CLS_SITE	0
#define OBJ_CLS_BTS		1
#define OBJ_CLS_CAR		2
#define OBJ_CLS_CHAN	3
#define OBJ_CLS_TRX		4
#define OBJ_CLS_NSE		0xf0	/* ip.access-ism */
#define OBJ_CLS_CELL	0xf1	/* ip.access-ism */
#define OBJ_CLS_NSVC	0xf2	/* ip.access-ism */

static const value_string admin_state_vals[] = {
	{ 1, "Locked" },
	{ 2, "Unlocked" },
	{ 3, "Shutting Down" },
	{ 0, NULL }
};

static const value_string conn_fail_crit_vals[] = {
	{ 1, "Radio Link Timeout" },
	{ 2, "RXLEV/RXQUAL" },
	{ 0, NULL }
};

static const value_string event_type_vals[] = {
	{ 0, "Communications Failure" },
	{ 1, "QoS Failure" },
	{ 2, "Processing Failure" },
	{ 3, "Equipment Failure" },
	{ 4, "Environmental Failure" },
	{ 0, NULL }
};

static const value_string chan_comb_vals[] = {
	{ 0, "TCH/F" },
	{ 1, "TCH/H" },
	{ 2, "TCH/H + TCH/H" },
	{ 3, "SDCCH/8" },
	{ 4, "Main BCCH" },
	{ 5, "BCCH+SDCCH/4" },
	{ 6, "BCH" },
	{ 7, "BCCH+SDCCH/4+CBCH" },
	{ 8, "SDCCH/8+CBCH" },
	{ 0x0d,"PDCH" },
	{ 0x0c,"cPDCH (inc. PCCCH)" },
	{ 0x0b,"bPDCH (inc. PBCCH)" },
	{ 0x80, "TCH/F or PDCH" },
	{ 0, NULL }
};

static const value_string nack_cause_vals[] = {
	{ 0x01, "Incorrect Message Structure" },
	{ 0x02, "Invalid Message Type" },
	{ 0x05, "Invalid Object class value" },
	{ 0x06, "Object class not supported" },
	{ 0x07, "BTS No. unknown" },
	{ 0x08, "Baseband Transceiver unknown" },
	{ 0x09, "Object Instance unknown" },
	{ 0x0c, "Invalid attribute identifier value" },
	{ 0x0d, "Attribute Identifier not supported" },
	{ 0x0e, "Parameter values outside permitted range" },
	{ 0x0f, "Inconsistency in attribute list" },
	{ 0x10, "Specified implementation not supported" },
	{ 0x11, "Message cannot be performed" },
	{ 0x19, "Resource not implemented" },
	{ 0x1a, "Resource not available" },
	{ 0x1b, "Frequency not available" },
	{ 0x1c, "Test not supported" },
	{ 0x1d, "Capacity restrictions" },
	{ 0x1e, "Physical configuration cannot be performed" },
	{ 0x1f, "Test not initiated" },
	{ 0x20, "Physical configuration cannot be restored" },
	{ 0x21, "No such Test" },
	{ 0x22, "Test cannot be stopped" },
	{ 0x23, "Message inconsistent with physical config" },
	{ 0x25, "Complete file not received" },
	{ 0x26, "File not available at destination" },
	{ 0x27, "File cannot be activated" },
	{ 0x28, "Request not granted" },
	{ 0x29, "Wait" },
	{ 0x2a, "Nothing reportable existing" },
	{ 0x2b, "Measurement not supported" },
	{ 0x2c, "Measurement not started" },
	{ 0x80, "Test precedence Limitation" },
	{ 0, NULL }
};

static const value_string op_state_vals[] = {
	{ 1, "Disabled" },
	{ 2, "Enabled" },
	{ 0, NULL }
};

static const value_string t200_vals[] = {
	{ 0, "SDCCH" },
	{ 1, "FACCH/Full rate" },
	{ 2, "FACCH/Half rate" },
	{ 3, "SACCH with TCH SAPI=0" },
	{ 4, "SACCH with SDCCH" },
	{ 5, "SDCCH SAPI=3" },
	{ 6, "SACCH with TCH SAPI=3" },
	{ 0, NULL }
};

static const value_string perceived_severity_vals[] = {
	{ 0, "Ceased" },
	{ 1, "Critical" },
	{ 2, "Major" },
	{ 3, "Minor" },
	{ 4, "Warning" },
	{ 5, "Indeterminate" },
	{ 0x80, "Logged Ceased" },
	{ 0x81, "Logged Critical" },
	{ 0x82, "Logged Major" },
	{ 0x83, "Logged Minor" },
	{ 0x84, "Logged Warning" },
	{ 0x85, "Logged Indeterminate" },
	{ 0, NULL }
};

static const value_string availability_status_vals[] = {
	{ 0, "In Test" },
	{ 1, "Failed" },
	{ 2, "Power Off" },
	{ 3, "Offline" },
	{ 4, "Off Duty" },
	{ 5, "Dependency" },
	{ 6, "Degraded" },
	{ 7, "Not Installed" },
	{ 0, NULL }
};

static const value_string stream_id_vals[] = {
	{ 0, "RSL" },
	{ 1, "RSL 1" },
	{ 128, "NULL" },
	{ 254, "Channel Control" },
	{ 255, "OML" },
	{ 0, NULL }
};

static const value_string emi_vals[] = {
	{ 0x01, "attPageReq"},
	{ 0x02, "pageDisc"},
	{ 0x03, "cumPchQueue"},
	{ 0x04, "chanReq"},
	{ 0x05, "cumSdcchQueue"},
	{ 0x06, "timeDlPcMax"},
	{ 0x07, "timeUlPcMax"},
	{ 0x08, "cumDlRxlev"},
	{ 0x09, "cumUlRxlev"},
	{ 0x0a, "cumDlRxqual"},
	{ 0x0b, "cumUlRxqual"},
	{ 0x0c, "cumTa"},
	{ 0x0d, "ulPkts"},
	{ 0x0e, "dlPkts"},
	{ 0x0f, "ulPktsLost"},
	{ 0x10, "dlPktsLost"},
	{ 0x11, "ulPktsNotSent"},
	{ 0x12, "dlPktsDtx"},
	{ 0x13, "cumDlJit"},
	{ 0x14, "cumUlJit"},
	{ 0x15, "cumTxDelay"},
	{ 0x16, "maxDlJit"},
	{ 0x17, "maxUlJit"},
	{ 0x18, "maxTxDelay"},
	{ 0x19, "pktPageCcch"},
	{ 0x1a, "pktChanReq"},
	{ 0x1b, "pktDlAssCcch"},
	{ 0x1c, "sucPktDlAssCcch"},
	{ 0x1d, "pktUlAssCcch"},
	{ 0x1e, "sucPktUlAssCcch"},
	{ 0x1f, "cumDurPktDlAssCcch"},
	{ 0x20, "attTxPktPageReqPpch"},
	{ 0x21, "pktPageDiscPpch"},
	{ 0x22, "cumPpchQueue"},
	{ 0x23, "pktChanReqCause"},
	{ 0x24, "psPageReqSgsn"},
	{ 0x25, "csPageReqSgsn"},
	{ 0x26, "pageProcRrPacch"},
	{ 0x27, "ulTbfReqDlTbf"},
	{ 0x28, "dlTbfReqUlTbf"},
	{ 0x2b, "attUlTbf"},
	{ 0x2c, "estUlTbf"},
	{ 0x2d, "sucUlTbf"},
	{ 0x2e, "attDlTbf"},
	{ 0x2f, "estDlTbf"},
	{ 0x30, "sucDlTbf"},
	{ 0x31, "failDlTbfCause"},
	{ 0x32, "failUlTbfCause"},
	{ 0x33, "cumDlTbf"},
	{ 0x34, "cumUlTbf"},
	{ 0x35, "maxDlTbf"},
	{ 0x36, "maxUlTbf"},
	{ 0x37, "maxDlTbfTime"},
	{ 0x38, "maxUlTbfTime"},
	{ 0x39, "dlTbfPdch"},
	{ 0x3a, "ulTbfPdch"},
	{ 0x3b, "dlLlcOct"},
	{ 0x3c, "ulLlcOct"},
	{ 0x3d, "dlLlcPdu"},
	{ 0x3e, "ulLlcPdu"},
	{ 0x3f, "flushReq"},
	{ 0x40, "dlLlcPduDisc"},
	{ 0x41, "ulLlcPduDisc"},
	{ 0x42, "cumLlcPduQueue"},
	{ 0x43, "dlEmptyBlks"},
	{ 0x44, "dlPacchBlocks"},
	{ 0x45, "dlPccchBlks"},
	{ 0x46, "dlPbcchBlks"},
	{ 0x47, "dlPdtchBlks"},
	{ 0x48, "ulPdtchBlksUnused"},
	{ 0x49, "ulPdtchBlksEmpty"},
	{ 0x4a, "ulPdtchBlksCrpt"},
	{ 0x4b, "ulPacchBlks"},
	{ 0x4c, "ulPdtchBlks"},
	{ 0x4d, "dlPcMaxTime"},
	{ 0x4e, "ulPcMaxTime"},
	{ 0x4f, "dlCsChange"},
	{ 0x50, "ulCsChange"},
	{ 0x51, "dlBlksCs"},
	{ 0x52, "ulBlksCs"},
	{ 0x53, "dlBlksErrCs"},
	{ 0x54, "ulBlksErrCs"},
	{ 0x56, "connTime"},
	{ 0x57, "connTime"},
	{ 0x58, "dlTchFrBlkCodingScheme"},
	{ 0x59, "ulTchFrBlkCodingScheme"},
	{ 0x5a, "dlTchHrBlkCodingScheme"},
	{ 0x5b, "ulTchHrBlkCodingScheme"},
	{ 0x5c, "dlAmrFrCsChange"},
	{ 0x5d, "dlAmrHrCsChange"},
	{ 0x5e, "ulAmrFrCsChange"},
	{ 0x5f, "ulAmrHrCsChange"},
	{ 0x60, "ulTchFrBlkErrCodingScheme"},
	{ 0x61, "ulTchHrBlkErrCodingScheme"},
	{ 0x62, "connTime"},
	{ 0x63, "meanDlJit"},
	{ 0x64, "meanUlJit"},
	{ 0x65, "meanTxDelay"},
	{ 0x66, "maxDlJit"},
	{ 0x67, "maxUlJit"},
	{ 0x68, "maxTxDelay"},
	{ 0x69, "ulMuxPktsTime"},
	{ 0x6a, "ulMuxPktsSize"},
	{ 0x6b, "ulMuxPktsEmpty"},
	{ 0x6c, "dlMuxPkts"},
	{ 0x6d, "ulMuxPktsLost"},
	{ 0x6e, "dlMuxPktsLost"},
	{ 0x6f, "ulSrtpPkts"},
	{ 0x70, "dlSrtpPkts"},
	{ 0x71, "dlSrtpPktsDisc"},
	{ 0x72, "ulSrtpOctets"},
	{ 0x73, "dlSrtpOctets"},
	{ 0x74, "ulGprsOctSq"},
	{ 0x75, "dlGprsOctSq"},
	{ 0x76, "ulEgprsOctSq"},
	{ 0x77, "dlEgprsOctSq"},
	{ 0x78, "ulGprsOct"},
	{ 0x79, "dlGprsOct"},
	{ 0x7a, "meanDlRxLevSdcch"},
	{ 0x7b, "meanUlRxLevSdcch"},
	{ 0x7c, "meanDlRxLevTch"},
	{ 0x7d, "meanUlRxLevTch"},
	{ 0x82, "ulAbisVolume"},
	{ 0x83, "dlAbisVolume"},
	{ 0x84, "ulPeakAbisRatePerSec"},
	{ 0x85, "dlPeakAbisRatePerSec"},
	{ 0x86, "ulPeakAbisRatePer10Sec"},
	{ 0x87, "dlPeakAbisRatePer10Sec"},
	{ 0x8a, "abTermRrAmrTchFrRfLastUlCodecMode"},
	{ 0x8b, "abTermRrAmrTchFrRfLastDlCodecMode"},
	{ 0x8c, "abTermRrAmrTchHrRfLastUlCodecMode"},
	{ 0x8d, "abTermRrAmrTchHrRfLastDlCodecMode"},
	{ 0x8e, "ulEgprsOct"},
	{ 0x8f, "dlEgprsOct"},
	{ 0x90, "preemptUlTbf"},
	{ 0x91, "preemptDlTbf"},
	{ 0x92, "ulRxQualDetail"},
	{ 0x93, "dlRxQualDetail"},
	{ 0x94, "connTime"},
	{ 0x95, "meanTxDelay"},
	{ 0x96, "maxTxDelay"},
	{ 0x97, "ulGprsMuxPktsTime"},
	{ 0x98, "ulGprsMuxPktsSize"},
	{ 0x99, "ulGprsMuxPktsEmpty"},
	{ 0x9a, "dlGprsMuxPkts"},
	{ 0x9b, "ulGprsMuxPktsLost"},
	{ 0x9c, "dlGprsMuxPktsLost"},
	{ 0, NULL }
};

static const value_string measurement_id_vals[] = {
	{ 0x40, "CCCH Utilisation MF"},
	{ 0x41, "SDCCH Usage MF"},
	{ 0x42, "Measurement Processing MF"},
	{ 0x43, "RTP Usage MF"},
	{ 0x44, "RTP Performance MF"},
	{ 0x45, "GPRS CCCH MF"},
	{ 0x46, "PCCCH Utilisation MF"},
	{ 0x47, "GPRS CCCH Details MF"},
	{ 0x48, "Establishment MF"},
	{ 0x49, "Uplink TBF MF"},
	{ 0x4a, "Downlink TBF MF"},
	{ 0x4b, "TBF Details MF"},
	{ 0x4c, "TBF Usage MF"},
	{ 0x4d, "LLC Data MF"},
	{ 0x4e, "PDCH Usage MF"},
	{ 0x4f, "Power Control MF"},
	{ 0x50, "Link Adaptation MF"},
	{ 0x51, "TCH Usage Details MF"},
	{ 0x52, "AMR MF"},
	{ 0x53, "RTP Multiplex Performance MF"},
	{ 0x54, "RTP Multiplex Usage MF"},
	{ 0x55, "SRTP Multiplex Usage MF"},
	{ 0x56, "Abis Traffic MF"},
	{ 0x58, "GPRS Multiplex Performance MF"},
	{ 0x59, "GPRS Multiplex Usage MF"},
	{ 0, NULL }
};

static const value_string ie_ei_vals[] = {
	{ 0x01, "Abis Channel" },
	{ 0x02, "Additional Info" },
	{ 0x03, "Additional Text" },
	{ 0x04, "Administrative State" },
	{ 0x05, "ARFCN List" },
	{ 0x06, "Autonomously Report" },
	{ 0x07, "Availability Status" },
	{ 0x08, "BCCH ARFCN" },
	{ 0x09, "BSIC" },
	{ 0x0a, "BTS Air Timer" },
	{ 0x0b, "CCCH Load Indication Period" },
	{ 0x0c, "CCCH Load Threshold" },
	{ 0x0d, "Channel Combination" },
	{ 0x0e, "Connection Failure Criterion" },
	{ 0x0f, "Destination" },
	{ 0x11, "Event Type" },
	{ 0x12, "File Id" },
	{ 0x13, "File Version" },
	{ 0x14, "GSM Time" },
	{ 0x15, "HSN" },
	{ 0x16, "HW Configuration" },
	{ 0x17, "HW Description" },
	{ 0x18, "Intave Parameter" },
	{ 0x19, "Interference Level Boundaries" },
	{ 0x1a, "List of Required Attributes" },
	{ 0x1b, "MAIO" },
	{ 0x1c, "Manufacturer Dependent State" },
	{ 0x1d, "Manufacturer Dependent Thresholds" },
	{ 0x1e, "Manufacturer Id" },
	{ 0x1f, "Max Timing Advance" },
	{ 0x20, "Multi-drop BSC Link" },
	{ 0x21, "Multi-drop next BTS Link" },
	{ 0x22, "Nack Causes" },
	{ 0x23, "Ny1" },
	{ 0x24, "Operational State" },
	{ 0x25, "Overload Period" },
	{ 0x26, "Physical Config" },
	{ 0x27, "Power Class" },
	{ 0x28, "Power Output Thresholds" },
	{ 0x29, "Probable Cause" },
	{ 0x2a, "RACH Busy Threshold" },
	{ 0x2b, "RACH Load Averaging Slots" },
	{ 0x2c, "Radio Sub Channel" },
	{ 0x2d, "RF Max Power Reduction" },
	{ 0x2e, "Site Inputs" },
	{ 0x2f, "Site Outputs" },
	{ 0x30, "Source" },
	{ 0x31, "Specific Problems" },
	{ 0x32, "Starting Time" },
	{ 0x33, "T200" },
	{ 0x34, "TEI" },
	{ 0x35, "Test Duration" },
	{ 0x36, "Test No" },
	{ 0x37, "Test Report Info" },
	{ 0x38, "VSWR Thresholds" },
	{ 0x39, "Window Size" },
	{ 0x40, "TSC" },
	{ 0x41, "SW Configuration" },
	{ 0x42, "SW Description" },
	{ 0x43, "Perceived Severity" },
	{ 0x44, "Get Attributes Response Info" },
	{ 0x45, "HW Conf Change Info" },
	{ 0x46, "Outstanding Alarm Sequence" },
	{ 0x47, "File Data" },
	{ 0x48, "Measurement Result" },
	{ 0x49, "Measurement Type" },
/* ip.access defined additions */
	{ 0x80, "Destination IP Address" },
	{ 0x81, "Destination IP Port" },
	{ 0x82, "SSRC" },
	{ 0x83, "RTP Payload Type" },
	{ 0x84, "Baseband Identifier" },
	{ 0x85, "Stream Identifier" },
	{ 0x86, "NV Config Flags" },
	{ 0x87, "Frequency Control" },
	{ 0x88, "Primary Config" },
	{ 0x89, "Secondary Config" },
	{ 0x8a, "IP Interface Config" },
	{ 0x8b, "IP Gateway Config" },
	{ 0x8c, "In Service Time" },
	{ 0x8d, "TRX BTS Assignment" },
	{ 0x8e, "BTS Location" },
	{ 0x8f, "Paging Configuration" },
	{ 0x90, "File Data" },
	{ 0x91, "Unit Id" },
	{ 0x92, "Parent Unit Id" },
	{ 0x93, "Unit Name" },
	{ 0x94, "SNMP Config" },
	{ 0x95, "Primary OML Config List" },
	{ 0x96, "Primary OML Fallback Timeout" },
	{ 0x97, "Current SW Configuration" },
	{ 0x98, "Timing Bus" },
	{ 0x99, "CGI" },
	{ 0x9A, "RAC" },
	{ 0x9B, "Object Version" },
	{ 0x9C, "GPRS Paging Configuration" },
	{ 0x9D, "NSEI" },
	{ 0x9E, "BVCI" },
	{ 0x9F, "NSVCI" },
	{ 0xA0, "NS Configuration" },
	{ 0xA1, "BSSGP Configuration" },
	{ 0xA2, "NS Link Configuration" },
	{ 0xA3, "RLC Configuration" },
	{ 0xA4, "Alarm Threshold List" },
	{ 0xA5, "Monitored Value List" },
	{ 0xA6, "Timing Bus Control" },
	{ 0xA7, "Supported Features" },
	{ 0xA8, "Coding Schemes" },
	{ 0xA9, "RLC Configuration 2" },
	{ 0xAA, "Heartbeat Timeout" },
	{ 0xAB, "Up Time" },
	{ 0xAC, "RLC Configuration 3" },
	{ 0xAD, "SSL Configuration" },
	{ 0xAE, "Security Possible" },
	{ 0xAF, "IML SSL State" },
	{ 0xB0, "Revocation Date" },
	{ 0, NULL }
};

static const value_string connection_failure_criterion_vals[] = {
	{ 0x01, "Uplink SACCH error rate (RADIO_LINK_TIMEOUT)" },
	{ 0x02, "RXLEV/RXQUAL" },
	{ 0, NULL }
};

static const value_string ssl_control_vals[] = {
	{ 0x00, "SSL Disabled" },
	{ 0x01, "SSL Optional" },
	{ 0x02, "SSL Mandatory" },
	{ 0, NULL }
};

static const value_string security_possible_vals[] = {
	{ 0x00, "No Certificate" },
	{ 0x01, "Certificate Available but not in use" },
	{ 0x02, "Certificate in use" },
	{ 0, NULL }
};

static const value_string iml_ssl_vals[] = {
	{ 0x00, "Insecure" },
	{ 0x01, "Secure" },
	{ 0, NULL }
};

static const value_string timing_bus_control_vals[] = {
	{ 0x00, "Auto" },
	{ 0x01, "TIB out Enabled" },
	{ 0x02, "TIB out Disabled" },
	{ 0x03, "Reserved" },
	{ 0, NULL }
};

static const value_string test_precedence_vals[] = {
	{ 0x01, "Object must be Locked" },
	{ 0x02, "Object Idle" },
	{ 0x03, "Object Unconfigured" },
	{ 0, NULL }
};

static const value_string test_result_code_vals[] = {
	{ 0x00, "Success" },
	{ 0x01, "Timeout" },
	{ 0x02, "No Suitable Channels" },
	{ 0x03, "Partial Results - more to follow" },
	{ 0x04, "Test Stopped" },
	{ 0, NULL }
};

static const value_string test_no_vals[] = {
	{ 0x01, "Radio Loop test via antenna" },
	{ 0x02, "Radio Loop test via transceiver" },
	{ 0x03, "BTS Functional object self test" },
	{ 0x40, "Channel Usage" },
	{ 0x41, "BCCH Channel Usage" },
	{ 0x42, "Frequency Synchronisation" },
	{ 0x43, "BCCH Information" },
	{ 0x44, "Transmit Beacon" },
	{ 0x45, "SysInfo Monitor" },
	{ 0x46, "BCCH & CCCH Monitor" },
	{ 0, NULL }
};

static const ie_def_t ie_defs[] = {
	{ 0x01, "Abis Channel", 3, NULL, NULL },
	{ 0x02, "Additional Info", -2, dissect_gsm1221_eie, NULL },
	{ 0x03, "Additional Text", -2, dissect_ie_text, NULL },
	{ 0x04, "Administrative State", 1, NULL, (void *)admin_state_vals },
	{ 0x05, "ARFCN List", -2, dissect_ie_arfcn_list, NULL },
	{ 0x06, "Autonomously Report", 1, NULL, NULL },
	{ 0x07, "Availability Status", -2, NULL, (void *)availability_status_vals },
	{ 0x08, "BCCH ARFCN", 2, dissect_ie_bcch_arfcn, NULL },
	{ 0x09, "BSIC", 1, NULL, NULL },
	{ 0x0a, "BTS Air Timer", 1, dissect_ie_octet_units, "*10ms" },
	{ 0x0b, "CCCH Load Indication Period", 1, dissect_ie_octet_units, "s" },
	{ 0x0c, "CCCH Load Threshold", 1, dissect_ie_octet_units, "%" },
	{ 0x0d, "Channel Combination", 1, NULL, (void *)chan_comb_vals },
	{ 0x0e, "Connection Failure Criterion", -2, dissect_ie_connection_failure_criterion, NULL },
	{ 0x0f, "Destination", -2, NULL, NULL },
	{ 0x11, "Event Type", 1, NULL, (void *)event_type_vals },
	{ 0x12, "File Id", -2, dissect_ie_text, NULL },
	{ 0x13, "File Version", -2, dissect_ie_text, NULL },
	{ 0x14, "GSM Time", 2, dissect_ie_short_units, " FN (modulo 42432)" },
	{ 0x15, "HSN", 1, NULL, NULL },
	{ 0x16, "HW Configuration", -2, dissect_ie_hw_configuration, NULL }, /* fudge: actually carries a set of HW Descriptions IEs */
	{ 0x17, "HW Description", -2, NULL, NULL },	/* fudge: actually a set of 2 octet length + data[n] fields */
	{ 0x18, "Intave Parameter", 1, dissect_ie_octet_units, " SACCH multiframes" },
	{ 0x19, "Interference Level Boundaries", 6, dissect_ie_interference_level_boundaries, NULL },
	{ 0x1a, "List of Required Attributes", -2, NULL, (void *)ie_ei_vals },
	{ 0x1b, "MAIO", 1, NULL, NULL },
	{ 0x1c, "Manufacturer Dependent State", 1, NULL, NULL },
	{ 0x1d, "Manufacturer Dependent Thresholds", -2, NULL, NULL },
	{ 0x1e, "Manufacturer Id", -2, dissect_ie_text, NULL },
	{ 0x1f, "Max Timing Advance", 1, dissect_ie_octet_units, " bits" },
	{ 0x20, "Multi-drop BSC Link", 2, NULL, NULL },
	{ 0x21, "Multi-drop next BTS Link", 2, NULL, NULL },
	{ 0x22, "Nack Causes", 1, NULL, (void *)nack_cause_vals },
	{ 0x23, "Ny1", 1, dissect_ie_octet_units, " repetitions of Physical Info" },
	{ 0x24, "Operational State", 1, NULL, (void *)op_state_vals },
	{ 0x25, "Overload Period", -2, NULL, NULL },
	{ 0x26, "Physical Config", -2, dissect_ie_physical_config, NULL },
	{ 0x27, "Power Class", 1, dissect_ie_octet_units, "mobile power class (05.05)" },
	{ 0x28, "Power Output Thresholds", 3, NULL, NULL },
	{ 0x29, "Probable Cause", 3, NULL, NULL },
	{ 0x2a, "RACH Busy Threshold", 1, dissect_ie_octet_units, " neg dBm" },
	{ 0x2b, "RACH Load Averaging Slots", 2, dissect_ie_short_units, " slots" },
	{ 0x2c, "Radio Sub Channel", 1, NULL, NULL },
	{ 0x2d, "RF Max Power Reduction", 1, dissect_ie_octet_units, "*2dB" },
	{ 0x2e, "Site Inputs", -2, NULL, NULL },
	{ 0x2f, "Site Outputs", -2, NULL, NULL },
	{ 0x30, "Source", -2, NULL, NULL },
	{ 0x31, "Specific Problems", 1, NULL, NULL },
	{ 0x32, "Starting Time", 2, dissect_ie_short_units, " FN (modulo 42432)" },
	{ 0x33, "T200", 7, dissect_ie_t200, NULL },
	{ 0x34, "TEI", 1, NULL, NULL },
	{ 0x35, "Test Duration", 2, dissect_ie_short_units, "s" },
	{ 0x36, "Test No", 1, NULL, (void *)test_no_vals },
	{ 0x37, "Test Report Info", -2, dissect_ie_test_report_info, NULL },
	{ 0x38, "VSWR Thresholds", 2, NULL, NULL },
	{ 0x39, "Window Size", 1, NULL, NULL },
	{ 0x40, "TSC", 1, NULL, NULL },
	{ 0x41, "SW Configuration", -2, dissect_ie_recurse, NULL },
	{ 0x42, "SW Description", 0, NULL, NULL }, /* fudge: actually a File Id and a File Version follow directly */
	{ 0x43, "Perceived Severity", 1, NULL, (void *)perceived_severity_vals },
	{ 0x44, "Get Attributes Response Info", -2, NULL, NULL },
	{ 0x45, "HW Conf Change Info", -2, NULL, NULL },
	{ 0x46, "Outstanding Alarm Sequence", 1, NULL, NULL },
	{ 0x47, "File Data", -2, NULL, NULL },
	{ 0x48, "Measurement Result", -2, dissect_ie_measurement_result, NULL },
	{ 0x49, "Measurement Type", 1, NULL, (void *)measurement_id_vals },
	{ 0x80, "Destination IP Address", 4, dissect_ie_ip_address, NULL },
	{ 0x81, "Destination IP Port", 2, dissect_ie_short_units, "" },
	{ 0x82, "SSRC", 4, NULL, NULL },
	{ 0x83, "RTP Payload Type", 2, NULL, NULL },
	{ 0x84, "Baseband Identifier", 1, NULL, NULL },
	{ 0x85, "Stream Identifier", 1, NULL, (void *)stream_id_vals },
	{ 0x86, "NV Config Flags", -2, NULL, NULL },
	{ 0x87, "Frequency Control", 2, dissect_ie_short_units, "ppb" },
	{ 0x88, "Primary Config", 6, NULL, NULL },
	{ 0x89, "Secondary Config", 6, NULL, NULL },
	{ 0x8a, "IP Interface Config", 8, NULL, NULL },
	{ 0x8b, "IP Gateway Config", 12, NULL, NULL },
	{ 0x8c, "In Service Time", 4, NULL, NULL },
	{ 0x8d, "TRX BTS Assignment", 2, NULL, NULL },
	{ 0x8e, "BTS Location", -2, dissect_ie_text, NULL },
	{ 0x8f, "Paging Configuration", 2, NULL, NULL },
	{ 0x90, "File Data", -2, dissect_ie_text, NULL },
	{ 0x91, "Unit Id", -2, dissect_ie_text, NULL },
	{ 0x92, "Parent Unit Id", -2, dissect_ie_text, NULL },
	{ 0x93, "Unit Name", -2, dissect_ie_text, NULL },
	{ 0x94, "SNMP Config", -2, dissect_gsm1221_eie, NULL },
	{ 0x95, "Primary OML Config List", -2, dissect_ie_recurse, NULL },
	{ 0x96, "Primary OML Fallback Timeout", -2, dissect_ie_short_units, "s" },
	{ 0x97, "Current SW Configuration", -2, dissect_ie_recurse, NULL },
	{ 0x98, "Timing Bus", -2, NULL, NULL },
	{ 0x99, "CGI", -2, dissect_ie_cgi, NULL },
	{ 0x9A, "RAC", -2, dissect_ie_octet_units, "" },
	{ 0x9B, "Object Version", -2, NULL, NULL },
	{ 0x9C, "GPRS Paging Configuration", -2, NULL, NULL },
	{ 0x9D, "NSEI", -2, dissect_ie_short_units, "" },
	{ 0x9E, "BVCI", -2, dissect_ie_short_units, "" },
	{ 0x9F, "NSVCI", -2, dissect_ie_short_units, "" },
	{ 0xA0, "NS Configuration", -2, NULL, NULL },
	{ 0xA1, "BSSGP Configuration", -2, NULL, NULL },
	{ 0xA2, "NS Link Configuration", -2, NULL, NULL },
	{ 0xA3, "RLC Configuration", -2, NULL, NULL },
	{ 0xA4, "Alarm Threshold List", -2, NULL, NULL },
	{ 0xA5, "Monitored Value List", -2, NULL, NULL },
	{ 0xA6, "Timing Bus Control", -2, NULL, (void *)timing_bus_control_vals },
	{ 0xA7, "Supported Features", -2, dissect_gsm1221_eie, NULL },
	{ 0xA8, "Coding Schemes", -2, NULL, NULL },
	{ 0xA9, "RLC Configuration 2", -2, NULL, NULL },
	{ 0xAA, "Heartbeat Timeout", -2, NULL, NULL },
	{ 0xAB, "Up Time", -2, NULL, NULL },
	{ 0xAC, "RLC Configuration 3", -2, NULL, NULL },
	{ 0xAD, "SSL Configuration", -2, NULL, (void *)ssl_control_vals },
	{ 0xAE, "Security Possible", -2, NULL, (void *)security_possible_vals },
	{ 0xAF, "IML SSL State", -2, NULL, (void *)iml_ssl_vals },
	{ 0xB0, "Revocation Date", -2, dissect_ie_text, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

/*  Embedded IEs */
static const value_string eie_ei_vals[] = {
	{ 0x01, "ARFCN Whitelist" },
	{ 0x02, "ARFCN Blacklist" },
	{ 0x03, "Frequency Error List" },
	{ 0x04, "Channel Usage List" },
	{ 0x05, "BCCH Information Type" },
	{ 0x06, "BCCH Information" },
	{ 0x07, "Configuration" },
	{ 0x08, "Result Details" },
	{ 0x09, "RXLEV Threshold" },
	{ 0x0a, "Frequency Sync Options" },
	{ 0x0b, "MAC Address" },
	{ 0x0c, "HW/SW Compatibility" },
	{ 0x0d, "Manufacturer Serial Number" },
	{ 0x0e, "OEM Id" },
	{ 0x0f, "Date and Time of Manufacture" },
	{ 0x10, "Date and Time of Calibration" },
	{ 0x11, "Beacon Information" },
	{ 0x12, "Frequency Error" },
	{ 0x13, "SNMP Community String" },
	{ 0x14, "SNMP Trap Address" },
	{ 0x15, "SNMP Trap Port" },
	{ 0x16, "SNMP Manager Address" },
	{ 0x17, "SNMP System Contact" },
	{ 0x18, "Factory Id" },
	{ 0x19, "Factory Serial Number" },
	{ 0x1a, "Logged Event Indicator" },
	{ 0x1b, "Localised Additional Text Information" },
	{ 0x1c, "Frequency Bands" },
	{ 0x1d, "Max Timing Advance" },
	{ 0x1e, "Ciphering Algorithms" },
	{ 0x1f, "Channel Types" },
	{ 0x20, "Channel Modes" },
	{ 0x21, "GPRS Coding Schemes" },
	{ 0x22, "RTP Features" },
	{ 0x23, "RSL Features" },
	{ 0x24, "BTS Hardware Class" },
	{ 0x25, "BTS Identifier" },
	{ 0x26, "Broadcast L3 Message" },
	{ 0, NULL }
};

static const ie_def_t eie_defs[] = {
	{ 0x01, "ARFCN Whitelist", -2, NULL, NULL },
	{ 0x02, "ARFCN Blacklist", -2, NULL, NULL },
	{ 0x03, "Frequency Error List", -2, NULL, NULL },
	{ 0x04, "Channel Usage List", -2, NULL, NULL },
	{ 0x05, "BCCH Information Type", -2, NULL, NULL },
	{ 0x06, "BCCH Information", -2, NULL, NULL },
	{ 0x07, "Configuration", -2, NULL, NULL },
	{ 0x08, "Result Details", -2, NULL, NULL },
	{ 0x09, "RXLEV Threshold", -2, NULL, NULL },
	{ 0x0a, "Frequency Sync Options", -2, NULL, NULL },
	{ 0x0b, "MAC Address", -2, NULL, NULL },
	{ 0x0c, "HW/SW Compatibility", -2, NULL, NULL },
	{ 0x0d, "Manufacturer Serial Number", -2, NULL, NULL },
	{ 0x0e, "OEM Id", -2, NULL, NULL },
	{ 0x0f, "Date and Time of Manufacture", -2, NULL, NULL },
	{ 0x10, "Date and Time of Calibration", -2, NULL, NULL },
	{ 0x11, "Beacon Information", -2, NULL, NULL },
	{ 0x12, "Frequency Error", -2, NULL, NULL },
	{ 0x13, "SNMP Community String", -2, NULL, NULL },
	{ 0x14, "SNMP Trap Address", -2, NULL, NULL },
	{ 0x15, "SNMP Trap Port", -2, NULL, NULL },
	{ 0x16, "SNMP Manager Address", -2, NULL, NULL },
	{ 0x17, "SNMP System Contact", -2, NULL, NULL },
	{ 0x18, "Factory Id", -2, NULL, NULL },
	{ 0x19, "Factory Serial Number", -2, NULL, NULL },
	{ 0x1a, "Logged Event Indicator", -2, NULL, NULL },
	{ 0x1b, "Localised Additional Text Information", -2, NULL, NULL },
	{ 0x1c, "Frequency Bands", -2, NULL, NULL },
	{ 0x1d, "Max Timing Advance", -2, NULL, NULL },
	{ 0x1e, "Ciphering Algorithms", -2, NULL, NULL },
	{ 0x1f, "Channel Types", -2, NULL, NULL },
	{ 0x20, "Channel Modes", -2, NULL, NULL },
	{ 0x21, "GPRS Coding Schemes", -2, NULL, NULL },
	{ 0x22, "RTP Features", -2, dissect_eie_rtp_features, NULL },
	{ 0x23, "RSL Features", -2, NULL, NULL },
	{ 0x24, "BTS Hardware Class", -2, NULL, NULL },
	{ 0x25, "BTS Identifier", -2, NULL, NULL },
	{ 0x26, "Broadcast L3 Message", -2, NULL, NULL },
	{ 0, NULL, 0, NULL, NULL }
};

/* Code to dissect the packets */
static void
dissect_gsm1221(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_desc, placement, seq_no, len;
	tvbuff_t *next_tvb;
    fragment_data *frag_head;
    guint32 frag_id;

	/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	/* load the display labels 	*/
	pinfo->current_proto = "gsm1221";

	gsmstat_set_interface(GSMSTAT_INTERFACE_A_BIS,pinfo);

	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM");

	msg_len = tvb_length_remaining(tvb,offset);
	if (main_tree) {
		/* create display subtree for the protocol */
		ti = proto_tree_add_item(main_tree, proto_gsm1221, tvb, offset, msg_len, FALSE);

		/* add an item to the subtree, see section 1.6 for more information */
		tree = proto_item_add_subtree(ti, ett_gsm1221);
	}

	gsmstat_set_proto(GSMSTAT_PROTO_GSM1221);

	msg_desc = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_msg_desc, tvb, offset, 1, msg_desc);
	offset += 1;

	placement = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_placement, tvb, offset, 1, placement);
	offset += 1;

	seq_no = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_seq_no, tvb, offset, 1, seq_no);
	offset += 1;

	len = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_len, tvb, offset, 1, len);
	offset += 1;

	if (check_col(pinfo->cinfo, COL_INFO))
		col_append_str(pinfo->cinfo, COL_INFO, "12.21:");

	if (placement == PLACEMENT_FULL) {
	    /* create a new subset TVB for disecting the payload */
	    next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
    }
    else {
        frag_id = pinfo->srcport + pinfo->destport;  /* identifies the stream in case of overlapping fragments on different streams */

        frag_head = fragment_add_seq(tvb, offset, pinfo, frag_id, fragment_table,
                                     seq_no, len, (placement!=PLACEMENT_LAST));

        if ((frag_head)&&(pinfo->fd->num == frag_head->reassembled_in)) {
            next_tvb = tvb_new_real_data (frag_head->data, frag_head->len, frag_head->len);
            tvb_set_child_real_data_tvbuff(tvb, next_tvb);

            /* Add the defragmented data to the data source list. */
            add_new_data_source(pinfo, next_tvb, "Reassembled GSM1221");
        }
		else {
			next_tvb = NULL;
			if (check_col(pinfo->cinfo, COL_INFO))
				col_append_fstr(pinfo->cinfo, COL_INFO, "(Segment %d)", seq_no);
		}
    }

    if (next_tvb) {
	    if (msg_desc==MSG_DESC_FORM)
		    dissect_gsm1221_form(next_tvb, pinfo, tree);
	    else if (msg_desc==MSG_DESC_MANU)
		    dissect_gsm1221_manu(next_tvb, pinfo, tree);

	    tap_queue_packet(gsmstat_tap,pinfo,NULL);
    }
	else if ((frag_head) && (pinfo->fd->num != frag_head->reassembled_in)) {
		if (tree)
			proto_tree_add_uint(tree, hf_gsm1221_reassembled_in, tvb, offset, tvb_length_remaining(tvb,offset), frag_head->reassembled_in);
	}

	return;
}

/* Code to dissect the Standard GSM Formatted O&M packets */
static void
dissect_gsm1221_form(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len;
	guint8 msg_type, obj_class;
	guint8 obj_bts, obj_trx, obj_chan;
	tvbuff_t *next_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	msg_len = tvb_length_remaining(tvb,offset);

	if (main_tree) {
		ti = proto_tree_add_text(main_tree, tvb, offset, msg_len, "" );
		tree = proto_item_add_subtree(ti, ett_gsm1221);
	}

	msg_type = tvb_get_guint8(tvb, offset);
	gsmstat_set_msg_type(msg_type);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_msg_type, tvb, offset, 1, msg_type);
	offset += 1;

	obj_class = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_uint(tree, hf_gsm1221_obj_class, tvb, offset, 1, obj_class);
	offset += 1;

	obj_bts = tvb_get_guint8(tvb, offset);
	if ((obj_class==OBJ_CLS_BTS)||(obj_class==OBJ_CLS_TRX)||(obj_class==OBJ_CLS_CAR)||(obj_class==OBJ_CLS_CHAN)||(obj_class==OBJ_CLS_NSE)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
		if (tree)
			proto_tree_add_uint(tree, hf_gsm1221_obj_bts, tvb, offset, 1, obj_bts);
	offset += 1;

	obj_trx = tvb_get_guint8(tvb, offset);
	if ((obj_class==OBJ_CLS_TRX)||(obj_class==OBJ_CLS_CAR)||(obj_class==OBJ_CLS_CHAN)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
		if (tree)
			proto_tree_add_uint(tree, hf_gsm1221_obj_trx, tvb, offset, 1, obj_trx);
	offset += 1;

	obj_chan = tvb_get_guint8(tvb, offset);
	if (obj_class==OBJ_CLS_CHAN)
		if (tree)
			proto_tree_add_uint(tree, hf_gsm1221_obj_chan, tvb, offset, 1, obj_chan);
	offset += 1;

	if (check_col(pinfo->cinfo, COL_INFO)) {
		if (obj_class!=OBJ_CLS_SITE) {
			if ((obj_class==OBJ_CLS_BTS)||(obj_class==OBJ_CLS_NSE))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s(%s_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts);
			else if ((obj_class==OBJ_CLS_TRX)||(obj_class==OBJ_CLS_CAR)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s(%s_%d_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts,obj_trx);
			else	/* used for GSM Channel */
				col_append_fstr(pinfo->cinfo, COL_INFO, "%s(%s_%d_%d_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts,obj_trx,obj_chan);
		}
		else
			col_append_fstr(pinfo->cinfo, COL_INFO, "%s(%s)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals));
	}
	if (tree) {
		if (obj_class!=OBJ_CLS_SITE) {
			if ((obj_class==OBJ_CLS_BTS)||(obj_class==OBJ_CLS_NSE))
				proto_item_set_text(ti, "12.21: %s(%s_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts);
			else if ((obj_class==OBJ_CLS_TRX)||(obj_class==OBJ_CLS_CAR)||(obj_class==OBJ_CLS_CELL)||(obj_class==OBJ_CLS_NSVC))
				proto_item_set_text(ti, "12.21: %s(%s_%d_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts,obj_trx);
			else	/* used for GSM Channel */
				proto_item_set_text(ti, "12.21: %s(%s_%d_%d_%d)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals),obj_bts,obj_trx,obj_chan);
		}
		else
			proto_item_set_text(ti, "12.21: %s(%s)", rd_match_strval(msg_type,msg_type_vals), rd_match_strval(obj_class,obj_class_vals));
	}
	if (msg_len>offset) {
		next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
		dissect_gsm1221_ie(msg_type, ie_defs, next_tvb, pinfo, tree);
	}

	return;
}

/* Code to dissect the ip.access Manufacturer-specific O&M packets */
static void
dissect_gsm1221_manu(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset=0, msg_len;
	guint8 manu_id_len;
	tvbuff_t *next_tvb;

	msg_len = tvb_length_remaining(tvb,offset);

	manu_id_len = tvb_get_guint8(tvb, offset);
	offset += 1;

	offset += manu_id_len; /* skip Manu Id - ip.a use string */

	next_tvb = tvb_new_subset(tvb, offset, msg_len-offset, -1);
	dissect_gsm1221_form(next_tvb, pinfo, tree); /* all ip.a extensions conform to normal formatting rules */
	return;
}

/* Code to dissect the IEs */
static void
dissect_gsm1221_ie( guint8 msg_type, const ie_def_t *ie_defs, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
	int offset=0, msg_len, i;
	guint8	ie_ei, ie_len8;
	guint16 ie_len16, ie_len;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	ie_ei = tvb_get_guint8(tvb, offset);
	if (main_tree)
		proto_tree_add_uint_hidden(main_tree, hf_gsm1221_ie_ei, tvb, offset, 1, ie_ei);
	offset += 1;

	for (p_ie_def=ie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
		if (p_ie_def->ie_ei==ie_ei)
			break;
	}

	if (p_ie_def->ie_name!=0) {
		if (p_ie_def->ie_len == 0){
			ie_len=0;
			if (main_tree)
				ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
		}
		else if (p_ie_def->ie_len > 0) {
			ie_len = p_ie_def->ie_len;
			if (tvb_length_remaining(tvb,offset)<ie_len) {
				if (main_tree)
					proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len, tvb_length_remaining(tvb,offset));
				ie_len=tvb_length_remaining(tvb,offset);
			}

			if (main_tree) {
				if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset),(value_string *)(p_ie_def->info)));
					for (i=2; i<ie_len+1; i++)
						proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
				else {
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, ie_len, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_gsm1221_ie);
				}
			}
		}
		else if (p_ie_def->ie_len == -1) {
			ie_len8 = tvb_get_guint8(tvb, offset);
			if (tvb_length_remaining(tvb,offset+1)<ie_len8) {
				if (main_tree)
					proto_tree_add_text(main_tree, tvb, 0, tvb_length_remaining(tvb,offset+1)+2, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len8, tvb_length_remaining(tvb,offset+1));
				ie_len8=tvb_length_remaining(tvb,offset+1);
			}

			if (main_tree) {
				proto_tree_add_uint_hidden(main_tree, hf_gsm1221_ie_len8, tvb, offset, 1, ie_len8);
				if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
					if (ie_len8==0) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: <empty>", p_ie_def->ie_name);
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+1),(value_string *)(p_ie_def->info)));
						for (i=3; i<ie_len8+2; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
				}
				else {
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len8+2, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+1, ie_len8, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_gsm1221_ie);
				}
			}
			offset += 1;
			ie_len = ie_len8;
		}

		else if (p_ie_def->ie_len == -2) {
			ie_len16 = tvb_get_ntohs(tvb, offset);

			if (tvb_length_remaining(tvb,offset+2)<ie_len16) {
				if (main_tree)
					proto_tree_add_text(main_tree, tvb, 0, tvb_length_remaining(tvb,offset+2)+3, "IE Length error: %s Length %d, but only %d remaining!", p_ie_def->ie_name, ie_len16, tvb_length_remaining(tvb,offset+2));
				ie_len16=tvb_length_remaining(tvb,offset+2);
			}

			if (main_tree) {
				proto_tree_add_uint_hidden(main_tree, hf_gsm1221_ie_len16, tvb, offset, 2, ie_len16);
				if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
					if (ie_len16==0) {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: <empty>", p_ie_def->ie_name);
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+2),(value_string *)(p_ie_def->info)));
						for (i=4; i<ie_len16+3; i++)
							proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
				}
				else {
					ti = proto_tree_add_text(main_tree, tvb, 0, ie_len16+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, ie_len16, str, sizeof(str) ));
					tree = proto_item_add_subtree(ti, ett_gsm1221_ie);
				}
			}
			offset += 2;
			ie_len = ie_len16;
		}
		else {			/* should never happen */
			ie_len=0; /* safety! */
		}

		if (ie_len>0) {
			new_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
			if (p_ie_def->ie_disector!=NULL)
				(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, main_tree, p_ie_def->info);
			else if (ie_len > MAX_DATA_STR)
				call_dissector(data_handle, new_tvb, pinfo, tree);
			offset+=ie_len;
		}

			/* only recurse to this disector if this ie was decoded OK */
		if (tvb_length_remaining(tvb,offset)>0) {
			/* use a new subset TVB for the recursion */
			new_tvb = tvb_new_subset(tvb, offset, -1, -1);
			dissect_gsm1221_ie(msg_type, ie_defs, new_tvb, pinfo, main_tree);
		}
	}

	return;
}

/* Code to dissect the Embedded IEs */
static void
dissect_gsm1221_eie( int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree, void *extra)
{
	int offset=0, msg_len, i, eie_len;
	guint8	eie_ei;
	const ie_def_t	*p_ie_def;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;
	gchar	str[3*MAX_DATA_STR+4];

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>0) {
		eie_ei = tvb_get_guint8(tvb, offset);
		if (main_tree)
			proto_tree_add_uint_hidden(main_tree, hf_gsm1221_eie_ei, tvb, offset, 1, eie_ei);
		offset += 1;

		for (p_ie_def=eie_defs; p_ie_def->ie_name!=0; p_ie_def++) {
			if (p_ie_def->ie_ei==eie_ei)
				break;
		}

		if (p_ie_def->ie_name!=0) {
			if (p_ie_def->ie_len == 0){
				eie_len=0;
				if (main_tree)
					ti = proto_tree_add_text(main_tree, tvb, 0, 1, p_ie_def->ie_name);
			}
			else if (p_ie_def->ie_len > 0) {
				eie_len = p_ie_def->ie_len;

				if (main_tree) {
					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s:", p_ie_def->ie_name);
						for (i=1; i<eie_len+1; i++)
							if (i==1)
								proto_item_append_text(ti, " %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
							else
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+1, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm1221_eie);
					}
				}
			}
			else if (p_ie_def->ie_len == -2) {
				eie_len = tvb_get_ntohs(tvb, offset);

				if (main_tree) {
					proto_tree_add_uint_hidden(main_tree, hf_gsm1221_eie_len, tvb, offset, 2, eie_len);

					if (p_ie_def->ie_disector==NULL && p_ie_def->info) {
						if (eie_len==0) {
							ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+3, "%s: <empty>", p_ie_def->ie_name);
						}
						else {
							ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+3, "%s: %s", p_ie_def->ie_name, rd_match_strval(tvb_get_guint8(tvb, offset+2),(value_string *)(p_ie_def->info)));
							for (i=4; i<eie_len+3; i++)
								proto_item_append_text(ti, ", %s", rd_match_strval(tvb_get_guint8(tvb, i),(value_string *)(p_ie_def->info)));
						}
					}
					else {
						ti = proto_tree_add_text(main_tree, tvb, 0, eie_len+3, "%s: %s", p_ie_def->ie_name, data_str( tvb, offset+2, eie_len, str, sizeof(str) ));
						tree = proto_item_add_subtree(ti, ett_gsm1221_eie);
					}
				}
				offset += 2;
			}

			if (eie_len>0) {
				new_tvb = tvb_new_subset(tvb, offset, eie_len, -1);
				if (p_ie_def->ie_disector!=NULL)
					(*p_ie_def->ie_disector)(msg_type, new_tvb, pinfo, tree, top_tree, p_ie_def->info);
				else
					call_dissector(data_handle, new_tvb, pinfo, tree);
				offset+=eie_len;
			}

				/* only recurse to this disector if this ie was decoded OK */
			if (tvb_length_remaining(tvb,offset)>0) {

				/* use a new subset TVB for the recursion */
				new_tvb = tvb_new_subset(tvb, offset, -1, -1);

				dissect_gsm1221_eie(msg_type, new_tvb, pinfo, main_tree, top_tree, extra);
			}
		}
	}
	return;
}

/* Code to dissect the Embedded IEs */
static void
dissect_ie_recurse( int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	tvbuff_t *new_tvb;
	proto_item *ti=NULL;
	proto_tree *tree=NULL;

	if (tvb_length_remaining(tvb,offset)>0) {
		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, -1, -1);
		dissect_gsm1221_ie((guint8)msg_type, ie_defs, new_tvb, pinfo, main_tree);
	}

	return;
}

static void dissect_ie_measurement_result(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len, emi, em_len, i, cause, value;
	tvbuff_t *new_tvb;

	msg_len = tvb_length_remaining(tvb,offset);

	if (msg_len>=3) {
		emi = tvb_get_guint8(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, hf_gsm1221_emi, tvb, offset, 1, emi);
		offset += 1;
		em_len = tvb_get_ntohs(tvb, offset);
		if (tree)
			proto_tree_add_uint_hidden(tree, hf_gsm1221_em_len, tvb, offset, 2, em_len);
		offset += 2;

		if (em_len==4) {
			value = tvb_get_ntohl(tvb, offset);
			if (tree) {
				proto_tree_add_uint_hidden(tree, hf_gsm1221_em_value, tvb, offset, 4, value);
				proto_tree_add_none_format(tree, hf_gsm1221_em, tvb, 0, offset+4, "%s=%d", rd_match_strval(emi,emi_vals), value);
			}
			offset += em_len;
		}
		else if (em_len==0){	/* empty array */
			if (tree) {
				proto_tree_add_none_format(tree, hf_gsm1221_em, tvb, offset-3, 3, "%s(empty)", rd_match_strval(emi,emi_vals));
			}
		}
		else if (em_len%5==0){
			for (i=0; i<em_len/5; i++) {
				cause = tvb_get_guint8(tvb, offset);
				if (tree)
					proto_tree_add_uint_hidden(tree, hf_gsm1221_em_cause, tvb, offset, 1, cause);
				offset += 1;
				value = tvb_get_ntohl(tvb, offset);
				if (tree) {
					proto_tree_add_uint_hidden(tree, hf_gsm1221_em_value, tvb, offset, 4, value);
					proto_tree_add_none_format(tree, hf_gsm1221_em, tvb, offset-1, 5, "%s(%d)=%d", rd_match_strval(emi,emi_vals), cause, value);
				}
				offset += 4;
			}
		}

		/* only recurse to this disector if this ie was decoded OK */
		if (tvb_length_remaining(tvb,offset)>0) {
			/* use a new subset TVB for the recursion */
			new_tvb = tvb_new_subset(tvb, offset, -1, -1);
			dissect_ie_measurement_result(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
	}
	return;
}

static void dissect_ie_physical_config(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, precedence;
	tvbuff_t *new_tvb;

	precedence = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_text(tree, tvb, offset, 1, "Test Precendence = %s", rd_match_strval(precedence,test_precedence_vals));
	offset += 1;

	if (tvb_length_remaining(tvb,offset)>0) {
		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, -1, -1);
		dissect_gsm1221_eie(msg_type, new_tvb, pinfo, tree, main_tree, info);
	}
	return;
}

static void dissect_ie_test_report_info(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, result_code;
	tvbuff_t *new_tvb;

	result_code = tvb_get_guint8(tvb, offset);
	if (tree)
		proto_tree_add_text(tree, tvb, offset, 1, "Result Code = %s", rd_match_strval(result_code,test_result_code_vals));
	offset += 1;

	if (tvb_length_remaining(tvb,offset)>0) {
		/* use a new subset TVB for the recursion */
		new_tvb = tvb_new_subset(tvb, offset, -1, -1);
		dissect_gsm1221_eie(msg_type, new_tvb, pinfo, tree, main_tree, info);
	}
	return;
}

static void dissect_ie_hw_configuration(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len, equ_id_len, equ_type_len, equ_ver_len, loc_len, manu_info_len;
	guint8 ie_ei, *equ_id, *equ_type, *equ_ver, *loc;
	tvbuff_t *new_tvb;

	msg_len = tvb_length_remaining(tvb,offset);

	while (offset<msg_len) {	/* the HW Descriptions in the HW Configuration IE are not proper TLVs - they contain no overall length! */
		ie_ei = tvb_get_guint8(tvb, offset);
		if (ie_ei == 0x17) {	/* All HW Descriptions start with a single octet IEI, but NO following overall length - lots of sub-fields with lengths follow */
			if (tree)
				proto_tree_add_text(tree, tvb, offset, 1, "HW Description:");
		}
		else {
			if (tree)
				proto_tree_add_text(tree, tvb, offset, 1, "Protocol error: missing HW Description Attribute Id");
			return;
		}
		offset += 1;

		equ_id_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		equ_id = tvb_get_string(tvb, offset, equ_id_len);
		if (tree)
			proto_tree_add_text(tree, tvb, offset, equ_id_len, "Equipment Id: %s", equ_id);
		g_free(equ_id);
		offset += equ_id_len;

		equ_type_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		equ_type = tvb_get_string(tvb, offset, equ_type_len);
		if (tree)
			proto_tree_add_text(tree, tvb, offset, equ_type_len, "Equipment Type: %s", equ_type);
		g_free(equ_type);
		offset += equ_type_len;

		equ_ver_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		equ_ver = tvb_get_string(tvb, offset, equ_ver_len);
		if (tree)
			proto_tree_add_text(tree, tvb, offset, equ_ver_len, "Equipment Version: %s", equ_ver);
		g_free(equ_ver);
		offset += equ_ver_len;

		loc_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		loc = tvb_get_string(tvb, offset, loc_len);
		if (tree)
			proto_tree_add_text(tree, tvb, offset, loc_len, "Location: %s", loc);
		g_free(loc);
		offset += loc_len;

		manu_info_len = tvb_get_ntohs(tvb, offset);
		offset += 2;
		if (manu_info_len>0) {
			/* use a new subset TVB for the recursion */
			new_tvb = tvb_new_subset(tvb, offset, manu_info_len, manu_info_len);
			dissect_gsm1221_eie(msg_type, new_tvb, pinfo, tree, main_tree, info);
		}
		offset += manu_info_len;
	}

	return;
}

static void dissect_ie_text(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;

	msg_len = tvb_length_remaining(tvb,offset);

	if (tree) {
		proto_tree_add_item(tree, hf_gsm1221_text, tvb, offset, msg_len, FALSE);
	}
	offset += msg_len;

	return;
}

static void dissect_ie_arfcn_list(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len, arfcn;

	msg_len = tvb_length_remaining(tvb,offset);

	for (offset=0; offset<msg_len; offset+=2) {
		arfcn = tvb_get_ntohs(tvb, offset);
		if (tree) {
			proto_tree_add_item(tree, hf_gsm1221_arfcn, tvb, offset, 2, FALSE);
		}
	}
	return;
}

static void dissect_ie_bcch_arfcn(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, arfcn;

	arfcn = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_item(tree, hf_gsm1221_arfcn, tvb, offset, 2, FALSE);
	}
	return;
}

static void dissect_ie_octet_units(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	guint8	octet;

	octet = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "Value=%d%s", octet, (char *)info);
	}
	return;
}

static void dissect_ie_short_units(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	guint16	sh;

	sh = tvb_get_ntohs(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 2, "Value=%d%s", sh, (char *)info);
	}
	return;
}

static void dissect_ie_connection_failure_criterion(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	guint8	octet, sacch;

	octet = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_item(tree, hf_gsm1221_connection_failure_criterion, tvb, offset, 1, FALSE);
	}
	offset++;

	if (octet==0x01) {	/* SACCH multiframes follows */
		sacch = tvb_get_guint8(tvb, offset);
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 1, "SACCH Multiframes = %d", sacch);
		}
		offset++;
	}
	return;
}

static void dissect_ie_interference_level_boundaries(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	guint8	boundary;

	msg_len = tvb_length_remaining(tvb,offset);

	for (offset=0; offset<msg_len; offset++) {
		boundary = tvb_get_guint8(tvb, offset);
		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 1, "Boundary X%d = -%ddBm", offset, boundary);
		}
	}

	return;
}

static void dissect_ie_t200(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0, msg_len;
	guint16	t200, units;

	msg_len = tvb_length_remaining(tvb,offset);

	for (offset=0; offset<msg_len; offset++) {
		t200 = tvb_get_guint8(tvb, offset);
		if ((offset==3) || (offset==4) || (offset==6))
			units=10;
		else
			units=5;

		if (tree) {
			proto_tree_add_text(tree, tvb, offset, 1, "%s = %dms", rd_match_strval(offset,t200_vals), t200*units);
		}
	}

	return;
}

static void dissect_ie_ip_address(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	guint32 ip_address;

	tvb_memcpy(tvb, (unsigned char *)&ip_address, offset, 4);
	if (tree) {
		proto_tree_add_ipv4(tree, hf_gsm1221_ip_address, tvb, 0, 4, ip_address);
	}

	return;
}

static void dissect_ie_cgi(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *top_tree, void *extra)
{
	int offset=0;
	guint8 o1, o2, o3;

	o1 = tvb_get_guint8(tvb, offset);
	o2 = tvb_get_guint8(tvb, offset+1);
	o3 = tvb_get_guint8(tvb, offset+2);
	if (tree) {
		if (o2>>4==0x0f)	/* 2 digit MNC */
			proto_tree_add_text(tree, tvb, offset, 3, "MCC-MNC: %d%d%d-%d%d", o1&0x0f, o1>>4, o2&0x0f, o3&0x0f, o3>>4);
		else				/* 3 digit MNC */
			proto_tree_add_text(tree, tvb, offset, 3, "MCC-MNC: %d%d%d-%d%d%d", o1&0x0f, o1>>4, o2&0x0f, o2>>4, o3&0x0f, o3>>4);
		}
	offset += 3;
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 2, "LAC: %d", tvb_get_ntohs(tvb,offset));
	}
	offset += 2;

	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 2, "CI: %d", tvb_get_ntohs(tvb,offset));
	}
	offset += 2;

	return;
}


static void dissect_eie_rtp_features(int msg_type, tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, proto_tree *main_tree, void *info)
{
	int offset=0;
	guint8 octet;

	octet = tvb_get_guint8(tvb, offset);
	if (tree) {
		proto_tree_add_text(tree, tvb, offset, 1, "C:         %d", (octet>>0)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "IR 8kb/s:  %d", (octet>>1)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "IR 16kb/s: %d", (octet>>2)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "IR 32kb/s: %d", (octet>>3)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "IR 64kb/s: %d", (octet>>4)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "Mux:       %d", (octet>>6)&1);
		proto_tree_add_text(tree, tvb, offset, 1, "SRTP:      %d", (octet>>7)&1);
	}

	return;
}

/* init routine to setup/clear the fragment table */
static void defragment_init(void)
{
    fragment_table_init (&fragment_table);
}


/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm1221(void)
{

/* Setup list of header fields  See Section 1.6.1 for details*/
	static hf_register_info hf[] = {
		{ &hf_gsm1221_msg_desc,
			{ "Msg Desc", "gsm1221.msg_desc",
			FT_UINT8, BASE_HEX, VALS(msg_desc_vals), 0x0,
			"Message Descriminator", HFILL }
		},
		{ &hf_gsm1221_msg_type,
			{ "Msg Type", "gsm1221.msg_type",
			FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,
			"Message Type", HFILL }
		},
		{ &hf_gsm1221_placement,
			{ "Placement", "gsm1221.placement",
			FT_UINT8, BASE_HEX, VALS(placement_vals), 0x0,
			"Placement", HFILL }
		},
		{ &hf_gsm1221_seq_no,
			{ "Sequence No.", "gsm1221.seq_no",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Sequence Number", HFILL }
		},
		{ &hf_gsm1221_len,
			{ "Len", "gsm1221.len",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Length", HFILL }
		},
		{ &hf_gsm1221_obj_class,
			{ "Obj Class", "gsm1221.obj_class",
			FT_UINT8, BASE_HEX, VALS(obj_class_vals), 0x0,
			"Object Class", HFILL }
		},
		{ &hf_gsm1221_obj_bts,
			{ "BTS", "gsm1221.obj_bts",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object BTS", HFILL }
		},
		{ &hf_gsm1221_obj_trx,
			{ "TRX", "gsm1221.obj_trx",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object TRX", HFILL }
		},
		{ &hf_gsm1221_obj_chan,
			{ "Chan", "gsm1221.obj_chan",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"Object Channel", HFILL }
		},
		{ &hf_gsm1221_ie_ei,
			{ "IE", "gsm1221.ie.ei",
			FT_UINT8, BASE_HEX, VALS(ie_ei_vals), 0x0,
			"IE Element Identifier", HFILL }
		},
		{ &hf_gsm1221_ie_len8,
			{ "IE Len", "gsm1221.ie.len8",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_gsm1221_ie_len16,
			{ "IE Len", "gsm1221.ie.len16",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"IE Length", HFILL }
		},
		{ &hf_gsm1221_eie_ei,
			{ "EIE", "gsm1221.eie.ei",
			FT_UINT8, BASE_HEX, VALS(eie_ei_vals), 0x0,
			"Embedded IE Element Identifier", HFILL }
		},
		{ &hf_gsm1221_eie_len,
			{ "EIE Len", "gsm1221.eie.len",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"Embedded IE Length", HFILL }
		},
		{ &hf_gsm1221_text,
			{ "Text", "gsm1221.ie.text",
			FT_STRING, BASE_HEX, NULL, 0x0,
			"Text", HFILL }
		},
		{ &hf_gsm1221_arfcn,
			{ "ARFCN", "gsm1221.ie.arfcn",
			FT_INT16, BASE_DEC, NULL, 0x0,
			"ARFCN", HFILL }
		},
		{ &hf_gsm1221_ip_address,
			{ "IP", "gsm1221.ie.ip_address",
			FT_IPv4, BASE_DEC, NULL, 0x0,
			"IP Address", HFILL }
		},
		{ &hf_gsm1221_connection_failure_criterion,
			{ "Conn. Fail. Crit.", "gsm1221.ie.connection_failure_criterion",
			FT_UINT8, BASE_HEX, VALS(connection_failure_criterion_vals), 0x0,
			"Connection Failure Criterion", HFILL }
		},
		{ &hf_gsm1221_em,
			{ "Embedded Measurement", "gsm1221.ie.em",
			FT_NONE, NULL, NULL, 0x0,
			"EM", HFILL }
		},
		{ &hf_gsm1221_emi,
			{ "Embedded Measurement Identifier", "gsm1221.ie.em.id",
			FT_UINT8, BASE_HEX, VALS(emi_vals), 0x0,
			"EMI", HFILL }
		},
		{ &hf_gsm1221_em_len,
			{ "Embedded Measurement Length", "gsm1221.ie.em.len",
			FT_UINT16, BASE_DEC, NULL, 0x0,
			"EM Length", HFILL }
		},
		{ &hf_gsm1221_em_cause,
			{ "Embedded Measurement Cause", "gsm1221.ie.em.cause",
			FT_UINT8, BASE_DEC, NULL, 0x0,
			"EM Cause", HFILL }
		},
		{ &hf_gsm1221_em_value,
			{ "Embedded Measurement Value", "gsm1221.ie.em.value",
			FT_UINT32, BASE_DEC, NULL, 0x0,
			"EM Value", HFILL }
		},
		{ &hf_gsm1221_reassembled_in,
			{ "Reassembled in Frame", "gsm1221.reassembled_in",
			FT_FRAMENUM, BASE_DEC, NULL, 0x0,
			"Reassembled in", HFILL }
		},
	};

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_gsm1221,
		&ett_gsm1221_ie,
		&ett_gsm1221_eie,
	};

	proto_gsm1221 = proto_register_protocol("GSM 12.21 A-bis Interface O&M Layer 3", "GSM 12.21", "gsm1221");

	register_dissector("gsm1221", dissect_gsm1221, proto_gsm1221);

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_gsm1221, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

    register_init_routine(defragment_init);
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm1221(void)
{
    gsm1221_handle = find_dissector("gsm1221");
    data_handle = find_dissector("data");

	dissector_add("ipa.stream", OML_STREAM, gsm1221_handle);
}

