/* packet-gsm24008gmm.c
 * Routines for GSM 24.08 GMM Gprs Mobility Management)
 * Copyright 2001, ip.access Ltd <Calvin.Monk@ipaccess.com>
 *
 * $Id: epan/dissectors/packet-gsm24008gmm.c 1.2 2006/08/09 18:41:13BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 *
 * Copied from README.developer
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-gsm.h"

/* Initialize the protocol and registered fields */
static int proto_gsm24008gmm = -1;
static int hf_gsm24008gmm_pd = -1;
static int hf_gsm24008gmm_skip = -1;
static int hf_gsm24008gmm_ti_flag = -1;
static int hf_gsm24008gmm_tio = -1;
static int hf_gsm24008gmm_tie = -1;
static int hf_gsm24008gmm_msg_type = -1;
static int hf_gsm24008gmm_ie_tag = -1;
static int hf_gsm24008gmm_ie_len1 = -1;
static int hf_gsm24008gmm_ie_len2 = -1;
/* IEs */
static int hf_gsm24008gmm_ac_reference_number_hi = -1;
static int hf_gsm24008gmm_attach_result_low = -1;
static int hf_gsm24008gmm_attach_type_attach_type_low = -1;
static int hf_gsm24008gmm_attach_type_follow_on_low = -1;
static int hf_gsm24008gmm_ciphering_algorithm_low = -1;
static int hf_gsm24008gmm_ciphering_key_sn_hi = -1;
static int hf_gsm24008gmm_ciphering_key_sn_low = -1;
static int hf_gsm24008gmm_detach_type_detach_type_low = -1;
static int hf_gsm24008gmm_detach_type_power_off_low = -1;
static int hf_gsm24008gmm_imeisv_request_hi = -1;
static int hf_gsm24008gmm_force_to_standby_hi = -1;
static int hf_gsm24008gmm_force_to_standby_low = -1;
static int hf_gsm24008gmm_gprs_timer_units = -1;
static int hf_gsm24008gmm_gprs_timer_val = -1;
static int hf_gsm24008gmm_gmm_cause = -1;
static int hf_gsm24008gmm_identity_type_low = -1;
static int hf_gsm24008gmm_llc_sapi = -1;
static int hf_gsm24008gmm_mobile_id_odd = -1;
static int hf_gsm24008gmm_mobile_id_type = -1;
static int hf_gsm24008gmm_nsapi = -1;
static int hf_gsm24008gmm_network_feature_support_low = -1;
static int hf_gsm24008gmm_radio_priority_hi = -1;
static int hf_gsm24008gmm_radio_priority_low = -1;
static int hf_gsm24008gmm_sm_cause = -1;
static int hf_gsm24008gmm_tear_down_indicator_low = -1;
static int hf_gsm24008gmm_tmsi_status = -1;
static int hf_gsm24008gmm_update_result_hi = -1;
static int hf_gsm24008gmm_update_type_update_type_low = -1;
static int hf_gsm24008gmm_update_type_follow_on_low = -1;


/* Initialize the subtree pointers */
static gint ett_gsm24008gmm = -1;
static gint ett_gsm24008gmm_ie = -1;

static dissector_handle_t gsm24008gmm_handle;
static dissector_handle_t data_handle;

#define SKIP_MASK                                   (0xF0)
#define TI_FLAG_MASK                                (0x80)
#define TIO_MASK                                    (0x70)
#define TIE_MASK                                    (0x7F)
#define PD_MASK                                     (0x0F)

/*****************/
/* Message types */
/*****************/

/* Mobility Management */
#define ATTACH_REQUEST                              (0x01)
#define ATTACH_ACCEPT                               (0x02)
#define ATTACH_COMPLETE                             (0x03)
#define ATTACH_REJECT                               (0x04)
#define DETACH_REQUEST                              (0x05)
#define DETACH_ACCEPT                               (0x06)
#define ROUTING_AREA_UPDATE_REQUEST                 (0x08)
#define ROUTING_AREA_UPDATE_ACCEPT                  (0x09)
#define ROUTING_AREA_UPDATE_COMPLETE                (0x0a)
#define ROUTING_AREA_UPDATE_REJECT                  (0x0b)
#define SERVICE_REQUEST                             (0x0c)
#define SERVICE_ACCEPT                              (0x0d)
#define SERVICE_REJECT                              (0x0e)
#define P_TMSI_REALLOCATION_COMMAND                 (0x10)
#define P_TMSI_REALLOCATION_RESPONSE                (0x11)
#define AUTHENTICATION_AND_CIPHERING_REQUEST        (0x12)
#define AUTHENTICATION_AND_CIPHERING_RESPONSE       (0x13)
#define AUTHENTICATION_AND_CIPHERING_REJECT         (0x14)
#define AUTHENTICATION_AND_CIPHERING_FAILURE        (0x1c)
#define IDENTITY_REQUEST                            (0x15)
#define IDENTITY_RESPONSE                           (0x16)
#define GMM_STATUS                                  (0x20)
#define GMM_INFORMATION                             (0x21)
/* Session Management */
#define ACTIVATE_PDP_CONTEXT_REQUEST                (0x41)
#define ACTIVATE_PDP_CONTEXT_ACCEPT                 (0x42)
#define ACTIVATE_PDP_CONTEXT_REJECT                 (0x43)
#define REQUEST_PDP_CONTEXT_ACTIVATION              (0x44)
#define REQUEST_PDP_CONTEXT_ACTIVATION_REJECT       (0x45)
#define DEACTIVATE_PDP_CONTEXT_REQUEST              (0x46)
#define DEACTIVATE_PDP_CONTEXT_ACCEPT               (0x47)
#define MODIFY_PDP_CONTEXT_REQUEST_NETWORK_TO_MS    (0x48)
#define MODIFY_PDP_CONTEXT_ACCEPT_MS_TO_NETWORK     (0x49)
#define MODIFY_PDP_CONTEXT_REQUEST_MS_TO_NETWORK    (0x4a)
#define MODIFY_PDP_CONTEXT_ACCEPT_NETWORK_TO_MS     (0x4b)
#define MODIFY_PDP_CONTEXT_REJECT                   (0x4c)
#define ACTIVATE_SECONDARY_PDP_CONTEXT_REQUEST      (0x4d)
#define ACTIVATE_SECONDARY_PDP_CONTEXT_ACCEPT       (0x4e)
#define ACTIVATE_SECONDARY_PDP_CONTEXT_REJECT       (0x4f)
#define SM_RESERVED_1                               (0x50)
#define SM_RESERVED_2                               (0x51)
#define SM_RESERVED_3                               (0x52)
#define SM_RESERVED_4                               (0x53)
#define SM_RESERVED_5                               (0x54)
#define SM_STATUS                                   (0x55)

/* Value Strings */
static const value_string msg_type_vals[] = {
  { ATTACH_REQUEST,                           "ATTACH-REQUEST" },
  { ATTACH_ACCEPT,                            "ATTACH-ACCEPT" },
  { ATTACH_COMPLETE,                          "ATTACH-COMPLETE" },
  { ATTACH_REJECT,                            "ATTACH-REJECT" },
  { DETACH_REQUEST,                           "DETACH-REQUEST" },
  { DETACH_ACCEPT,                            "DETACH-ACCEPT" },
  { ROUTING_AREA_UPDATE_REQUEST,              "ROUTING-AREA-UPDATE-REQUEST" },
  { ROUTING_AREA_UPDATE_ACCEPT,               "ROUTING-AREA-UPDATE-ACCEPT" },
  { ROUTING_AREA_UPDATE_COMPLETE,             "ROUTING-AREA-UPDATE-COMPLETE" },
  { ROUTING_AREA_UPDATE_REJECT,               "ROUTING-AREA-UPDATE-REJECT" },
  { SERVICE_REQUEST,                          "SERVICE-REQUEST" },
  { SERVICE_ACCEPT,                           "SERVICE-ACCEPT" },
  { SERVICE_REJECT,                           "SERVICE-REJECT" },
  { P_TMSI_REALLOCATION_COMMAND,              "P-TMSI-REALLOCATION-COMMAND" },
  { P_TMSI_REALLOCATION_RESPONSE,             "P-TMSI-REALLOCATION-RESPONSE" },
  { AUTHENTICATION_AND_CIPHERING_REQUEST,     "AUTHENTICATION-AND-CIPHERING-REQUEST" },
  { AUTHENTICATION_AND_CIPHERING_RESPONSE,    "AUTHENTICATION-AND-CIPHERING-RESPONSE" },
  { AUTHENTICATION_AND_CIPHERING_REJECT,      "AUTHENTICATION-AND-CIPHERING-REJECT" },
  { AUTHENTICATION_AND_CIPHERING_FAILURE,     "AUTHENTICATION-AND-CIPHERING-FAILURE" },
  { IDENTITY_REQUEST,                         "IDENTITY-REQUEST" },
  { IDENTITY_RESPONSE,                        "IDENTITY-RESPONSE" },
  { GMM_STATUS,                                "GMM-STATUS" },
  { GMM_INFORMATION,                           "GMM-INFORMATION" },
  { ACTIVATE_PDP_CONTEXT_REQUEST,             "ACTIVATE-PDP-CONTEXT-REQUEST" },
  { ACTIVATE_PDP_CONTEXT_ACCEPT,              "ACTIVATE-PDP-CONTEXT-ACCEPT" },
  { ACTIVATE_PDP_CONTEXT_REJECT,              "ACTIVATE-PDP-CONTEXT-REJECT" },
  { REQUEST_PDP_CONTEXT_ACTIVATION,           "REQUEST-PDP-CONTEXT-ACTIVATION" },
  { REQUEST_PDP_CONTEXT_ACTIVATION_REJECT,    "REQUEST-PDP-CONTEXT-ACTIVATION-REJECT" },
  { DEACTIVATE_PDP_CONTEXT_REQUEST,           "DEACTIVATE-PDP-CONTEXT-REQUEST" },
  { DEACTIVATE_PDP_CONTEXT_ACCEPT,            "DEACTIVATE-PDP-CONTEXT-ACCEPT" },
  { MODIFY_PDP_CONTEXT_REQUEST_NETWORK_TO_MS, "MODIFY-PDP-CONTEXT-REQUES (Network to MS)" },
  { MODIFY_PDP_CONTEXT_ACCEPT_MS_TO_NETWORK,  "MODIFY-PDP-CONTEXT-ACCEPT (MS to Network)" },
  { MODIFY_PDP_CONTEXT_REQUEST_MS_TO_NETWORK, "MODIFY-PDP-CONTEXT-REQUEST (MS to Network)" },
  { MODIFY_PDP_CONTEXT_ACCEPT_NETWORK_TO_MS,  "MODIFY-PDP-CONTEXT-ACCEPT (Network to MS)" },
  { MODIFY_PDP_CONTEXT_REJECT,                "MODIFY-PDP-CONTEXT-REJECT" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_REQUEST,   "ACTIVATE-SECONDARY-PDP-CONTEXT-REQUEST" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_ACCEPT,    "ACTIVATE-SECONDARY-PDP-CONTEXT-ACCEPT" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_REJECT,    "ACTIVATE-SECONDARY-PDP-CONTEXT-REJECT" },
  { SM_RESERVED_1,                            "RESERVED" },
  { SM_RESERVED_2,                            "RESERVED" },
  { SM_RESERVED_3,                            "RESERVED" },
  { SM_RESERVED_4,                            "RESERVED" },
  { SM_RESERVED_5,                            "RESERVED" },
  { SM_STATUS,                                "SM-STATUS" },
  { 0, NULL }
};

/* Value Strings */
static const value_string msg_type_short_vals[] = {
  { ATTACH_REQUEST,                           "ATTACH-REQ" },
  { ATTACH_ACCEPT,                            "ATTACH-ACCEPT" },
  { ATTACH_COMPLETE,                          "ATTACH-COMP" },
  { ATTACH_REJECT,                            "ATTACH-REJ" },
  { DETACH_REQUEST,                           "DETACH-REQ" },
  { DETACH_ACCEPT,                            "DETACH-ACCEPT" },
  { ROUTING_AREA_UPDATE_REQUEST,              "RAU-REQ" },
  { ROUTING_AREA_UPDATE_ACCEPT,               "RAU-ACCEPT" },
  { ROUTING_AREA_UPDATE_COMPLETE,             "RAU-COMP" },
  { ROUTING_AREA_UPDATE_REJECT,               "RAU-REJ" },
  { SERVICE_REQUEST,                          "SERVICE-REQ" },
  { SERVICE_ACCEPT,                           "SERVICE-ACCEPT" },
  { SERVICE_REJECT,                           "SERVICE-REJ" },
  { P_TMSI_REALLOCATION_COMMAND,              "P-TMSI-REALLOC-CMD" },
  { P_TMSI_REALLOCATION_RESPONSE,             "P-TMSI-REALLOC-RESP" },
  { AUTHENTICATION_AND_CIPHERING_REQUEST,     "AUTH-REQ" },
  { AUTHENTICATION_AND_CIPHERING_RESPONSE,    "AUTH-RESP" },
  { AUTHENTICATION_AND_CIPHERING_REJECT,      "AUTH-REJ" },
  { AUTHENTICATION_AND_CIPHERING_FAILURE,     "AUTH-FAIL" },
  { IDENTITY_REQUEST,                         "ID-REQ" },
  { IDENTITY_RESPONSE,                        "ID-RESP" },
  { GMM_STATUS,                               "GMM-STATUS" },
  { GMM_INFORMATION,                          "GMM-INFO" },
  { ACTIVATE_PDP_CONTEXT_REQUEST,             "ACT-PDP-CNTX-REQ" },
  { ACTIVATE_PDP_CONTEXT_ACCEPT,              "ACT-PDP-CNTX-ACCEPT" },
  { ACTIVATE_PDP_CONTEXT_REJECT,              "ACT-PDP-CNTX-REJ" },
  { REQUEST_PDP_CONTEXT_ACTIVATION,           "REQ-PDP-CNTX-ACT" },
  { REQUEST_PDP_CONTEXT_ACTIVATION_REJECT,    "REQ-PDP-CNTX-ACT-REJ" },
  { DEACTIVATE_PDP_CONTEXT_REQUEST,           "DEACT-PDP-CNTX-REQ" },
  { DEACTIVATE_PDP_CONTEXT_ACCEPT,            "DEACT-PDP-CNTX-ACCEPT" },
  { MODIFY_PDP_CONTEXT_REQUEST_NETWORK_TO_MS, "MOD-PDP-CNTX-REQ" },
  { MODIFY_PDP_CONTEXT_ACCEPT_MS_TO_NETWORK,  "MOD-PDP-CNTX-ACCEPT" },
  { MODIFY_PDP_CONTEXT_REQUEST_MS_TO_NETWORK, "MOD-PDP-CNTX-REQ" },
  { MODIFY_PDP_CONTEXT_ACCEPT_NETWORK_TO_MS,  "MOD-PDP-CNTX-ACCEPT" },
  { MODIFY_PDP_CONTEXT_REJECT,                "MOD-PDP-CNTX-REJ" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_REQUEST,   "ACT-SEC-PDP-CNTX-REQ" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_ACCEPT,    "ACT-SEC-PDP-CNTX-ACCEPT" },
  { ACTIVATE_SECONDARY_PDP_CONTEXT_REJECT,    "ACT-SEC-PDP-CNTX-REJ" },
  { SM_RESERVED_1,                            "RESERVED" },
  { SM_RESERVED_2,                            "RESERVED" },
  { SM_RESERVED_3,                            "RESERVED" },
  { SM_RESERVED_4,                            "RESERVED" },
  { SM_RESERVED_5,                            "RESERVED" },
  { SM_STATUS,                                "SM-STATUS" },
  { 0, NULL }
};

static const value_string pd_vals[] = {
  { 0x00,                                     "Group call control" },
  { 0x01,                                     "Broadcaset call control" },
  { 0x02,                                     "Reserved" },
  { 0x03,                                     "Call control, call related SS message" },
  { 0x04,                                     "GPRS Tranparent Transport Protocol (GTTP)" },
  { 0x05,                                     "Modify management messages" },
  { 0x06,                                     "Radio resources messages" },
  { 0x08,                                     "GPRS mobility management messages" },
  { 0x09,                                     "SMS messages" },
  { 0x0A,                                     "GPRS session management messages" },
  { 0x0B,                                     "non call related SS messages" },
  { 0x0C,                                     "Location services" },
  { 0x0E,                                     "Reserved" },
  { 0x0F,                                     "Reserved" },
  { 0, NULL }
};

/* IE Values */

static const value_string attach_result_vals[] = {
  { 0x01,                                     "Gprs attach" },
  { 0x03,                                     "Combined Gprs/IMSI attach" },
  { 0, NULL }
};

static const value_string attach_type_vals[] = {
  { 0x01,                                     "Gprs attach" },
  { 0x02,                                     "Gprs attach while IMSI attached" },
  { 0x03,                                     "Combined Gprs/IMSI attach" },
  { 0, NULL }
};

static const value_string ciphering_algorithm_vals[] = {
  { 0x00,                                     "Ciphering not used" },
  { 0x01,                                     "GPRS encryption algorithm GEA/1" },
  { 0x02,                                     "GPRS encryption algorithm GEA/2" },
  { 0x03,                                     "GPRS encryption algorithm GEA/3" },
  { 0x04,                                     "GPRS encryption algorithm GEA/4" },
  { 0x05,                                     "GPRS encryption algorithm GEA/5" },
  { 0x06,                                     "GPRS encryption algorithm GEA/6" },
  { 0x07,                                     "GPRS encryption algorithm GEA/7" },
  { 0, NULL }
};

static const value_string detach_type_vals[] = {
  { 0x01,                                     "MS->N/W Gprs detach | N/W->MS re-attach required" },
  { 0x02,                                     "MS->N/W IMSI detached | N/W->MS re-attach not required" },
  { 0x03,                                     "MS->N/W Combined Gprs/IMSI detach | N/W->MS IMSI detach (after VLR failure)" },
  { 0, NULL }
};

static const value_string power_off_vals[] = {
  { 0x00,                                     "Normal detach" },
  { 0x01,                                     "Power switched off" },
  { 0, NULL }
};

static const value_string imeisv_request_vals[] = {
  { 0x00,                                     "IMEISV not requested" },
  { 0x01,                                     "IMEISV requested" },
  { 0, NULL }
};

static const value_string force_to_standby_vals[] = {
  { 0x00,                                     "Force to standby not indicated" },
  { 0x01,                                     "Force to standby indicated" },
  { 0, NULL }
};

static const value_string gprs_timer_units_vals[] = {
  { 0x00,                                     "Value is incremented in multiples of 2 second" },
  { 0x01,                                     "Value is incremented in multiples of 1 minute" },
  { 0x02,                                     "Value is incremented in multiples of decihours" },
  { 0x03,                                     "Value indicates that the timer is deactivated" },
  { 0, NULL }
};

static const value_string gmm_cause_vals[] = {
  { 0x02,                                     "IMSII unknown in H:R" },
  { 0x03,                                     "Illegal MS" },
  { 0x06,                                     "Illegal ME" },
  { 0x07,                                     "GPRS services not allowed" },
  { 0x08,                                     "GPRS services and non-GPRS services not allowed" },
  { 0x09,                                     "MS identity cannot be derived by the network" },
  { 0x0A,                                     "Implicitly detatched" },
  { 0x0B,                                     "PLMN not allowed" },
  { 0x0C,                                     "Location area not allowed" },
  { 0x0D,                                     "Roaming not allowed in this location area" },
  { 0x0E,                                     "GPRS services not allowed in this PLMN" },
  { 0x0F,                                     "No suitable cells in location area" },
  { 0x10,                                     "MSC temporarily not reachable" },
  { 0x11,                                     "Network failure" },
  { 0x14,                                     "MAC failure" },
  { 0x15,                                     "Synch failure" },
  { 0x16,                                     "Congestion" },
  { 0x17,                                     "GSM authentication unavailable" },
  { 0x28,                                     "No PDP activated" },
  { 0x30,                                     "Retry upon entry into a new cell" },
  { 0x31,                                     "Retry upon entry into a new cell" },
  { 0x32,                                     "Retry upon entry into a new cell" },
  { 0x33,                                     "Retry upon entry into a new cell" },
  { 0x34,                                     "Retry upon entry into a new cell" },
  { 0x35,                                     "Retry upon entry into a new cell" },
  { 0x36,                                     "Retry upon entry into a new cell" },
  { 0x37,                                     "Retry upon entry into a new cell" },
  { 0x38,                                     "Retry upon entry into a new cell" },
  { 0x39,                                     "Retry upon entry into a new cell" },
  { 0x3A,                                     "Retry upon entry into a new cell" },
  { 0x3B,                                     "Retry upon entry into a new cell" },
  { 0x3C,                                     "Retry upon entry into a new cell" },
  { 0x3D,                                     "Retry upon entry into a new cell" },
  { 0x3E,                                     "Retry upon entry into a new cell" },
  { 0x3F,                                     "Retry upon entry into a new cell" },
  { 0x5F,                                     "Semantically incorrect messsage" },
  { 0x60,                                     "Invalid mandatory IE" },
  { 0x61,                                     "Message type non-existant or not implemented" },
  { 0x62,                                     "Message type not compatible with the protocol state" },
  { 0x63,                                     "IE non-existant or not implemented" },
  { 0x64,                                     "Conditinal IE error" },
  { 0x65,                                     "Message not compatible with the protocol state" },
  { 0x6F,                                     "Protocol error, unspecified" },
  { 0, NULL }
};


static const value_string identity_type_vals[] = {
  { 0x01,                                     "IMSI" },
  { 0x02,                                     "IMEI" },
  { 0x03,                                     "IMEISV" },
  { 0x04,                                     "TMSI" },
  { 0, NULL }
};

static const value_string mobile_id_odd_vals[] = {
  { 0x00,                                     "Even number of bytes" },
  { 0x01,                                     "Odd number of bytes" },
  { 0, NULL }
};

static const value_string network_feature_support_vals[] = {
  { 0x00,                                     "LCS-MOLR via PS Domain not supported" },
  { 0x01,                                     "LCS-MOLR via PS Domain supported" },
  { 0, NULL }
};

static const value_string tmsi_status_vals[] = {
  { 0x00,                                     "No valid TMSI available" },
  { 0x01,                                     "Valid TMSI available" },
  { 0, NULL }
};

static const value_string mobile_id_type_vals[] = {
  { 0x00,                                     "No Id" },
  { 0x01,                                     "IMSI" },
  { 0x02,                                     "IMEI" },
  { 0x03,                                     "IMEISV" },
  { 0x04,                                     "(P)TMSI" },
  { 0, NULL }
};

static const value_string sm_cause_vals[] = {
  { 0x08,                                     "Operator determined barring" },
  { 0x19,                                     "LLC or SNDCP failure (GSM only)" },
  { 0x1A,                                     "Insufficient resources" },
  { 0x1B,                                     "Missing or unknown APN" },
  { 0x1C,                                     "Unknown PDP address or PDP type" },
  { 0x1D,                                     "User authentication failure" },
  { 0x1E,                                     "Activation rejected by GGSN" },
  { 0x1F,                                     "Activation rejected, unspecified" },
  { 0x20,                                     "Service option not supported" },
  { 0x21,                                     "Requested service option not subscribed" },
  { 0x22,                                     "Service option temporarily out of service" },
  { 0x23,                                     "NSAPI already used (not sent)" },
  { 0x24,                                     "Regular deactivation" },
  { 0x25,                                     "Qos not accepted" },
  { 0x26,                                     "Network failure" },
  { 0x27,                                     "Reactivation required" },
  { 0x28,                                     "Feature not supported" },
  { 0x29,                                     "Semantic error in the TFT operartion" },
  { 0x2A,                                     "Syntactical error in the TFT operartion" },
  { 0x2B,                                     "Unknown PDP context" },
  { 0x2C,                                     "Semantic errors in packet filter(s)" },
  { 0x2D,                                     "Syntactical errors in packet filter(s)" },
  { 0x2E,                                     "PDP context without TFT activated" },
  { 0x50,                                     "Invalid transaction identifier value" },
  { 0x5F,                                     "Semantically incorrect message" },
  { 0x60,                                     "Invalid mandatory information" },
  { 0x61,                                     "Message type non-existant or not implemented" },
  { 0x62,                                     "Message type not compatible with the protocol state" },
  { 0x63,                                     "Information element non-existant or not implemented" },
  { 0x64,                                     "Conditional IE error" },
  { 0x65,                                     "Message not compatible with the protocol state" },
  { 0x6F,                                     "Protocol error, unspecified" },
  { 0, NULL }
};

static const value_string tear_down_indicator_vals[] = {
  { 0x00,                                     "Tear down not requested" },
  { 0x01,                                     "Tear down requested" },
  { 0, NULL }
};

static const value_string update_result_vals[] = {
  { 0x00,                                     "RA updated" },
  { 0x01,                                     "Combined RA/LA updated" },
  { 0, NULL }
};

static const value_string update_type_vals[] = {
  { 0x00,                                     "RA updating" },
  { 0x01,                                     "Combined RA/LA updating" },
  { 0x02,                                     "Combined RA/LA updating with IMSI attach" },
  { 0x03,                                     "Periodic updating" },
  { 0, NULL }
};

int dissect_gsm24008gmm_fixed_len_ie(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, int length, char * text, int tag);

#define dissect_gsm24008gmm_v_ie(tvb, pinfo, main_tree, msg_len, offset, length, text) \
        dissect_gsm24008gmm_fixed_len_ie(tvb, pinfo, main_tree, msg_len, offset, length, text, FALSE)

#define dissect_gsm24008gmm_tv_ie(tvb, pinfo, main_tree, msg_len, offset, length, text) \
        dissect_gsm24008gmm_fixed_len_ie(tvb, pinfo, main_tree, msg_len, offset, length, text, TRUE)

int dissect_gsm24008gmm_fixed_len_gmm_cause(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag);

#define dissect_gsm24008gmm_v_gmm_cause(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_gmm_cause(tvb, pinfo, main_tree, msg_len, offset, text, FALSE)

#define dissect_gsm24008gmm_tv_gmm_cause(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_gmm_cause(tvb, pinfo, main_tree, msg_len, offset, text, TRUE)

int dissect_gsm24008gmm_fixed_len_sm_cause(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag);

#define dissect_gsm24008gmm_v_sm_cause(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_sm_cause(tvb, pinfo, main_tree, msg_len, offset, text, FALSE)

#define dissect_gsm24008gmm_tv_sm_cause(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_sm_cause(tvb, pinfo, main_tree, msg_len, offset, text, TRUE)

int dissect_gsm24008gmm_fixed_len_gprs_timer(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag, int len_present);

#define dissect_gsm24008gmm_v_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text, FALSE, FALSE)

#define dissect_gsm24008gmm_tv_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text, TRUE, FALSE)
#define dissect_gsm24008gmm_tlv_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_fixed_len_gprs_timer(tvb, pinfo, main_tree, msg_len, offset, text, TRUE, TRUE)


int dissect_gsm24008gmm_var_len_ie(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag);

#define dissect_gsm24008gmm_lv_ie(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_var_len_ie(tvb, pinfo, main_tree, msg_len, offset, text, FALSE)

#define dissect_gsm24008gmm_tlv_ie(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_var_len_ie(tvb, pinfo, main_tree, msg_len, offset, text, TRUE)

int dissect_gsm24008gmm_var_len_mobile_id(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag);

#define dissect_gsm24008gmm_lv_mobile_id(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_var_len_mobile_id(tvb, pinfo, main_tree, msg_len, offset, text, FALSE)

#define dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, main_tree, msg_len, offset, text) \
        dissect_gsm24008gmm_var_len_mobile_id(tvb, pinfo, main_tree, msg_len, offset, text, TRUE)

#define CHECK_OFFSET(LEN)           \
        if((offset + LEN) > msg_len)      \
        {           \
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR ***");           \
          if (check_col(pinfo->cinfo, COL_INFO))           \
          {           \
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");           \
          }           \
          offset = msg_len;           \
          break;              \
       }

/* Code to dissect the GMM messages */

#define UPDATE_OFFSET(LEN)           \
        if(LEN)           \
        {           \
           offset += LEN;           \
        }           \
        else           \
        {           \
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR ***");           \
          if (check_col(pinfo->cinfo, COL_INFO))           \
          {           \
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");           \
          }           \
          offset = msg_len;           \
          break;              \
       }

/* Code to dissect the GMM messages */
void
dissect_gsm24008gmm(tvbuff_t *tvb, packet_info *pinfo, proto_tree *main_tree)
{
  int offset=0, msg_len, ie_len;
  guint msg_byte0, msg_byte1 = 0;
  guint tag;
  guint8 msg_type;                          /* Message Type */
  guint8 ti_skip, skip = FALSE;
  tvbuff_t *next_tvb;                       
  int ti_extended = FALSE;

  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;
  proto_tree *ie_tree=NULL;

  /* load the display labels   */
  pinfo->current_proto = "gsm24008gmm";
  
  if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
    col_add_str(pinfo->cinfo, COL_PROTOCOL, "GSM GMM");
  if (check_col(pinfo->cinfo, COL_INFO))
    col_clear(pinfo->cinfo, COL_INFO);

  msg_len = tvb_length_remaining(tvb, offset) -3;

  if (main_tree) {
    /* create display subtree for the protocol */
    ti = proto_tree_add_item(main_tree, proto_gsm24008gmm, tvb, offset, msg_len, FALSE);
    tree = proto_item_add_subtree(ti, ett_gsm24008gmm);
  }

  msg_byte0 = tvb_get_guint8(tvb, offset);

  ti_skip = msg_byte0 & SKIP_MASK;

  if((ti_skip & TIO_MASK) == TIO_MASK)
  {
    ti_extended = TRUE;
  }

  if(!ti_extended)
  {
    msg_type = tvb_get_guint8(tvb, offset + 1);
  }
  else
  {
    msg_byte1 = tvb_get_guint8(tvb, offset + 1);
    msg_type = tvb_get_guint8(tvb, offset + 2);
  }

  switch(msg_type)
  {
    case ATTACH_REQUEST:
    case ATTACH_ACCEPT:
    case ATTACH_COMPLETE:
    case ATTACH_REJECT:
    case DETACH_REQUEST:
    case DETACH_ACCEPT:
    case ROUTING_AREA_UPDATE_REQUEST:
    case ROUTING_AREA_UPDATE_ACCEPT:
    case ROUTING_AREA_UPDATE_COMPLETE:
    case ROUTING_AREA_UPDATE_REJECT:
    case SERVICE_REQUEST:
    case SERVICE_ACCEPT:
    case SERVICE_REJECT:
    case P_TMSI_REALLOCATION_COMMAND:
    case P_TMSI_REALLOCATION_RESPONSE:
    case AUTHENTICATION_AND_CIPHERING_REQUEST:
    case AUTHENTICATION_AND_CIPHERING_RESPONSE:
    case AUTHENTICATION_AND_CIPHERING_REJECT:
    case AUTHENTICATION_AND_CIPHERING_FAILURE:
    case IDENTITY_REQUEST:
    case IDENTITY_RESPONSE:
    case GMM_STATUS:
    case GMM_INFORMATION:
    {
      skip = TRUE;
      break;
    }
    default:
    {
      skip = FALSE;
      break;
    }
  }


  if (tree) 
  {
    /* create display subtree for the protocol */
    if( (skip) && (ti_skip) )
    {
      proto_item_set_text(ti, "GSM GMM: %s (*** SKIP ***)", rd_match_strval(msg_type,msg_type_vals));
    }
    else
    {
      proto_item_set_text(ti, "GSM GMM: %s", rd_match_strval(msg_type,msg_type_vals));
    }

    if(skip)
    {
      proto_tree_add_uint(tree, hf_gsm24008gmm_pd, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_skip, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_msg_type, tvb, offset + 1, 1, msg_type);
    }
    else if (!ti_extended)
    {
      proto_tree_add_uint(tree, hf_gsm24008gmm_pd, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_ti_flag, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_tio, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_msg_type, tvb, offset + 1, 1, msg_type);
    }
    else
    {
      proto_tree_add_uint(tree, hf_gsm24008gmm_pd, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_ti_flag, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_tie, tvb, offset + 1, 1, msg_byte1);
      proto_tree_add_uint(tree, hf_gsm24008gmm_msg_type, tvb, offset + 2, 1, msg_type);
    }
  }
  if (check_col(pinfo->cinfo, COL_INFO))
  {
    if( (skip) && (ti_skip) )
    {
      col_append_fstr(pinfo->cinfo, COL_INFO, "+++%s(*** SKIP ***)(",rd_match_strval(msg_type,msg_type_short_vals));
    }
    else
    {
      col_append_fstr(pinfo->cinfo, COL_INFO, "+++%s(",rd_match_strval(msg_type,msg_type_short_vals));
    }
  }

  if(!ti_extended)
  {
    offset += 2;
  }
  else
  {
    offset += 3;
  }

  switch(msg_type)
  {
    case ATTACH_REQUEST:
    {
      /* Mandatory IEs */
      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "MS Network Capability: ");
      UPDATE_OFFSET(ie_len)

      /* 1/2 Byte Attach Type */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Attach Type: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_attach_type_attach_type_low, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_attach_type_follow_on_low, tvb, offset, 1, msg_byte0);
      /* 1/2 Byte Cipher Key Type */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "GPRS Ciphering Key Sequence No: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ciphering_key_sn_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 2, "DRX Parameters: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_lv_mobile_id(tvb, pinfo, tree, msg_len, offset, "P-TMSI or IMSI: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 6, "Old Routing Area Identification: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "MS Radio Access Capability: ");
      UPDATE_OFFSET(ie_len)
      
      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x19:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 4, "Old P-TMSI Signature: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x17:
          {
            ie_len = dissect_gsm24008gmm_tv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Requested READY Timer Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x33:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 3, "PS LCS Capability: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            if( (tag & 0xF0) == 0x90)
            {
              CHECK_OFFSET(1)
              msg_byte1 = tvb_get_guint8(tvb, offset);
              ti = proto_tree_add_text(tree, tvb, offset, 1, "TMSI Status: ");
              ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
              proto_tree_add_uint(ie_tree, hf_gsm24008gmm_tmsi_status, tvb, offset, 1, msg_byte0);
              offset++;
            }
            else
            {
              proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
              if (check_col(pinfo->cinfo, COL_INFO))
              {
                col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
              }
              offset = msg_len;
            }
            break;
          }
        }
      }
      break;
    }
    case ATTACH_ACCEPT:
    {
      /* Mandatory IEs */
      /* 1/2 Byte Attach Result */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Attach Result: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_attach_result_low, tvb, offset, 1, msg_byte0);

      /* 1/2 Byte Cipher Force to standby */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      ie_len = dissect_gsm24008gmm_v_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Periodic Update Timer: ");
      UPDATE_OFFSET(ie_len)

      /* 1/2 Byte Radio Pri SMS */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Radio Priority for SMS: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_radio_priority_low, tvb, offset, 1, msg_byte0);

      /* 1/2 Byte Radio Pri TOMB */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Radio Priority for TOMB: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_radio_priority_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 6, "Routing Area Identification: ");
      UPDATE_OFFSET(ie_len)

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x19:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 4, "P-TMSI Signature: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x17:
          {
            ie_len = dissect_gsm24008gmm_tv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Negotiated READY Timer Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x18:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "Allocated P-TMSI: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x23:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "MS Identity: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x25:
          {
            ie_len = dissect_gsm24008gmm_tv_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x2A:
          {
            ie_len = dissect_gsm24008gmm_tlv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "T3302 Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x8C:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 1, "Cell Notification: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x4A:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Equivalent PLMNs ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            if( (tag & 0xF0) == 0xB0)
            {
              CHECK_OFFSET(1)
              msg_byte1 = tvb_get_guint8(tvb, offset);
              ti = proto_tree_add_text(tree, tvb, offset, 1, "Network Feature Support: ");
              ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
              proto_tree_add_uint(ie_tree, hf_gsm24008gmm_network_feature_support_low, tvb, offset, 1, msg_byte0);
              offset++;
            }
            else
            {
              proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
              if (check_col(pinfo->cinfo, COL_INFO))
              {
                col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
              }
              offset = msg_len;
            }
            break;
          }
        }
      }
      break;
    }
    case ATTACH_COMPLETE:
    {
      break; /* Nothing in this one */
    }
    case ATTACH_REJECT:
    {
      ie_len = dissect_gsm24008gmm_v_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if(tag == 0x2A)
        {
          ie_len = dissect_gsm24008gmm_tlv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "T3302 Value: ");
          UPDATE_OFFSET(ie_len)
          break;
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case DETACH_REQUEST:
    {
      /* Mandatory IEs */
      /* 1/2 Byte Detach Type */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Detach Type: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_detach_type_detach_type_low, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_detach_type_power_off_low, tvb, offset, 1, msg_byte0);

      /* 1/2 Byte Force to Standby or spare */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby (N/W->MS) | Spare: (MS->NW)");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x25:
          {
            ie_len = dissect_gsm24008gmm_tv_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x18:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "P-TMSI: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x19:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "P-TMSI Signature: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
            if (check_col(pinfo->cinfo, COL_INFO))
            {
              col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
            }
            offset = msg_len;
            break;
          }
        }
      }
      break;
    }
    case DETACH_ACCEPT:
    {
      if(offset < msg_len)
      {
        /* Mandatory IEs  - N/W->MS Only */
        /* 1/2 Byte Force to Standby */
        CHECK_OFFSET(1)
        msg_byte0 = tvb_get_guint8(tvb, offset);
        ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
        ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
        proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_low, tvb, offset, 1, msg_byte0);
        offset++;
      }
      break;
    }
    case ROUTING_AREA_UPDATE_REQUEST:
    {
      /* 1/2 Byte Attach Type */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Update Type: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_update_type_update_type_low, tvb, offset, 1, msg_byte0);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_update_type_follow_on_low, tvb, offset, 1, msg_byte0);
      /* 1/2 Byte Cipher Key Type */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "GPRS Ciphering Key Sequence No: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ciphering_key_sn_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 6, "Old Routing Area Identification: ");

      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "MS Radio Access Capability: ");
      UPDATE_OFFSET(ie_len)
      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x19:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 4, "Old P-TMSI Signature: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x17:
          {
            ie_len = dissect_gsm24008gmm_tv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Requested READY Timer Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x27:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 3, "DRX parameter: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x18:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "P-TMSI: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x31:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "MS Network Capability: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x32:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "PDP Context status: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x33:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "PS LCS Capability: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            if( (tag & 0xF0) == 0x90)
            {
              CHECK_OFFSET(1)
              msg_byte1 = tvb_get_guint8(tvb, offset);
              ti = proto_tree_add_text(tree, tvb, offset, 1, "TMSI Status: ");
              ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
              proto_tree_add_uint(ie_tree, hf_gsm24008gmm_tmsi_status, tvb, offset, 1, msg_byte0);
              offset++;
            }
            else
            {
              proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
              if (check_col(pinfo->cinfo, COL_INFO))
              {
                col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
              }
              offset = msg_len;
            }
            break;
          }
        }
      }
      break;
    }
    case ROUTING_AREA_UPDATE_ACCEPT:
    {
      /* 1/2 Byte Force to Standby */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_low, tvb, offset, 1, msg_byte0);
      /* 1/2 Byte Update Result */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Update Result: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_update_result_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      ie_len = dissect_gsm24008gmm_v_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Periodic Update Timer: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 6, "Routing Area Identification: ");
      UPDATE_OFFSET(ie_len)

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x19:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 4, "P-TMSI Signature: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x18:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "Allocated P-TMSI: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x23:
          {
            ie_len = dissect_gsm24008gmm_tlv_mobile_id(tvb, pinfo, tree, msg_len, offset, "MS Identity: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x26:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "List of Receive N-PDU Numbers: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x17:
          {
            ie_len = dissect_gsm24008gmm_tv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "Negotiated READY Timer Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x25:
          {
            ie_len = dissect_gsm24008gmm_tv_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x2A:
          {
            ie_len = dissect_gsm24008gmm_tlv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "T3302 Value: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x8C:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 1, "Cell Notification: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x4A:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Equivalent PLMNs: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x32:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "PDP context status: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            if( (tag & 0xF0) == 0xB0)
            {
              CHECK_OFFSET(1)
              msg_byte1 = tvb_get_guint8(tvb, offset);
              ti = proto_tree_add_text(tree, tvb, offset, 1, "Network Feature Support: ");
              ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
              proto_tree_add_uint(ie_tree, hf_gsm24008gmm_network_feature_support_low, tvb, offset, 1, msg_byte0);
              offset++;
            }
            else
            {
              proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
              if (check_col(pinfo->cinfo, COL_INFO))
              {
                col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
              }
              offset = msg_len;
            }
            break;
          }
        }
      }
      break;
    }
    case ROUTING_AREA_UPDATE_COMPLETE:
    {
      if(offset < msg_len)
      {
        ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "List of Receive N-PDU Numbers: ");
        UPDATE_OFFSET(ie_len)
      }
      break;
    }
    case ROUTING_AREA_UPDATE_REJECT:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_v_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
      UPDATE_OFFSET(ie_len)

      /* 1/2 Byte Force to Standby + spare */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_low, tvb, offset, 1, msg_byte0);
      offset++;

      if(offset < msg_len)
      {
        ie_len = dissect_gsm24008gmm_tlv_gprs_timer(tvb, pinfo, tree, msg_len, offset, "T3302 Value: ");
        UPDATE_OFFSET(ie_len)
      }
      break;
    }
    case SERVICE_REQUEST:
    case SERVICE_ACCEPT:
    case SERVICE_REJECT:
    {
      /* SERVICE... are UMTS Only */
      /* Default for unhandled messages */
      if(tree)
      {
        ti = proto_tree_add_text(tree, tvb, offset, -1, "UNDECODED: ");
        tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(data_handle, next_tvb, pinfo, tree);
      }
      break;
    }
    case P_TMSI_REALLOCATION_COMMAND:
    {
      /* Mandatory IEs */
      ie_len = dissect_gsm24008gmm_lv_mobile_id(tvb, pinfo, tree, msg_len, offset, "Allocated P-TMSI: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_v_ie(tvb, pinfo, tree, msg_len, offset, 6, "Routing Area Identification: ");
      UPDATE_OFFSET(ie_len)

      /* 1/2 Byte Force to Standby + spare */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_low, tvb, offset, 1, msg_byte0);
      offset++;

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if(tag == 0x19)
        {
          ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 4, "Old P-TMSI Signature: ");
          UPDATE_OFFSET(ie_len)
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case P_TMSI_REALLOCATION_RESPONSE:
    {
      break; /* Nothing in this one */
    }
    case AUTHENTICATION_AND_CIPHERING_REQUEST:
    {
      /* 1/2 Byte Ciphering Algorithm */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Ciphering Algorithm: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ciphering_algorithm_low, tvb, offset, 1, msg_byte0);
      /* 1/2 Byte IMEISV Request */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "IMEISV Request: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_imeisv_request_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      /* 1/2 Byte Force to Standby */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_low, tvb, offset, 1, msg_byte0);
      /* 1/2 Byte A&C Reference Number */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "A&C Reference Number: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ac_reference_number_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x21:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 17, "Authentication Parameter RAND ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x28:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Authentication Parameter AUTN: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            if( (tag & 0xF0) == 0x80)
            {
              CHECK_OFFSET(1)
              msg_byte1 = tvb_get_guint8(tvb, offset);
              ti = proto_tree_add_text(tree, tvb, offset, 1, "Ciphering Key Sequence Number: ");
              ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
              proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ciphering_key_sn_low, tvb, offset, 1, msg_byte0);
              offset++;
            }
            else
            {
              proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
              if (check_col(pinfo->cinfo, COL_INFO))
              {
                col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
              }
              offset = msg_len;
            }
            break;
          }
        }
      }
      break;
    }
    case AUTHENTICATION_AND_CIPHERING_RESPONSE:
    {
      /* 1/2 Byte Ciphering Algorithm */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Ciphering Algorithm: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_ciphering_algorithm_low, tvb, offset, 1, msg_byte0);
      offset++; /* For both 1/2 byte and spare  */

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x22:
          {
            ie_len = dissect_gsm24008gmm_tv_ie(tvb, pinfo, tree, msg_len, offset, 5, "Authentication Parameter Response ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x23:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "IMEISV: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x29:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Authentication Response Parameter (Extension): ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
            if (check_col(pinfo->cinfo, COL_INFO))
            {
              col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
            }
            offset = msg_len;
            break;
          }
        }
      }
      break;
    }
    case AUTHENTICATION_AND_CIPHERING_REJECT:
    {
      break; /* Nothing in this one */
    }
    case AUTHENTICATION_AND_CIPHERING_FAILURE:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_v_gmm_cause(tvb, pinfo, tree, msg_len, offset, "GMM Cause: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if(tag == 0x30)
        {
          ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Authentication Failure Parameter: ");
          UPDATE_OFFSET(ie_len)
          break;
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case IDENTITY_REQUEST:
    {
      /* Mandatory IEs */
      /* 1/2 Byte Attach Result */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Identity Type: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_identity_type_low, tvb, offset, 1, msg_byte0);

      /* 1/2 Byte Cipher Force to standby */
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Force to Standby: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_force_to_standby_hi, tvb, offset, 1, msg_byte0);

      offset++; /* For both 1/2 bytes */

      break;
    }
    case IDENTITY_RESPONSE:
    {
      /* Mandatory IEs */
      ie_len = dissect_gsm24008gmm_lv_mobile_id(tvb, pinfo, tree, msg_len, offset, "Mobile Identity: ");
      UPDATE_OFFSET(ie_len)

      break;
    }
    case GMM_STATUS:
    case GMM_INFORMATION:
    {
      /* Default for unhandled messages */
      if(tree)
      {
        ti = proto_tree_add_text(tree, tvb, offset, -1, "UNDECODED: ");
        tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(data_handle, next_tvb, pinfo, tree);
      }
      break;
    }
    case ACTIVATE_PDP_CONTEXT_REQUEST:
    {
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Requested NSAPI: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_nsapi, tvb, offset, 1, msg_byte0);
      offset++;

      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Requested LLC SAPI: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_llc_sapi, tvb, offset, 1, msg_byte0);
      offset++;

      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "Requested QOS: ");
      UPDATE_OFFSET(ie_len)

      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "Requested PDP Address: ");
      UPDATE_OFFSET(ie_len)

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x28:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Access Point Name: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x27:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Protocol Configuration Options: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
            if (check_col(pinfo->cinfo, COL_INFO))
            {
              col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
            }
            offset = msg_len;
            break;
          }
        }
      }
      break;
    }
    case ACTIVATE_PDP_CONTEXT_ACCEPT:
    {
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Negotiated LLC SAPI: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_llc_sapi, tvb, offset, 1, msg_byte0);
      offset++;

      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "Negotiated QOS: ");
      UPDATE_OFFSET(ie_len)

      /* 1/2 Byte Radio Pri */
      CHECK_OFFSET(1)
      msg_byte0 = tvb_get_guint8(tvb, offset);
      ti = proto_tree_add_text(tree, tvb, offset, 1, "Radio Priority: ");
      ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
      proto_tree_add_uint(ie_tree, hf_gsm24008gmm_radio_priority_low, tvb, offset, 1, msg_byte0);
      offset++;

      /* Optional IEs */
      while(offset < msg_len)
      {
        CHECK_OFFSET(1)
        tag = tvb_get_guint8(tvb, offset);
        switch(tag)
        {
          case 0x2B:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "PDP Address: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x27:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Protocol Configuration Options: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          case 0x34:
          {
            ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Packet Flow Identifier: ");
            UPDATE_OFFSET(ie_len)
            break;
          }
          default:
          {
            proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
            if (check_col(pinfo->cinfo, COL_INFO))
            {
              col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
            }
            offset = msg_len;
            break;
          }
        }
      }
      break;
    }
    case ACTIVATE_PDP_CONTEXT_REJECT:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_v_sm_cause(tvb, pinfo, tree, msg_len, offset, "SM Cause: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if(tag == 0x27)
        {
          ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Protocol Configuration Optione: ");
          UPDATE_OFFSET(ie_len)
          break;
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case REQUEST_PDP_CONTEXT_ACTIVATION:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_lv_ie(tvb, pinfo, tree, msg_len, offset, "Offered PDP Address: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if(tag == 0x28)
        {
          ie_len = dissect_gsm24008gmm_tlv_ie(tvb, pinfo, tree, msg_len, offset, "Access Point Name: ");
          UPDATE_OFFSET(ie_len)
          break;
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case REQUEST_PDP_CONTEXT_ACTIVATION_REJECT:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_v_sm_cause(tvb, pinfo, tree, msg_len, offset, "SM Cause: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR ***");
        if (check_col(pinfo->cinfo, COL_INFO))
        {
          col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
        }
      }
      break;
    }
    case DEACTIVATE_PDP_CONTEXT_REQUEST:
    {
      /* Mandatory offset */
      ie_len = dissect_gsm24008gmm_v_sm_cause(tvb, pinfo, tree, msg_len, offset, "SM Cause: ");
      UPDATE_OFFSET(ie_len)

      if(offset < msg_len)
      {
        tag = tvb_get_guint8(tvb, offset);
        if((tag & 0xF0) == 0x90)
        {
          CHECK_OFFSET(1)
          msg_byte1 = tvb_get_guint8(tvb, offset);
          ti = proto_tree_add_text(tree, tvb, offset, 1, "Tear Down Indicator: ");
          ie_tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
          proto_tree_add_uint(ie_tree, hf_gsm24008gmm_tear_down_indicator_low, tvb, offset, 1, msg_byte0);
          offset++;
          break;
        }
        else
        {
          proto_tree_add_text(main_tree, tvb, offset, -1, "*** ERROR in optional ie's***");
          if (check_col(pinfo->cinfo, COL_INFO))
          {
            col_append_fstr(pinfo->cinfo, COL_INFO, "*** ERROR ***");
          }
        }
      }
      break;
    }
    case DEACTIVATE_PDP_CONTEXT_ACCEPT:
    {
      break; /* Nothing in this one */
    }
    case MODIFY_PDP_CONTEXT_REQUEST_NETWORK_TO_MS:
    case MODIFY_PDP_CONTEXT_ACCEPT_MS_TO_NETWORK:
    case MODIFY_PDP_CONTEXT_REQUEST_MS_TO_NETWORK:
    case MODIFY_PDP_CONTEXT_ACCEPT_NETWORK_TO_MS:
    case MODIFY_PDP_CONTEXT_REJECT:
    case ACTIVATE_SECONDARY_PDP_CONTEXT_REQUEST:
    case ACTIVATE_SECONDARY_PDP_CONTEXT_ACCEPT:
    case ACTIVATE_SECONDARY_PDP_CONTEXT_REJECT:
    case SM_RESERVED_1:
    case SM_RESERVED_2:
    case SM_RESERVED_3:
    case SM_RESERVED_4:
    case SM_RESERVED_5:
    case SM_STATUS:
    {
      /* Default for unhandled messages */
      if(tree)
      {
        ti = proto_tree_add_text(tree, tvb, offset, -1, "UNDECODED: ");
        tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);
        next_tvb = tvb_new_subset(tvb, offset, -1, -1);
        call_dissector(data_handle, next_tvb, pinfo, tree);
      }
      break;
    }
    default:
      break; /* Will be caught elsewhere */
  }

  if (check_col(pinfo->cinfo, COL_INFO))
    col_append_fstr(pinfo->cinfo, COL_INFO, ")");

  return;
}

/* Displays a V or TV format ie - returns length passed in */
int dissect_gsm24008gmm_fixed_len_ie(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, int length, char * text, int tag)
{
  guint8 ie_tag = 0;
  tvbuff_t *next_tvb;                       

  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  if( (offset + length) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, length, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    ie_tag = tvb_get_guint8(tvb, offset);
    proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
    if(length > 1)
    {
      next_tvb = tvb_new_subset(tvb, offset + 1, length - 1, -1);
      call_dissector(data_handle, next_tvb, pinfo, tree);
    }
  }
  else
  {
    next_tvb = tvb_new_subset(tvb, offset, length, -1);
    call_dissector(data_handle, next_tvb, pinfo, tree);
  }

  return(length);
}

/* Displays a V or TV format gmm cause returns length */
int dissect_gsm24008gmm_fixed_len_gmm_cause(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag)
{
  int length;
  guint8 ie_tag = 0;
  guint8 msg_byte0 = 0;

  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  length = (tag) ? 2: 1;

  if( (offset + length) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, length, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    ie_tag = tvb_get_guint8(tvb, offset);
    msg_byte0 = tvb_get_guint8(tvb, offset +1);
    proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
    proto_tree_add_uint(tree, hf_gsm24008gmm_gmm_cause, tvb, offset + 1, 1, msg_byte0);
  }
  else
  {
    msg_byte0 = tvb_get_guint8(tvb, offset );
    proto_tree_add_uint(tree, hf_gsm24008gmm_gmm_cause, tvb, offset, 1, msg_byte0);
  }

  if (check_col(pinfo->cinfo, COL_INFO))
    col_append_fstr(pinfo->cinfo, COL_INFO, "Cause=%s",rd_match_strval(msg_byte0, gmm_cause_vals));

  return(length);
}

/* Displays a V or TV format sm cause returns length */
int dissect_gsm24008gmm_fixed_len_sm_cause(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag)
{
  int length;
  guint8 ie_tag = 0;
  guint8 msg_byte0 = 0;

  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  length = (tag) ? 2: 1;

  if( (offset + length) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, length, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    ie_tag = tvb_get_guint8(tvb, offset);
    msg_byte0 = tvb_get_guint8(tvb, offset +1);
    proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
    proto_tree_add_uint(tree, hf_gsm24008gmm_sm_cause, tvb, offset + 1, 1, msg_byte0);
  }
  else
  {
    msg_byte0 = tvb_get_guint8(tvb, offset );
    proto_tree_add_uint(tree, hf_gsm24008gmm_sm_cause, tvb, offset, 1, msg_byte0);
  }

  if (check_col(pinfo->cinfo, COL_INFO))
    col_append_fstr(pinfo->cinfo, COL_INFO, "Cause=%s",rd_match_strval(msg_byte0, sm_cause_vals));

  return(length);
}

/* Displays a V or TV format GPRS timer returns length */
int dissect_gsm24008gmm_fixed_len_gprs_timer(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag, int len_present)
{
  int length;
  guint8 units = 0;
  guint8 value = 0;
  guint8 ie_tag = 0;
  guint8 msg_byte0 = 0;
  guint8 ie_len = 0;

  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  length = (tag) ? 2: 1;
  if(len_present)
  {
    length++;
  }

  if( (offset + length) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, length, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    if(len_present)
    {
      ie_tag = tvb_get_guint8(tvb, offset);
      ie_len = tvb_get_guint8(tvb, offset +1);
      msg_byte0 = tvb_get_guint8(tvb, offset +2);
      proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
      proto_tree_add_uint(tree, hf_gsm24008gmm_ie_len1, tvb, offset + 1, 1, ie_len);
      proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_units, tvb, offset + 2, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_val, tvb, offset + 2, 1, msg_byte0);
    }
    else
    {
      ie_tag = tvb_get_guint8(tvb, offset);
      msg_byte0 = tvb_get_guint8(tvb, offset +1);
      proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
      proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_units, tvb, offset + 1, 1, msg_byte0);
      proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_val, tvb, offset + 1, 1, msg_byte0);
    }
  }
  else
  {
    msg_byte0 = tvb_get_guint8(tvb, offset );
    proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_units, tvb, offset, 1, msg_byte0);
    proto_tree_add_uint(tree, hf_gsm24008gmm_gprs_timer_val, tvb, offset, 1, msg_byte0);
  }

  switch(msg_byte0 & 0xE0)
  {
    case 0x00:
    {
      ti = proto_tree_add_text(tree, tvb, offset, length, 
               "Duration: %d seconds", ((int)(msg_byte0 & 0x1F)) * 2);
      break;
    }
    case 0x20:
    {
      ti = proto_tree_add_text(tree, tvb, offset, length, 
               "Duration: %d minutes", ((int)(msg_byte0 & 0x1F)));
      break;
    }
    case 0x40:
    {
      ti = proto_tree_add_text(tree, tvb, offset, length, 
               "Duration: %d minutes", ((int)(msg_byte0 & 0x1F)) * 6);
      break;
    }
    default:
    {
      break;
    }
  }

  return(length);
}

/* Displays a LV or TLV format ie returning it's length */
int dissect_gsm24008gmm_var_len_ie(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag)
{
  guint8 ie_tag = 0, ie_len1;
  guint16 ie_len2 = 0;
  int ie_len = 0;
  int tag_len;
  tvbuff_t *next_tvb;                       
  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  if(tag)
  {
    if( (offset + 1) > msg_len)
    {
      return(0);
    }

    ie_tag = tvb_get_guint8(tvb, offset);
    tag_len = 1;
  }
  else
  {
    tag_len = 0;
  }

  if( (offset + tag_len + 1) > msg_len)
  {
    return(0);
  }

  ie_len1 = tvb_get_guint8(tvb, offset + tag_len);

  ie_len = ie_len1;

  if( (offset + tag_len + 1 + ie_len) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, ie_len + 1 + tag_len, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
    offset++;
  }

  proto_tree_add_uint(tree, hf_gsm24008gmm_ie_len1, tvb, offset, 1, ie_len1);
  offset ++;

  next_tvb = tvb_new_subset(tvb, offset, ie_len, -1);
  call_dissector(data_handle, next_tvb, pinfo, tree);

  return(tag_len + 1 + ie_len);
}

/* Displays a LV or TLV format mobile_id ie returning it's length */
int dissect_gsm24008gmm_var_len_mobile_id(tvbuff_t *tvb, packet_info *pinfo, 
    proto_tree *main_tree, int msg_len, int offset, char * text, int tag)
{
  char buffer[20];
  int i;
  guint8 ie_tag = 0, ie_len1;
  guint8 id_type, id_byte, id_val;
  int id_digits;
  int ie_len;
  int tag_len;
  /* Set up structures needed to add the protocol subtree and manage it */
  proto_item *ti=NULL;
  proto_tree *tree=NULL;

  if(tag)
  {
    if( (offset + 1) > msg_len)
    {
      return(0);
    }

    ie_tag = tvb_get_guint8(tvb, offset);
    tag_len = 1;
  }
  else
  {
    tag_len = 0;
  }

  if( (offset + tag_len + 1) > msg_len)
  {
    return(0);
  }

  ie_len1 = tvb_get_guint8(tvb, offset + tag_len);

  ie_len = ie_len1;

  if( (offset + tag_len + ie_len) > msg_len)
  {
    return(0);
  }

  ti = proto_tree_add_text(main_tree, tvb, offset, ie_len + 1 + tag_len, text);
  tree = proto_item_add_subtree(ti, ett_gsm24008gmm_ie);

  if(tag)
  {
    proto_tree_add_uint(tree, hf_gsm24008gmm_ie_tag, tvb, offset, 1, ie_tag);
    offset++;
  }

  proto_tree_add_uint(tree, hf_gsm24008gmm_ie_len1, tvb, offset, 1, ie_len1);
  offset ++;

  id_byte = tvb_get_guint8(tvb, offset);
  id_type = id_byte & 0x07;
  proto_tree_add_uint(tree, hf_gsm24008gmm_mobile_id_type, tvb, offset, 1, id_byte);
  proto_tree_add_uint(tree, hf_gsm24008gmm_mobile_id_odd, tvb, offset, 1, id_byte);

	if(id_byte & 0x04)
	{
		/* (P)TMSI */
      
		id_digits = (ie_len -1) * 2;

    if(id_byte & 0x08)
    {
      /* Odd number of bytes 
			 * Not expected, but we'll handle it anyway */
      id_digits--;
    }

    /* Build a string with ID in it */
    for(i = 0; (i < id_digits); i++)
    {
      if(i % 2)
      {
        id_val = id_byte & 0x0F;
      }
      else
      {
        /* Need new byte */
        id_byte = tvb_get_guint8(tvb, offset + 1 + (i / 2));
        id_val = (id_byte >> 4) & 0x0F;
      }
      buffer[i] = (id_val < 10) ? (id_val + '0') : (id_val - 10 + 'A');
    }
    buffer[i] = '\0';
	}
	else
	{
		/* IMSI etc */
    if(id_byte & 0x08)
    {
      /* Odd number of bytes */
      id_digits = (ie_len * 2) -1;
    }
    else
    {
      /* Even number of bytes */
      id_digits = (ie_len -1) * 2;
    }
  
    /* Build a string with ID in it */
    for(i = 0; (i < id_digits); i++)
    {
      if(i % 2)
      {
        /* Need new byte */
        id_byte = tvb_get_guint8(tvb, offset + 1 + (i / 2));
        id_val = id_byte & 0x0F;
      }
      else
      {
        id_val = (id_byte >> 4) & 0x0F;
      }
      buffer[i] = (id_val < 10) ? (id_val + '0') : (id_val - 10 + 'A');
    }
    buffer[i] = '\0';
	}

  proto_tree_add_text(tree, tvb, offset, ie_len, "%s: %s", rd_match_strval(id_type, mobile_id_type_vals), buffer);

  if (check_col(pinfo->cinfo, COL_INFO))
    col_append_fstr(pinfo->cinfo, COL_INFO, "%s=%s",rd_match_strval(id_type, mobile_id_type_vals), buffer);

  return(tag_len + 1 + ie_len);
}
/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_gsm24008gmm(void)
{

/* Setup list of header fields  See Section 1.6.1 for details*/
  static hf_register_info hf[] = {
    { &hf_gsm24008gmm_msg_type,
      { "Msg Type", "gsm24008gmm.msg_type",
      FT_UINT8, BASE_HEX, VALS(msg_type_vals), 0x0,    
      "Message Type", HFILL }
    },
    { &hf_gsm24008gmm_skip,
      { "Skip", "gsm24008gmm.skip",
      FT_UINT8, BASE_HEX, NULL, SKIP_MASK,
      "Skip Indicator", HFILL }
    },
    { &hf_gsm24008gmm_ti_flag,
      { "TI Flag", "gsm24008gmm.ti_flag",
      FT_UINT8, BASE_DEC, NULL, TI_FLAG_MASK,
      "Transaction Id Flag", HFILL }
    },
    { &hf_gsm24008gmm_tio,
      { "TI", "gsm24008gmm.tio",
      FT_UINT8, BASE_HEX, NULL, TIO_MASK,
      "Transaction Id", HFILL }
    },
    { &hf_gsm24008gmm_tie,
      { "TI", "gsm24008gmm.tie",
      FT_UINT8, BASE_HEX, NULL, TIE_MASK,
      "Transaction Id (Extended)", HFILL }
    },
    { &hf_gsm24008gmm_pd,
      { "PD", "gsm24008gmm.pd",
      FT_UINT8, BASE_HEX, VALS(pd_vals), PD_MASK,
      "Protocol Discriminator", HFILL }
    },
    { &hf_gsm24008gmm_ie_tag,
      { "IEI", "gsm24008gmm.ie_tag",
      FT_UINT8, BASE_HEX, NULL, 0x0,    
      "Information Element Id", HFILL }
    },
    { &hf_gsm24008gmm_ie_len1,
      { "IE Length", "gsm24008gmm.ie_len1",
      FT_UINT8, BASE_DEC, NULL, 0x0,    
      "Information Element Lenghth", HFILL }
    },
    { &hf_gsm24008gmm_ie_len2,
      { "IE Length", "gsm24008gmm.ie_len2",
      FT_UINT16, BASE_DEC, NULL, 0x0,    
      "Information Element Lenghth", HFILL }
    },
    { &hf_gsm24008gmm_ac_reference_number_hi,
      { "Authentication and Ciphering Reference Number", "gsm24008gmm.ac_reference_number",
      FT_UINT8, BASE_HEX, NULL, 0xF0,
      "Authentication and Ciphering Reference Number IE", HFILL }
    },
    { &hf_gsm24008gmm_attach_result_low,
      { "Attach Result", "gsm24008gmm.attach_result",
      FT_UINT8, BASE_HEX, VALS(attach_result_vals), 0x0F,
      "Attach Result IE", HFILL }
    },
    { &hf_gsm24008gmm_attach_type_attach_type_low,
      { "Attach Type", "gsm24008gmm.attach_type",
      FT_UINT8, BASE_HEX, VALS(attach_type_vals), 0x03,
      "Attach Type IE - Attach type", HFILL }
    },
    { &hf_gsm24008gmm_attach_type_follow_on_low,
      { "Follow On", "gsm24008gmm.follow_on",
      FT_UINT8, BASE_HEX, NULL, 0x01,
      "Attach Type IE - Follow on pending", HFILL }
    },
    { &hf_gsm24008gmm_ciphering_algorithm_low,
      { "Ciphering Algorithm", "gsm24008gmm.ciphering_algorithm",
      FT_UINT8, BASE_HEX, VALS(ciphering_algorithm_vals), 0x0F,
      "Ciphering Algorithm", HFILL }
    },
    { &hf_gsm24008gmm_ciphering_key_sn_hi,
      { "Ciphering Key Sequence Number", "gsm24008gmm.ciphering_key_sn",
      FT_UINT8, BASE_HEX, NULL, 0xF0,
      "Ciphering Key Sequence Number IE", HFILL }
    },
    { &hf_gsm24008gmm_ciphering_key_sn_low,
      { "Ciphering Key Sequence Number", "gsm24008gmm.ciphering_key_sn",
      FT_UINT8, BASE_HEX, NULL, 0x0F,
      "Ciphering Key Sequence Number IE", HFILL }
    },
    { &hf_gsm24008gmm_detach_type_detach_type_low,
      { "Detach Type", "gsm24008gmm.detach_type",
      FT_UINT8, BASE_HEX, VALS(detach_type_vals), 0x07,
      "Detach Type IE - Detach Type", HFILL }
    },
    { &hf_gsm24008gmm_detach_type_power_off_low,
      { "Power Off", "gsm24008gmm.power_off",
      FT_UINT8, BASE_HEX, VALS(power_off_vals), 0x08,
      "Detach Type IE - Power Off", HFILL }
    },
    { &hf_gsm24008gmm_imeisv_request_hi,
      { "IMEISV Request", "gsm24008gmm.imeisv_request",
      FT_UINT8, BASE_HEX, VALS(imeisv_request_vals), 0xF0,
      "IMEISV RequestForce to Standby IE", HFILL }
    },
    { &hf_gsm24008gmm_force_to_standby_hi,
      { "Force to Standby", "gsm24008gmm.force_to_standby",
      FT_UINT8, BASE_HEX, VALS(force_to_standby_vals), 0xF0,
      "Force to Standby IE" , HFILL}
    },
    { &hf_gsm24008gmm_force_to_standby_low,
      { "Force to Standby", "gsm24008gmm.force_to_standby",
      FT_UINT8, BASE_HEX, VALS(force_to_standby_vals), 0x0F,
      "Force to Standby IE", HFILL }
    },
    { &hf_gsm24008gmm_gprs_timer_units,
      { "Units", "gsm24008gmm.gprs_timer_units",
      FT_UINT8, BASE_HEX, VALS(gprs_timer_units_vals), 0xE0,
      "GPRS Timer IE - Units", HFILL }
    },
    { &hf_gsm24008gmm_gprs_timer_val,
      { "Value", "gsm24008gmm.gprs_timer_val",
      FT_UINT8, BASE_HEX, NULL, 0x1F,
      "GPRS Timer IE - Value", HFILL }
    },
    { &hf_gsm24008gmm_gmm_cause,
      { "GMM Cause", "gsm24008gmm.gmm_cause",
      FT_UINT8, BASE_HEX, VALS(gmm_cause_vals), 0x0,
      "GMM Cause IE", HFILL }
    },
    { &hf_gsm24008gmm_identity_type_low,
      { "GMM Cause", "gsm24008gmm.identity_type",
      FT_UINT8, BASE_HEX, VALS(identity_type_vals), 0x03,
      "GMM Cause IE", HFILL }
    },
    { &hf_gsm24008gmm_llc_sapi,
      { "LLC Service Access Point Identifier", "gsm24008gmm.nsapi",
      FT_UINT8, BASE_HEX, NULL, 0x0F,
      "LLC Service Access Point Identifier IE", HFILL }
    },
    { &hf_gsm24008gmm_mobile_id_odd,
      { "Odd/Even Ind", "gsm24008gmm.mobile_id_odd",
      FT_UINT8, BASE_HEX, VALS(mobile_id_odd_vals), 0x08,
      "Mobile ID IE - Odd/Even Indication", HFILL }
    },
    { &hf_gsm24008gmm_mobile_id_type,
      { "Id Type", "gsm24008gmm.mobile_id_type",
      FT_UINT8, BASE_HEX, VALS(mobile_id_type_vals), 0x07,
      "Mobile ID IE - Type of identity", HFILL }
    },
    { &hf_gsm24008gmm_nsapi,
      { "Network Service Access Point Identifier", "gsm24008gmm.nsapi",
      FT_UINT8, BASE_HEX, NULL, 0x0F,
      "Network Service Access Point Identifier IE", HFILL }
    },
    { &hf_gsm24008gmm_network_feature_support_low,
      { "Units", "gsm24008gmm.network_feature_support",
      FT_UINT8, BASE_HEX, VALS(network_feature_support_vals), 0x80,
      "Network Feature Support IE", HFILL }
    },
    { &hf_gsm24008gmm_radio_priority_low,
      { "Radio Priority", "gsm24008gmm.radio_priority",
      FT_UINT8, BASE_HEX, NULL, 0x07,
      "Radio Priority IE", HFILL }
    },
    { &hf_gsm24008gmm_radio_priority_hi,
      { "Radio Priority", "gsm24008gmm.radio_priority",
      FT_UINT8, BASE_HEX, NULL, 0x70,
      "Radio Priority IE" , HFILL}
    },
    { &hf_gsm24008gmm_sm_cause,
      { "SM Cause", "gsm24008gmm.sm_cause",
      FT_UINT8, BASE_HEX, VALS(sm_cause_vals), 0x0,
      "SM Cause IE", HFILL }
    },
    { &hf_gsm24008gmm_tmsi_status,
      { "TMSI Status", "gsm24008gmm.tmsi_status",
      FT_UINT8, BASE_HEX, VALS(tmsi_status_vals), 0x01,
      "TMSI Status", HFILL }
    },
    { &hf_gsm24008gmm_tear_down_indicator_low,
      { "Tear Down Indicator", "gsm24008gmm.tear_down_indicator",
      FT_UINT8, BASE_HEX, VALS(tear_down_indicator_vals), 0x01,
      "Tear Down Indicator IE ", HFILL }
    },
    { &hf_gsm24008gmm_update_result_hi,
      { "Update Result", "gsm24008gmm.update_result",
      FT_UINT8, BASE_HEX, VALS(update_result_vals), 0x03,
      "Update Result IE", HFILL }
    },
    { &hf_gsm24008gmm_update_type_update_type_low,
      { "Update Type", "gsm24008gmm.update_type",
      FT_UINT8, BASE_HEX, VALS(update_type_vals), 0x03,
      "Update Type IE - Update type", HFILL }
    },
    { &hf_gsm24008gmm_update_type_follow_on_low,
      { "Follow On", "gsm24008gmm.follow_on",
      FT_UINT8, BASE_HEX, NULL, 0x01,
      "Update Type IE - Follow on request", HFILL }
    },
  };

/* Setup protocol subtree array */
  static gint *ett[] = {
    &ett_gsm24008gmm,
    &ett_gsm24008gmm_ie,
  };

  proto_gsm24008gmm = proto_register_protocol("GSM 24.008 GMM - GPRS Mobility Management Protocol", "GSM 24.008 GMM", "gsm24008gmm");

  register_dissector("gsm24008gmm", dissect_gsm24008gmm, proto_gsm24008gmm);

/* Required function calls to register the header fields and subtrees used */
  proto_register_field_array(proto_gsm24008gmm, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_gsm24008gmm(void)
{
    gsm24008gmm_handle = find_dissector("gsm24008gmm");
    data_handle = find_dissector("data");
}
