<?xml version='1.0' encoding="UTF-8"?>
<xsl:stylesheet 
  xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
  xmlns:redirect="org.apache.xalan.lib.Redirect"
  extension-element-prefixes="redirect"
  version="1.0">
  
<xsl:output omit-xml-declaration='yes'/>
<xsl:output method="text"/>

<!-- an xsl implementation of toUpper()/toLower() using XPath translate() -->
<xsl:variable name="lcletters">abcdefghijklmnopqrstuvwxyz</xsl:variable>
<xsl:variable name="ucletters">ABCDEFGHIJKLMNOPQRSTUVWXYZ</xsl:variable>

<xsl:template match="/MIB_Schema">
  <xsl:call-template name="gen-h"/>
</xsl:template>

<!--
    It turns out that using this index to look up Dns for
	each class is *much* faster.
-->

<xsl:key name="dn-index" match="//Dn" use="Label"/>






<xsl:template name="gen-h">
<redirect:open file="ipa-bscmgmt-mib.h"/>
<redirect:write file="ipa-bscmgmt-mib.h">
/* ipa-bscmgmt-mib.h
 * Definitions for ip.access BSC Management dissection
 * Copyright 2004, ip.access Ltd &lt;Neil.Piercy@ipaccess.com>
 *
 * $Id: ipa-bscmgmt-mib.xsl 1.1 2006/07/13 18:50:45BST Neil P Piercy (npp) Exp  $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs &lt;gerald@unicom.net>
 * Copyright 1998 Gerald Combs
 */

#ifndef IPA_BSCMGMT_MIB_H

static const value_string dn_vals[] = {
<xsl:for-each select="Class[Instantiable='true']">
	<xsl:sort select="DnReference"/>
	<xsl:if test="not (contains($ucletters, substring(Label, 1, 1)))">
	
		<xsl:text>    { </xsl:text>
		<xsl:value-of select="key('dn-index',DnReference)/RegisteredAs"/>
		<xsl:text>, "</xsl:text>
		<xsl:value-of select="DnReference"/>
		<xsl:text>" },
</xsl:text>
	</xsl:if>
</xsl:for-each>
	{ 0, NULL }
};

static const value_string obj_class_vals[] = {
<xsl:for-each select="Class">
		<xsl:if test="not (contains($ucletters, substring(Label, 1, 1)))">
			<xsl:text>    { </xsl:text>
			<xsl:value-of select="RegisteredAs"/>
			<xsl:text>, "</xsl:text>
			<xsl:value-of select="Label"/>
			<xsl:text>" },
</xsl:text>
		</xsl:if>
	</xsl:for-each>
	{ 0, NULL }
};

static const value_string attr_id_vals[] = {
<xsl:for-each select="Attribute">
<xsl:if test="not (contains($ucletters, substring(Label, 1, 1)))">
<xsl:text>    { </xsl:text>
<xsl:value-of select="RegisteredAs"/>
<xsl:text>, "</xsl:text>
<xsl:value-of select="Label"/>
<xsl:text>" },
</xsl:text>
</xsl:if>
</xsl:for-each>
	{ 0, NULL }
};

#define IPA_BSCMGMT_MIB_H
#endif
</redirect:write>
<redirect:close file="ipa-bscmgmt-mib.h"/>
</xsl:template>


<!--
     Utility templates
-->

<xsl:template name="tokenise-toupper">
  <xsl:param name="string" />
  <xsl:variable name="test-char" select="substring($string, 1, 1)" />
  <xsl:if test="$test-char">
  <xsl:choose>
    <xsl:when test="contains($ucletters, $test-char)">
      <xsl:text>_</xsl:text><xsl:value-of select="translate(substring($string, 1, 1), $lcletters, $ucletters)" />
      <xsl:call-template name="tokenise-toupper">
        <xsl:with-param name="string" select="substring($string, 2)" />
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="translate(substring($string, 1, 1), $lcletters, $ucletters)" />
      <xsl:call-template name="tokenise-toupper">
        <xsl:with-param name="string" select="substring($string, 2)" />
      </xsl:call-template>
    </xsl:otherwise>
  </xsl:choose>
  </xsl:if>
</xsl:template>

<xsl:template name="tokenise-tolower">
  <xsl:param name="string" />
  <xsl:variable name="test-char" select="substring($string, 1, 1)" />
  <xsl:if test="$test-char">
  <xsl:choose>
    <xsl:when test="contains($ucletters, $test-char)">
      <xsl:value-of select="substring($string, 1, 1)" />
      <xsl:call-template name="tokenise-tolower">
        <xsl:with-param name="string" select="substring($string, 2)" />
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="substring($string, 1, 1)" />
      <xsl:call-template name="tokenise-tolower">
        <xsl:with-param name="string" select="substring($string, 2)" />
      </xsl:call-template>
    </xsl:otherwise>
  </xsl:choose>
  </xsl:if>
</xsl:template>

<xsl:template name="initial-cap">
    <xsl:param name="string"/>
    <xsl:value-of select="translate(substring($string, 1, 1), $lcletters, $ucletters)" />
    <xsl:value-of select="substring($string,2)"/>
</xsl:template>

</xsl:stylesheet>
